c----------------------------------------------------------------------
c
c     Computer code for USGS Open-File Report 20121226:
c
c     Thermodynamic Method for Generating Random Stress Distributions
c     on an Earthquake Fault
c
c     by Michael Barall (Invisible Software Inc.)
c     and Ruth A. Harris (U.S. Geological Survey)
c
c     PUBLIC DOMAIN -- The software is in the public domain.  You are
c     free to use, copy, modify, and redistribute the software.
c     
c     NO WARRANTY -- The software is provided "AS IS" and without
c     warranty of any kind, expressed or implied, including warranties
c     of merchantability and fitness for a particular purpose.
c     
c     DISCLAIMER -- In no event will the USGS, Invisible Software, or
c     any other developer or distributor of the software be liable to
c     you for any damages arising out of the use or inability to use
c     the software (including but not limited to lost profits, lost
c     savings, lost data, inaccurate results, repair costs, incidental
c     or consequential damages, and losses sustained by third
c     parties), even if advised of the possibility of such damages.
c
c----------------------------------------------------------------------
c
c     Version 1.00
c
c----------------------------------------------------------------------
c
c     File genths.f -- Generate thermodynamic heterogeneous stress.
c
c     To run this program, you must use a Fortran 95 (or later version
c     of Fortran) compiler to compile and link the following files:
c       genths.f -- This file, the main program.
c       mt19937ar.f -- Mersenne Twister random number generator.
c     When compiling, you must set compiler options so that the Fortran
c     INTEGER data type is 64-bits or 8-bytes in size.
c
c     The program reads a response file from standard input.  The first
c     line of the response file contains a command, which specifies
c     which function is to be performed.  The remainder of the response
c     file contains parameters for the selected command.

      program genths
      
      implicit none
      
      call ths_gen_main()
      
      stop      
      end program genths








c----------------------------------------------------------------------
c
c     Fast Fourier Transform.








c----------------------------------------------------------------------
c
c     fft_ndim - Fast Fourier transform in n dimensions.
c     Parameters:
c      z = Data array.
c      nn = Array giving extent of the data array in each dimension.
c      ndim = Number of dimensions = Rank of data array.
c      isign = +1 for forward fft, -1 for reverse fft.
c     Each extent must be a power of 2.
c     In the data array, each adjacent pair of array elements holds
c      real and imaginary parts of a complex value (real part first).
c      The complex values are stored in column-major (Fortran) order,
c      that is, first index varies most rapidly.
c      (To use data stored in row-major (C/C++) order, reverse nn.)
c     The transformed data is stored in z, replacing the original.
c     The number N of complex values is the product of the elements
c      of nn.  The number of (real) elements of z is 2*N.
c     No scaling is applied on either forward or reverse fft.  So,
c      if you apply forward and then reverse fft (or vice versa) you
c      need to divide by N to recover the original values.
c     The sign convention is that forward fft is:
c      Z(k) = SUM(n)(z(n) * exp(2*pi*i*k.n/N))
c     and reverse fft is:
c      z(n) = SUM(k)(Z(k) * exp(-2*pi*i*k.n/N))
c     where k.n/N is the dot product of the index vectors k and n/N.
c     The notation n/N means (n1/N1, n2/N2, ...) where Ni = nn(i).
      
      subroutine fft_ndim (z, nn, ndim, isign)
      
      implicit none
      real(8), intent(inout) :: z(*)
      integer, intent(in) :: nn(ndim)
      integer, intent(in) :: ndim
      integer, intent(in) :: isign
      
      integer :: nsize, idim, i_stride, next_stride, i_extent
      integer :: i, i_rev, i_bit, m1, m2, k1, k2, j1, j2, i1, i2
      real(8) :: tr, ti
      real(8) :: ar, ai, br, bi, wr, wi, theta
      
      ! calculate total size of array, in elements
      
      nsize = 2    ! 2 elements for real and imaginary parts
      do idim = 1, ndim
        nsize = nsize*nn(idim)
      enddo
      
      ! loop over dimensions
      
      next_stride = 2
      
      do idim = 1, ndim
        i_stride = next_stride    ! element stride in current dim
        i_extent = nn(idim)    ! element extent in current dim
        next_stride = i_stride*i_extent   ! element stride in next dim
        
        ! bit reversal loop
        
        i_rev = i_extent/2    ! value of i, bit-reversed
        do i = 1, i_extent-2
        
          if (i .lt. i_rev) then
            do m1 = 0, nsize - next_stride, next_stride
              do m2 = 0, i_stride - 2, 2
                k1 = i*i_stride + m1 + m2 + 1
                k2 = i_rev*i_stride + m1 + m2 + 1
                
                ! swap elements at bit-reversed indexes
                
                tr = z(k1)
                ti = z(k1 + 1)
                z(k1) = z(k2)
                z(k1 + 1) = z(k2 + 1)
                z(k2) = tr
                z(k2 + 1) = ti
                
              enddo
            enddo
          endif
          
          ! increment i_rev, with bits reversed
          
          i_bit = i_extent/2
          do while (i_bit .le. i_rev)
            i_rev = i_rev - i_bit
            i_bit = i_bit/2
          enddo
          i_rev = i_rev + i_bit
        
        enddo    ! end bit reversal loop
        
        ! fft loop
        
        theta = 6.283185307179586476925286766559d0    ! 2*pi
        if (isign .lt. 0) then
          theta = -theta
        endif
        
        j2 = 1
        do while (j2 .lt. i_extent)
          j1 = j2
          j2 = j2*2
          
          theta = theta * 0.5d0
          
          ar = sin(theta * 0.5d0)
          ar = -2.0d0 * ar * ar    ! ar = cos(theta) - 1
          ai = sin(theta)
          
          br = 1.0d0
          bi = 0.0d0
          
          do i1 = 0, j1 - 1
            do i2 = 0, i_extent - j2, j2
              do m1 = 0, nsize - next_stride, next_stride
                do m2 = 0, i_stride - 2, 2
                  k1 = (i1 + i2)*i_stride + m1 + m2 + 1
                  k2 = (i1 + i2 + j1)*i_stride + m1 + m2 + 1
                  
                  ! butterfly transform
                  
                  tr = br*z(k2) - bi*z(k2 + 1)
                  ti = br*z(k2 + 1) + bi*z(k2)
                  z(k2) = z(k1) - tr
                  z(k2 + 1) = z(k1 + 1) - ti
                  z(k1) = z(k1) + tr
                  z(k1 + 1) = z(k1 + 1) + ti
                  
                enddo
              enddo
            enddo
            
            wr = br*ar - bi*ai
            wi = br*ai + bi*ar
            br = br + wr
            bi = bi + wi
            
          enddo
        
        enddo    ! end fft loop
      
      enddo    ! end loop over dimensions
      
      return
      end subroutine fft_ndim








c----------------------------------------------------------------------
c
c     test_fft_2dim - Test fast Fourier transform in 2 dimensions.
c     Parameters:
c      iret = Return code, 0 = failure, 1 = success.
c      ext_1 = Extent in dimension #1, must be a power of 2.
c      ext_2 = Extent in dimension #2, must be a power of 2.
c      init_1 = Initial index in dimension #1, range 0 to ext_1-1.
c      init_2 = Initial index in dimension #2, range 0 to ext_2-1.
c      step_1 = Amount to step index in dimension #1, range 0 to ext_1-1.
c      step_2 = Amount to step index in dimension #2, range 0 to ext_2-1.
c      n_test = Number of tests to run.
c     The function performs the 2-dimension FFT n_test times,
c     alternating between forward and reverse FFT.  Each FFT input has
c     one data element set to a non-zero value, with all others zero.
c     The first test uses data element (init_1,init2).  Each subsequent
c     test advances the data element index by (step_1,step2).  A complex
c     amplitude and phase is generated for the data element value.  The
c     function checks the entire output.
      
      subroutine test_fft_2dim (
     &  iret, ext_1, ext_2, init_1, init_2, step_1, step_2, n_test)
      
      implicit none
      integer, intent(out) :: iret
      integer, intent(in) :: ext_1
      integer, intent(in) :: ext_2
      integer, intent(in) :: init_1
      integer, intent(in) :: init_2
      integer, intent(in) :: step_1
      integer, intent(in) :: step_2
      integer, intent(in) :: n_test
      
      ! Data array
      
      real(8), allocatable :: z(:,:,:)
      integer, parameter :: i_re = 0
      integer, parameter :: i_im = 1
      
      ! Array to hold extents of each dimension
      
      integer :: ndim
      integer :: nn(2)
      
      ! Sign flag, +1 = forward FFT, -1 = reverse FFT
      
      integer :: isign
      
      ! Current index number
      
      integer :: n_1
      integer :: n_2
      
      ! Current test number
      
      integer :: i_test
      
      ! Wave number being checked
      
      integer :: k_1
      integer :: k_2
      
      ! Amplitude and phase
      
      real(8) :: amp
      real(8) :: phase
      
      ! Complex delta from expected value
      
      real(8) :: delta(0:1)
      
      ! Argument from fft formula
      
      real(8) :: fftarg
      
      ! Tolerance
      
      real(8), parameter :: tol = 1.0d-8
      
      ! Constant 2*pi
      
      real(8), parameter :: two_pi = 6.283185307179586476925286766559d0
      
      ! Assume failure
      
      iret = 0
      
      ! Allocate the data array
      
      allocate (z(0:1, 0:ext_1-1, 0:ext_2-1))
      
      ! Set up extents
      
      ndim = 2
      nn(1) = ext_1
      nn(2) = ext_2
      
      ! Initialize sign flag
      
      isign = 1
      
      ! Initialize current index number
      
      n_1 = init_1
      n_2 = init_2
      
      ! Loop over tests
      
      do i_test = 1, n_test
        
        ! Clear the data array
        
        do k_1 = 0, ext_1-1
          do k_2 = 0, ext_2-1
            z(i_re, k_1, k_2) = 0.0d0
            z(i_im, k_1, k_2) = 0.0d0
          enddo
        enddo
        
        ! Select an amplitude and phase
        
        amp = dble(n_1 * n_2 + 1)
!!        phase = two_pi * dble(i_test - 1) / dble(n_test)
        phase = two_pi * dble(i_test) / dble(n_test + 1)
        
        ! Insert complex number into current index
        
        z(i_re, n_1, n_2) = amp * cos(phase)
        z(i_im, n_1, n_2) = amp * sin(phase)
        
        ! Perform the Fourier transform
        
        call fft_ndim (z, nn, ndim, isign)
        
        ! Loop to check every result value
        
        do k_1 = 0, ext_1-1
          do k_2 = 0, ext_2-1
            
            ! Calculate argument from fft formula, including our phase
            
            fftarg = phase + two_pi * dble(isign) * (
     &          dble(mod(n_1 * k_1, ext_1)) / dble(ext_1)
     &        + dble(mod(n_2 * k_2, ext_2)) / dble(ext_2) )
            
            ! Calculate delta from expected value
            
            delta(i_re) = z(i_re, k_1, k_2) - amp * cos(fftarg)
            delta(i_im) = z(i_im, k_1, k_2) - amp * sin(fftarg)
            
            ! Check that delta is within tolerance
            
            if (sqrt(delta(i_re)*delta(i_re) + delta(i_im)*delta(i_im))
     &          > tol * amp) then
              
              ! Return failure
              
              iret = 0
              return
              
            endif
          
          enddo
        enddo
        
        ! Flip the sign flag
        
        isign = -isign
        
        ! Advance to next index
        
        n_1 = mod(n_1 + step_1, ext_1)
        n_2 = mod(n_2 + step_2, ext_2)
        
      enddo  ! end loop over tests

      ! Deallocate the data array
      
      deallocate (z) 
      
      ! Return success
      
      iret = 1     
      
      return
      end subroutine test_fft_2dim








c----------------------------------------------------------------------
c
c     I/O Unit Management.




c     io_unit_data - Data module.

      module io_unit_data
      implicit none
      
      ! Standard unit numbers
      
      integer, parameter :: stdin = 5
      integer, parameter :: stdout = 6
      integer, parameter :: stderr = 0
      
      ! Character string size to hold a file and path
      
      integer, parameter :: max_path = 512
      
      ! Structure to hold an available unit number
      
      type, private :: unit_list_item
        type(unit_list_item), pointer :: next  ! Next item in list
        integer :: iunit  ! unit number
      end type unit_list_item
      
      ! Linked list of empty structures
      
      type(unit_list_item), pointer, private, save ::
     &   unit_list_empty => null()
      
      ! Linked list of available unit numbers
      
      type(unit_list_item), pointer, private, save ::
     &   unit_list_available => null()
      
      ! Next unused unit number
      
      integer, private, save :: unit_next_unused = 20
      
      ! Procedure access
      
      public alloc_unit
      public free_unit
      
      public open_unit_r
      public open_unit_w
      public close_unit
      
      contains
      
      
      
      
c     alloc_unit - Allocate a new unit number.
c     Parameters:
c      iunit = Variable to receive Fortran unit number.

      subroutine alloc_unit (iunit)
      implicit none
      
      integer, intent(out) :: iunit
      
      type(unit_list_item), pointer :: p
      
      ! If there is an available unit number ...
      
      if (associated(unit_list_available)) then
      
        ! Remove the structure from the list
        
        p => unit_list_available
        unit_list_available => p%next
        
        ! Return the unit number
        
        iunit = p%iunit
        
        ! Put structure on the empty list
        
        p%next => unit_list_empty
        unit_list_empty => p
        
      ! Otherwise, no available unit number ...
      
      else
      
        ! Return the next available unit number
        
        iunit = unit_next_unused
        unit_next_unused = unit_next_unused + 1
        
      endif
      
      return
      end subroutine alloc_unit
      
      
      
      
c     free_unit - Free a unit number.
c     Parameters:
c      iunit = Fortran unit number.
c     The unit number must have been obtained from alloc_unit.

      subroutine free_unit (iunit)
      implicit none
      
      integer, intent(in) :: iunit
      
      type(unit_list_item), pointer :: p
      
      ! If there is an empty structure ...
      
      if (associated(unit_list_empty)) then
      
        ! Remove the structure from the list
        
        p => unit_list_empty
        unit_list_empty => p%next
        
      ! Otherwise, no empty structure ...
      
      else
      
        ! Allocate a new structure
        
        allocate(p)
        
      endif
      
      ! Save the unit number
      
      p%iunit = iunit
      
      ! Put structure on the available list
      
      p%next => unit_list_available
      unit_list_available => p
      
      return
      end subroutine free_unit
      
      
      
      
c     open_unit_r - Allocate a new unit number, and open file for reading.
c     Parameters:
c      iunit = Variable to receive Fortran unit number.
c      filename = Name of file to open.

      subroutine open_unit_r (iunit, filename)
      implicit none
      
      integer, intent(out) :: iunit
      character(*), intent(in) :: filename
      
      ! Allocate a unit number
      
      call alloc_unit (iunit)
      
      ! Open the file
      
      open(unit=iunit,status='old',form='formatted',
     &     file=filename)
      
      return
      end subroutine open_unit_r
      
      
      
      
c     open_unit_w - Allocate a new unit number, and open file for writing.
c     Parameters:
c      iunit = Variable to receive Fortran unit number.
c      filename = Name of file to open.

      subroutine open_unit_w (iunit, filename)
      implicit none
      
      integer, intent(out) :: iunit
      character(*), intent(in) :: filename
      
      ! Allocate a unit number
      
      call alloc_unit (iunit)
      
      ! Open the file
      
      open(unit=iunit,status='replace',form='formatted',
     &     file=filename)
      
      return
      end subroutine open_unit_w
      
      
      
      
c     close_unit - Close a file, and free its unit number.
c     Parameters:
c      iunit = Fortran unit number.

      subroutine close_unit (iunit)
      implicit none
      
      integer, intent(in) :: iunit
      
      ! Close the file
      
      close(unit=iunit)
      
      ! Free the unit number
      
      call free_unit (iunit)
      
      return
      end subroutine close_unit
      
      
      
      
c     make_realiz_filename - Make a filename containing a realization number.
c     Parameters:
c      filename = Variable to receive the constructed filename.
c      basename = Base filename, used as a prefix.
c      realiz = Realization number (non-negative, 5 digits).
c      sub_realiz = Sub-realization number (non-negative, 3 digits).
c      postname = Postfix filename to use.
c     A filename is constructed by combining the basename, realization
c      number, sub-realization number, and postname.
c     If sub_realiz is negative, it is not used.
c     If realiz is negative, neither it nor sub_realiz is used.

      subroutine make_realiz_filename (
     &  filename, basename, realiz, sub_realiz, postname)
      implicit none
      
      character(*), intent(out) :: filename
      character(*), intent(in)  :: basename
      integer, intent(in) :: realiz
      integer, intent(in) :: sub_realiz
      character(*), intent(in) :: postname
      
      ! Construct a combined name
      
      if (realiz < 0) then
      
        write(filename,'(A,A,A)')
     &    basename(1:len_trim(basename)),
     &    '_',
     &    postname(1:len_trim(postname))
      
      else if (sub_realiz < 0) then
      
        write(filename,'(A,A,I5.5,A,A)')
     &    basename(1:len_trim(basename)),
     &    '_',
     &    realiz,
     &    '_',
     &    postname(1:len_trim(postname))
      
      else
      
        write(filename,'(A,A,I5.5,A,I3.3,A,A)')
     &    basename(1:len_trim(basename)),
     &    '_',
     &    realiz,
     &    '_',
     &    sub_realiz,
     &    '_',
     &    postname(1:len_trim(postname))
      
      endif
      
      return
      end subroutine make_realiz_filename
      
      
      
      
      end module io_unit_data








c----------------------------------------------------------------------
c
c     Global constants.




c     math_constants - Mathematical constants module.

      module math_constants
      implicit none
      
      ! Mathematical constant pi
      
      real(8), parameter :: pi
     &                      = 3.1415926535897932384626433832795d0
      
      ! Mathematical constant 2*pi
      
      real(8), parameter :: two_pi
     &                      = 6.283185307179586476925286766559d0
      
      ! Mathematical constant 4*pi
      
      real(8), parameter :: four_pi
     &                      = 1.2566370614359172953850573533118d1
      
      ! Mathematical constant e
      
      real(8), parameter :: euler_e
     &                      = 2.7182818284590452353602874713527d0
      
      ! Index values for real and imaginary parts of a complex number
      
      integer, parameter :: i_re = 0
      integer, parameter :: i_im = 1
      
      contains
      
      
      
      
c     lin_interp - Linear interpolation
c     Parameters:
c      x = Argument value at which to compute interpolated value
c      x1 = Given argument #1
c      x2 = Given argument #2
c      y1 = Given value #1
c      y2 = Given value #2
c     Returns:
c      Interpolated value at x.
c     Note: The function assumes x1 /= x2.

      function lin_interp (x, x1, x2, y1, y2)
      implicit none
      
      real(8) :: lin_interp
      
      real(8), intent(in) :: x
      real(8), intent(in) :: x1
      real(8), intent(in) :: x2
      real(8), intent(in) :: y1
      real(8), intent(in) :: y2
      
      ! Interpolate
      
      lin_interp = ((x-x1)*y2 + (x2-x)*y1) / (x2-x1)
      
      end function lin_interp



      
      end module math_constants








c----------------------------------------------------------------------
c
c     rgen - Random number generator initialization.




c     rgen_data - Data module.

      module rgen_data
      implicit none
      
      ! Realization number
      
      integer, save :: rgen_realiz
      
      end module rgen_data




c     rgen_read_param - Read parameters for random number generator.
c     Parameters:
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       Realization number
c       Seed filename
c
c     Note: The seed file is assumed to contain 11 integers per line.
c     The realization number is the 0-based line number in the seed file.

      subroutine rgen_read_param (iunit, f_ok)
      
      use io_unit_data
      use rgen_data
      implicit none
      
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: rand_seed(11)
      character(max_path) :: seed_filename
      integer :: i
      
      integer :: seed_unit
      
      ! Read realization number
      
      read(iunit,*) rgen_realiz
      
      if (rgen_realiz < 0) then
        f_ok = .false.
        return
      endif
      
      ! Read seed filename
      
      read(iunit,*) seed_filename
      
      ! Open the random seed file and position it
      
      call open_unit_r (seed_unit, seed_filename)
      
      do i = 1, rgen_realiz
        read(seed_unit,*)
      enddo
      
      ! Set the random seed
      
      read(seed_unit,*) (rand_seed(i), i = 1,11)
      
      call init_by_array (rand_seed, 11)
      
      ! Close the seed file
      
      call close_unit (seed_unit)
      
      f_ok = .true.
      
      return
      end subroutine rgen_read_param




c     rgen_get_realiz - Get the realization number.
c     Parameters:
c      realiz = Variable to receive the realization number.

      subroutine rgen_get_realiz (realiz)
      
      use rgen_data
      implicit none
      
      integer, intent(out) :: realiz
      
      ! Return the realization number
      
      realiz = rgen_realiz
      
      return
      end subroutine rgen_get_realiz








c----------------------------------------------------------------------
c
c     value1d - Value that varies in 1 dimension.




c     value1d_include - Module to include.

      module value1d_include
      implicit none
      
      
      
      
      !===== Private data definitions =====
      
      
      
      
      ! value1d_constant - Constant value.
      
      type, private :: value1d_constant
      
        ! The value
        
        real(8) :: v
      
      end type value1d_constant
      
      
      
      
      ! value1d_pwlinear_item - Item for piecewise-linear list.
      
      type, private :: value1d_pwlinear_item
      
        ! Argument
        
        real(8) :: x
        
        ! Value on low side of argument
        
        real(8) :: v_lo
        
        ! Value on high side of argument
        
        real(8) :: v_hi
      
      end type value1d_pwlinear_item
      
      
      
      
      ! value1d_pwlinear - Piecewise linear value.
      
      type, private :: value1d_pwlinear
      
        ! Number of items, must be >= 1
        
        integer :: item_count
        
        ! List of items
        
        type(value1d_pwlinear_item), pointer :: item_list(:)
      
      end type value1d_pwlinear
      
      
      
      
      ! Virtual type codes
      
      integer, parameter, private :: value1d_vt_constant = 1
      integer, parameter, private :: value1d_vt_pwlinear = 2
      
      
      
      
      !===== Public data definitions =====
      
      
      
      
      ! value1d - Polymorphic value.
      
      type :: value1d
      
        ! Virtual type code
        
        integer :: vt_code
        
        ! Pointers for various types
        
        type(value1d_constant), pointer :: p_constant
        type(value1d_pwlinear), pointer :: p_pwlinear
        
      end type value1d
      
      
      
      
      !===== Subroutine access declarations =====
      
      
      
      
      private value1d_constant_new
      private value1d_constant_delete
      private value1d_constant_eval
      
      private value1d_pwlinear_new
      private value1d_pwlinear_delete
      private value1d_pwlinear_eval
      
      public value1d_new
      public value1d_delete
      public value1d_eval
      
      
      
      
      contains
      
      
      
      
      !===== Private subroutines =====
      
      
      
      
c     value1d_constant_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       value
      
      subroutine value1d_constant_new (this, iunit, f_ok)
      
      implicit none
      
      type(value1d_constant), pointer :: this
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: i
      
      ! Allocate the object
      
      allocate(this)
      
      ! Read value
        
      read(iunit,*) this%v
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine value1d_constant_new
      
      
      
      
c     value1d_constant_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine value1d_constant_delete (this)
      
      implicit none
      
      type(value1d_constant), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine value1d_constant_delete
      
      
      
      
c     value1d_constant_eval - Evaluate.
c     Parameters:
c      this = Pointer.
c      v = Variable to receive value.
c      x = Argument.
      
      subroutine value1d_constant_eval (this, v, x)
      
      implicit none
      
      type(value1d_constant), pointer :: this
      real(8), intent(out) :: v
      real(8), intent(in) :: x
      
      ! Return value
      
      v = this%v
      
      return
      end subroutine value1d_constant_eval
      
      
      
      
c     value1d_pwlinear_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       count
c       x, v_lo, v_hi  ["count" lines]
c
c     Note: Argument x must appear in increasing order.
      
      subroutine value1d_pwlinear_new (this, iunit, f_ok)
      
      implicit none
      
      type(value1d_pwlinear), pointer :: this
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: i
      
      ! Allocate the object
      
      allocate(this)
      this%item_list => null()
      
      ! Read size of list
        
      read(iunit,*) this%item_count
      
      if (this%item_count < 1) then
        call value1d_pwlinear_delete (this)
        f_ok = .false.
        return
      endif
      
      ! Allocate and read the list
      
      allocate(this%item_list(this%item_count))
        
      do i = 1, this%item_count
        read(iunit,*)
     &    this%item_list(i)%x,
     &    this%item_list(i)%v_lo,
     &    this%item_list(i)%v_hi
      enddo
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine value1d_pwlinear_new
      
      
      
      
c     value1d_pwlinear_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine value1d_pwlinear_delete (this)
      
      implicit none
      
      type(value1d_pwlinear), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
      
        ! Deallocate our held objects
        
        if (associated(this%item_list)) then
          deallocate(this%item_list)
        endif
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine value1d_pwlinear_delete
      
      
      
      
c     value1d_pwlinear_eval - Evaluate.
c     Parameters:
c      this = Pointer.
c      v = Variable to receive value.
c      x = Argument.
      
      subroutine value1d_pwlinear_eval (this, v, x)
      
      use math_constants
      implicit none
      
      type(value1d_pwlinear), pointer :: this
      real(8), intent(out) :: v
      real(8), intent(in) :: x
      
      integer :: i_lo
      integer :: i_hi
      integer :: i_mid
      
!      real(8) :: c
      
      ! Binary search to find interval containing argument
      
      i_lo = 0
      i_hi = this%item_count + 1
      
      do while (i_hi - i_lo > 1)
        i_mid = (i_lo + i_hi)/2
        if (x < this%item_list(i_mid)%x) then
          i_hi = i_mid
        else
          i_lo = i_mid
        endif
      enddo
      
      ! If below range of list ...
      
      if (i_lo < 1) then
        
        ! Use low value of first list entry
        
        v = this%item_list(1)%v_lo
        
      ! Otherwise, if above range of list ...
      
      else if (i_lo >= this%item_count) then
        
        ! Use high value of last list entry
        
        v = this%item_list(this%item_count)%v_hi
        
      ! Otherwise, within range of list ...
      
      else
        
        ! Interpolate within values
        
!        c = (x - this%item_list(i_lo)%x)
!     &        / (this%item_list(i_lo + 1)%x - this%item_list(i_lo)%x)
!        
!        v = (1.0 - c)*this%item_list(i_lo)%v_hi
!     &        + c*this%item_list(i_lo + 1)%v_lo
      
        v = lin_interp (
     &        x, this%item_list(i_lo)%x, this%item_list(i_lo + 1)%x,
     &        this%item_list(i_lo)%v_hi, this%item_list(i_lo + 1)%v_lo)
        
      endif
      
      return
      end subroutine value1d_pwlinear_eval
      
      
      
      
      !===== Public subroutines =====
      
      
      
      
c     value1d_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       virtual type code
c       parameters for specified type  [X lines]
      
      subroutine value1d_new (this, iunit, f_ok)
      
      implicit none
      
      type(value1d), pointer :: this
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: i
      
      ! Allocate the object
      
      allocate(this)
      this%p_constant => null()
      this%p_pwlinear => null()
      
      ! Read virtual type code
        
      read(iunit,*) this%vt_code
      
      ! Construct object for selected type
      
      select case (this%vt_code)
      
      case (value1d_vt_constant)
        call value1d_constant_new (this%p_constant, iunit, f_ok)
      
      case (value1d_vt_pwlinear)
        call value1d_pwlinear_new (this%p_pwlinear, iunit, f_ok)
      
      case default
        f_ok = .false.
      
      end select
      
      ! Check result
      
      if (.not. f_ok) then
        call value1d_delete (this)
        this => null()
      endif
      
      return
      end subroutine value1d_new
      
      
      
      
c     value1d_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine value1d_delete (this)
      
      implicit none
      
      type(value1d), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
      
        ! Deallocate our held objects
        
        call value1d_constant_delete (this%p_constant)
        call value1d_pwlinear_delete (this%p_pwlinear)
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine value1d_delete
      
      
      
      
c     value1d_eval - Evaluate.
c     Parameters:
c      this = Pointer.
c      v = Variable to receive value.
c      x = Argument.
      
      subroutine value1d_eval (this, v, x)
      
      implicit none
      
      type(value1d), pointer :: this
      real(8), intent(out) :: v
      real(8), intent(in) :: x
      
      ! Return value for selected type
      
      select case (this%vt_code)
      
      case (value1d_vt_constant)
        call value1d_constant_eval (this%p_constant, v, x)
      
      case (value1d_vt_pwlinear)
        call value1d_pwlinear_eval (this%p_pwlinear, v, x)
      
      end select
      
      return
      end subroutine value1d_eval
      


      
      end module value1d_include








c----------------------------------------------------------------------
c
c     value2d - Value that varies in 1 dimension.




c     value2d_include - Module to include.

      module value2d_include
      
      use value1d_include
      implicit none
      
      
      
      
      !===== Private data definitions =====
      
      
      
      
      ! value2d_constant - Constant value.
      
      type, private :: value2d_constant
      
        ! The value
        
        real(8) :: v
      
      end type value2d_constant
      
      
      
      
      ! value2d_product - Product of two 1d values.
      
      type, private :: value2d_product
      
        ! Value added to the product
        
        real(8) :: v_offset
      
        ! The x and y values
        
        type(value1d), pointer :: p_val_x
        type(value1d), pointer :: p_val_y
      
      end type value2d_product
      
      
      
      
      ! value2d_radial - Radially-applied 1d value.
      
      type, private :: value2d_radial
      
        ! The radial value
        
        type(value1d), pointer :: p_val_r
      
      end type value2d_radial
      
      
      
      
      ! Virtual type codes
      
      integer, parameter, private :: value2d_vt_constant = 1
      integer, parameter, private :: value2d_vt_product = 2
      integer, parameter, private :: value2d_vt_radial = 3
      
      
      
      
      !===== Public data definitions =====
      
      
      
      
      ! value2d - Polymorphic value.
      
      type :: value2d
      
        ! Virtual type code
        
        integer :: vt_code
        
        ! Linear transformation for x
        
        real(8) :: mxx
        real(8) :: mxy
        real(8) :: bx
        
        ! Linear transformation for y
        
        real(8) :: myx
        real(8) :: myy
        real(8) :: by
        
        ! Pointers for various types
        
        type(value2d_constant), pointer :: p_constant
        type(value2d_product), pointer :: p_product
        type(value2d_radial), pointer :: p_radial
        
      end type value2d
      
      
      
      
      !===== Subroutine access declarations =====
      
      
      
      
      private value2d_constant_new
      private value2d_constant_delete
      private value2d_constant_eval
      
      private value2d_product_new
      private value2d_product_delete
      private value2d_product_eval
      
      private value2d_radial_new
      private value2d_radial_delete
      private value2d_radial_eval
      
      public value2d_new
      public value2d_delete
      public value2d_eval
      
      
      
      
      contains
      
      
      
      
      !===== Private subroutines =====
      
      
      
      
c     value2d_constant_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       value
      
      subroutine value2d_constant_new (this, iunit, f_ok)
      
      implicit none
      
      type(value2d_constant), pointer :: this
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: i
      
      ! Allocate the object
      
      allocate(this)
      
      ! Read value
        
      read(iunit,*) this%v
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine value2d_constant_new
      
      
      
      
c     value2d_constant_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine value2d_constant_delete (this)
      
      implicit none
      
      type(value2d_constant), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine value2d_constant_delete
      
      
      
      
c     value2d_constant_eval - Evaluate.
c     Parameters:
c      this = Pointer.
c      v = Variable to receive value.
c      x = Argument x.
c      y = Argument y.
      
      subroutine value2d_constant_eval (this, v, x, y)
      
      implicit none
      
      type(value2d_constant), pointer :: this
      real(8), intent(out) :: v
      real(8), intent(in) :: x
      real(8), intent(in) :: y
      
      ! Return value
      
      v = this%v
      
      return
      end subroutine value2d_constant_eval
      
      
      
      
c     value2d_product_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       v_offset
c       x 1d value  [X lines]
c       y 1d value  [X lines]
      
      subroutine value2d_product_new (this, iunit, f_ok)
      
      implicit none
      
      type(value2d_product), pointer :: this
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      ! Allocate the object
      
      allocate(this)
      this%p_val_x => null()
      this%p_val_y => null()
      
      ! Read offset value
        
      read(iunit,*) this%v_offset
      
      ! Read x value
      
      call value1d_new (this%p_val_x, iunit, f_ok)
      
      if (.not. f_ok) then
        call value2d_product_delete (this)
        return
      endif
      
      ! Read y value
      
      call value1d_new (this%p_val_y, iunit, f_ok)
      
      if (.not. f_ok) then
        call value2d_product_delete (this)
        return
      endif
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine value2d_product_new
      
      
      
      
c     value2d_product_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine value2d_product_delete (this)
      
      implicit none
      
      type(value2d_product), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
      
        ! Deallocate our held objects
        
        call value1d_delete (this%p_val_x)
        call value1d_delete (this%p_val_y)
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine value2d_product_delete
      
      
      
      
c     value2d_product_eval - Evaluate.
c     Parameters:
c      this = Pointer.
c      v = Variable to receive value.
c      x = Argument.
      
      subroutine value2d_product_eval (this, v, x, y)
      
      implicit none
      
      type(value2d_product), pointer :: this
      real(8), intent(out) :: v
      real(8), intent(in) :: x
      real(8), intent(in) :: y
      
      real(8) :: vx
      real(8) :: vy
      
      ! Get the x and y values
      
      call value1d_eval (this%p_val_x, vx, x)
      call value1d_eval (this%p_val_y, vy, y)
      
      ! Return the product plus offset
      
      v = (vx * vy) + this%v_offset
      
      return
      end subroutine value2d_product_eval
      
      
      
      
c     value2d_radial_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       radial 1d value  [X lines]
      
      subroutine value2d_radial_new (this, iunit, f_ok)
      
      implicit none
      
      type(value2d_radial), pointer :: this
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      ! Allocate the object
      
      allocate(this)
      this%p_val_r => null()
      
      ! Read radial value
      
      call value1d_new (this%p_val_r, iunit, f_ok)
      
      if (.not. f_ok) then
        call value2d_radial_delete (this)
        return
      endif
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine value2d_radial_new
      
      
      
      
c     value2d_radial_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine value2d_radial_delete (this)
      
      implicit none
      
      type(value2d_radial), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
      
        ! Deallocate our held objects
        
        call value1d_delete (this%p_val_r)
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine value2d_radial_delete
      
      
      
      
c     value2d_radial_eval - Evaluate.
c     Parameters:
c      this = Pointer.
c      v = Variable to receive value.
c      x = Argument.
      
      subroutine value2d_radial_eval (this, v, x, y)
      
      implicit none
      
      type(value2d_radial), pointer :: this
      real(8), intent(out) :: v
      real(8), intent(in) :: x
      real(8), intent(in) :: y
      
      real(8) :: r
      
      ! Get the radial value
      
      r = sqrt(x*x + y*y)
      
      call value1d_eval (this%p_val_r, v, r)
      
      return
      end subroutine value2d_radial_eval
      
      
      
      
      !===== Public subroutines =====
      
      
      
      
c     value2d_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       virtual type code
c       mxx, mxy, bx
c       myx, myy, by
c       parameters for specified type  [X lines]
      
      subroutine value2d_new (this, iunit, f_ok)
      
      implicit none
      
      type(value2d), pointer :: this
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: i
      
      ! Allocate the object
      
      allocate(this)
      this%p_constant => null()
      this%p_product => null()
      this%p_radial => null()
      
      ! Read virtual type code
        
      read(iunit,*) this%vt_code
      
      ! Read x linear transformation
        
      read(iunit,*) this%mxx, this%mxy, this%bx
      
      ! Read y linear transformation
        
      read(iunit,*) this%myx, this%myy, this%by
      
      ! Construct object for selected type
      
      select case (this%vt_code)
      
      case (value2d_vt_constant)
        call value2d_constant_new (this%p_constant, iunit, f_ok)
      
      case (value2d_vt_product)
        call value2d_product_new (this%p_product, iunit, f_ok)
      
      case (value2d_vt_radial)
        call value2d_radial_new (this%p_radial, iunit, f_ok)
      
      case default
        f_ok = .false.
      
      end select
      
      ! Check result
      
      if (.not. f_ok) then
        call value2d_delete (this)
        this => null()
      endif
      
      return
      end subroutine value2d_new
      
      
      
      
c     value2d_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine value2d_delete (this)
      
      implicit none
      
      type(value2d), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
      
        ! Deallocate our held objects
        
        call value2d_constant_delete (this%p_constant)
        call value2d_product_delete (this%p_product)
        call value2d_radial_delete (this%p_radial)
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine value2d_delete
      
      
      
      
c     value2d_eval - Evaluate.
c     Parameters:
c      this = Pointer.
c      v = Variable to receive value.
c      x = Argument.
      
      subroutine value2d_eval (this, v, x, y)
      
      implicit none
      
      type(value2d), pointer :: this
      real(8), intent(out) :: v
      real(8), intent(in) :: x
      real(8), intent(in) :: y
      
      real(8) :: tx
      real(8) :: ty
      
      ! Apply x and y linear transformations
      
      tx = this%mxx * x + this%mxy * y + this%bx
      ty = this%myx * x + this%myy * y + this%by
      
      ! Return value for selected type
      
      select case (this%vt_code)
      
      case (value2d_vt_constant)
        call value2d_constant_eval (this%p_constant, v, tx, ty)
      
      case (value2d_vt_product)
        call value2d_product_eval (this%p_product, v, tx, ty)
      
      case (value2d_vt_radial)
        call value2d_radial_eval (this%p_radial, v, tx, ty)
      
      end select
      
      return
      end subroutine value2d_eval
      


      
      end module value2d_include








c----------------------------------------------------------------------
c
c     rolloff - Energy-per-mode rolloff function.




c     rolloff_data - Data module.

      module rolloff_data
      implicit none
      
      ! Rolloff formula selection
      
      integer, save :: rolo_form
      
      ! Formula = 1
      
      integer, parameter :: rolo_form_none = 1
      
      ! Formula = 1/(1 + (ell*k)^3 * (log(1 + zeta*k))^2)
      
      integer, parameter :: rolo_form_logsq = 2
      
      ! Parameter ell = rolloff wavelength / (2*pi)
      
      real(8), save :: rolo_ell
      
      ! Parameter zeta = seismogenic wavelength / (2*pi)
      
      real(8), save :: rolo_zeta
      
      ! Integral of (k^2)*g(k), or 0 if not convergent
      ! If g is not isotropic, the 2D integral of k*g(k0,k1)/(2*pi)
      
      real(8), save :: rolo_integ_ksq
      
      ! Correlation wavelength = (2*pi)/k where k is the wavenumber
      !                          where (k^2)*g(k) reaches its maximum
      
      real(8), save :: rolo_corr_lambda
      
      ! Mean-square shear stress, in the continuum
      
      real(8), save :: rolo_tau_ms_cont
      
      end module rolloff_data




c     rolloff_calc_g - Calculate the rolloff g function.
c     Parameters:
c      g = Variable to receive the value of the g function.
c      k0 = Wavenumber in direction 0 (perpendicular to slip).
c      k1 = Wavenumber in direction 1 (parallel to slip).
c     Wavenumbers equal 2*pi/wavelength, in units of radians/meter.
c     The g function is the fraction of the unconstrained energy per
c      mode that should be used for the given wavenumbers.  It ranges
c      from 0 to 1, with 1 meaning to use the unconstrained energy.
c     Note: Symmetry considerations require g(k0,k1) to depend only on
c      the absolute values of k0 and k1, that is,
c           g(k0,k1) == g(-k0,k1) == g(k0,-k1) == g(-k0,-k1)
c      This is always true if g depends only on k = sqrt(k0^2 + k1^2).

      subroutine rolloff_calc_g (g, k0, k1)
      
      use rolloff_data
      implicit none
      
      real(8), intent(out) :: g
      real(8), intent(in) :: k0
      real(8), intent(in) :: k1
      
      real(8) :: k  ! wavenumber magnitude
      real(8) :: ell_k  ! ell times k
      real(8) :: log_term  ! logarithmic term
      
      ! Use selected formula
      
      select case (rolo_form)
      
      ! No rolloff
      
      case (rolo_form_none)
        g = 1.0d0
      
      ! Log-square formula
      
      case (rolo_form_logsq)
        k = sqrt(k0*k0 + k1*k1)
        ell_k = rolo_ell * k
        log_term = log(1.0d0 + (rolo_zeta * k))
        g = 1.0d0 / (1.0d0 + (ell_k*ell_k*ell_k*log_term*log_term))
      
      end select
      
      return
      end subroutine rolloff_calc_g




c     rolloff_get_integ_ksq - Get the integral of (k^2)*g(k).
c     Parameters:
c      integ_ksq = Variable to receive integral of (k^2)*g(k).
c     If g is non-isotropic, the returned value is the 2D integral
c      of k*g(k0,k1)/(2*pi).

      subroutine rolloff_get_integ_ksq (integ_ksq)
      
      use rolloff_data
      implicit none
      
      real(8), intent(out) :: integ_ksq
      
      ! Return the integral of (k^2)*g(k)
      
      integ_ksq = rolo_integ_ksq
      
      return
      end subroutine rolloff_get_integ_ksq




c     rolloff_get_roll_lambda - Get the rolloff wavelength.
c     Parameters:
c      roll_lambda = Variable to receive rolloff wavelength, in meters.

      subroutine rolloff_get_roll_lambda (roll_lambda)
      
      use math_constants
      use rolloff_data
      implicit none
      
      real(8), intent(out) :: roll_lambda
      
      ! Return the rolloff wavelength
      
      roll_lambda = rolo_ell * two_pi
      
      return
      end subroutine rolloff_get_roll_lambda




c     rolloff_get_seis_lambda - Get the seismogenic wavelength.
c     Parameters:
c      seis_lambda = Variable to receive seismogeneic wavelength, in meters.

      subroutine rolloff_get_seis_lambda (seis_lambda)
      
      use math_constants
      use rolloff_data
      implicit none
      
      real(8), intent(out) :: seis_lambda
      
      ! Return the seismogenic wavelength
      
      seis_lambda = rolo_zeta * two_pi
      
      return
      end subroutine rolloff_get_seis_lambda




c     rolloff_get_corr_lambda - Get the correlation wavelength.
c     Parameters:
c      corr_lambda = Variable to receive correlation wavelength, in meters.

      subroutine rolloff_get_corr_lambda (corr_lambda)
      
      use rolloff_data
      implicit none
      
      real(8), intent(out) :: corr_lambda
      
      ! Return the correlation wavelength
      
      corr_lambda = rolo_corr_lambda
      
      return
      end subroutine rolloff_get_corr_lambda




c     rolloff_get_tau_ms_cont - Get the continuum mean-square shear stress.
c     Parameters:
c      tau_ms_cont = Variable to receive continuum mean-square shear stress, in Pascal.

      subroutine rolloff_get_tau_ms_cont (tau_ms_cont)
      
      use rolloff_data
      implicit none
      
      real(8), intent(out) :: tau_ms_cont
      
      ! Return the continuum mean-square shear stress
      
      tau_ms_cont = rolo_tau_ms_cont
      
      return
      end subroutine rolloff_get_tau_ms_cont




c     rolloff_write_param - Write parameters.
c     Parameters:
c      iunit = Fortran unit number.

      subroutine rolloff_write_param (iunit)
      
      use math_constants
      use rolloff_data
      implicit none
      
      integer, intent(in) :: iunit
      
      ! Write the parameters
      
      write(iunit,'(A)') '***** Parameters for rolloff function'
      write(iunit,*)
      
      write(iunit,'(A,I9)') 'rolo_form = ',rolo_form
      write(iunit,'(A,1pe15.6)') 'roll_lambda = ',rolo_ell * two_pi
      write(iunit,'(A,1pe15.6)') 'seis_lambda = ',rolo_zeta * two_pi
      write(iunit,'(A,1pe15.6)') 'integ_ksq = ',rolo_integ_ksq
      write(iunit,'(A,1pe15.6)') 'corr_lambda = ',rolo_corr_lambda
      write(iunit,'(A,1pe15.6)') 'tau_ms_cont = ',rolo_tau_ms_cont
      
      write(iunit,*)
      
      return
      end subroutine rolloff_write_param
      
      

      
c     rolloff_integ_ksq_range - Calculate integral of rolloff function.
c     Parameters:
c      integ = Variable to receive the integral.
c      k_lo = Lower limit of integration range.
c      k_hi = Upper limit of integration range.
c     Calculate the integral from k_lo to k_hi of (k^2)*g(k) dk.
c     This integral is evaluated using Simpson's rule.
c     This subroutine assumes g is isotropic.  For a non-isotropic g,
c      the subroutine should calculate the 2D integral of
c      k*g(k0,k1)/(2*pi) over the annulus k_lo <= k <= h_hi.

      subroutine rolloff_integ_ksq_range (integ, k_lo, k_hi)
      
      implicit none
      
      real(8), intent(out) :: integ
      real(8), intent(in) :: k_lo
      real(8), intent(in) :: k_hi
      
      real(8) :: prev_integ  ! Previous integral
      real(8) :: next_integ  ! Next integral
      real(8) :: s1  ! Sum of function values with coefficient 1
      real(8) :: s2  ! Sum of function values with coefficient 2
      real(8) :: s4  ! Sum of function values with coefficient 4
      integer :: n  ! Number of subintervals of [k_lo,k_hi]
      
      integer :: i  ! Current subinterval number
      real(8) :: k  ! Wavenumber
      real(8) :: g  ! Calculated function value
      
      real(8), parameter :: tol = 1.0d-8  ! tolerance for convergence
      
      ! Initialize for adaptive integration
      
      prev_integ = 0.0d0
      next_integ = 0.0d0
      s2 = 0.0d0
      s4 = 0.0d0
        
      call rolloff_calc_g (g, 0.0d0, k_lo)
      s1 = g*k_lo*k_lo
      call rolloff_calc_g (g, 0.0d0, k_hi)
      s1 = s1 + g*k_hi*k_hi
      
      ! Loop until maximum number of subintervals are considered
      
      n = 1
      
      do while (n < 1048576)  ! 1048576 = 2^20
        
        ! Double number of subintervals, and transfer new points
        
        n = n + n
        s2 = s2 + s4
        s4 = 0.0d0
        
        ! Loop over new points, and accumulate them
        
        do i = 1, n-1, 2
        
          k = ((k_hi - k_lo)*dble(i)/dble(n)) + k_lo
          call rolloff_calc_g (g, 0.0d0, k)
          
          s4 = g*k*k + s4
        
        enddo
        
        ! Save previous integral, and calculate the new one
        
        prev_integ = next_integ
        next_integ = (s1 + 2.0d0*s2 + 4.0d0*s4)*(k_hi-k_lo)/dble(n*3)
        
        ! Check for convergence
        
        if (n > 256 .and.
     &      abs(next_integ - prev_integ) < abs(tol*next_integ)) then
          exit
        endif
      
      enddo  ! end loop on subinterval
      
      ! Return the integral
      
      integ = next_integ
     
      return
      end subroutine rolloff_integ_ksq_range




c     rolloff_integ_logpow_k - Calculate g*(k^2) as a function of k.
c     This is only for g = 1/(1 + (ell*k)^3 * (log(1 + zeta*k))^m)
c     Parameters:
c      gksq = Variable to receive the value of the g function.
c      k = Wavenumber.
c      m = Logarithm power.

      subroutine rolloff_integ_logpow_k (gksq, k, m)
      
      use rolloff_data
      implicit none
      
      real(8), intent(out) :: gksq
      real(8), intent(in) :: k
      real(8), intent(in) :: m
      
      real(8) :: ell_k  ! ell times k
      real(8) :: log_term  ! logarithmic term
      
      ! Log-power formula
      
      ell_k = rolo_ell * k
      log_term = log(1.0d0 + (rolo_zeta * k))
      gksq = (k*k) / (1.0d0 + (ell_k*ell_k*ell_k*(log_term**m)))
      
      return
      end subroutine rolloff_integ_logpow_k




c     rolloff_integ_logpow_t - Calculate g*(k^2)*(dk/dt)
c                              as a function of t.
c     This is only for g = 1/(1 + (ell*k)^3 * (log(1 + zeta*k))^m)
c     Parameters:
c      gksq = Variable to receive the value of the g function.
c      t = Transformed wavenumber.
c      m = Logarithm power.
c      a = Cutoff value of k that corresponds to t=1.

      subroutine rolloff_integ_logpow_t (gksq, t, m, a)
      
      use math_constants
      use rolloff_data
      implicit none
      
      real(8), intent(out) :: gksq
      real(8), intent(in) :: t
      real(8), intent(in) :: m
      real(8), intent(in) :: a
      
      real(8) :: mm1  ! m minus 1
      real(8) :: zeta_a  ! zeta times a, over e
      real(8) :: ell_a  ! ell times a, over e
      real(8) :: t_pow  ! t**(1/(m-1))
      
      real(8) :: exp_term  ! exponential term
      real(8) :: log_term  ! logarithmic term
      
      ! Floating-point epsilon
      
      real(8), parameter :: fp_eps = 1.0d-16
      
      ! A number such that exp(-fp_exp_range) is approximately the
      ! smallest floating-point number
      
      real(8), parameter :: fp_exp_range = 690.0d0
      
      ! A number such that exp(fp_exp_eps) is approximately the
      ! reciprocal of the floating-point epsilon
      ! Note:  We assume fp_exp_eps <= fp_exp_range/3
      
      real(8), parameter :: fp_exp_eps = 37.0d0
      
      ! Calculate limit value, also the overall constant
      
      mm1 = m - 1.0d0
      zeta_a = rolo_zeta*a/euler_e
      ell_a = rolo_ell*a/euler_e
      
      ! If t is extremely small ...
      
      if (t <= fp_eps) then
      
        ! Exponential term -> 0 and logarithmic term -> 1
        
        exp_term = 0.0d0
        log_term = 1.0d0
        
      ! Otherwise, if t is small enough that the exponential term is
      ! effectively zero (implies log term can be approximated) ...
      
      else if (t <= (3.0d0/fp_exp_range)**mm1) then
      
        t_pow = t**(1.0d0/mm1)
        exp_term = 0.0d0
        log_term = log(zeta_a)*t_pow + 1.0d0
        
      ! Otherwise, if t is small enough so the logarithmic term can
      ! be approximated ...
      
      else if (t <= (1.0d0/fp_exp_eps)**mm1) then
      
        t_pow = t**(1.0d0/mm1)
        exp_term = exp(-3.0d0/t_pow) * (t*t_pow)
        log_term = log(zeta_a)*t_pow + 1.0d0
        
      ! Otherwise, evaluate formula directly
      
      else
      
        t_pow = t**(1.0d0/mm1)
        exp_term = exp(-3.0d0/t_pow) * (t*t_pow)
        log_term = log(zeta_a*exp(1.0d0/t_pow) + 1.0d0) * t_pow
      
      endif
      
      ! Compute the final result
      
      gksq = 1.0d0 / (rolo_ell*rolo_ell*rolo_ell*mm1*(
     &                exp_term/(ell_a*ell_a*ell_a) + log_term**m ))

      return
      end subroutine rolloff_integ_logpow_t
      
      

      
c     rolloff_integ_logpow - Calculate integral of rolloff function.
c     This is only for g = 1/(1 + (ell*k)^3 * (log(1 + zeta*k))^m)
c     Parameters:
c      m = Logarithm power.
c     The integral is stored in rolo_integ_ksq
c     Calculate the integral from 0 to infinity of (k^2)*g(k) dk.
c     Integrations are performed using Simpson's rule.  The function
c      (k^2)*g(k) is integrated from 0 to a.  Then the following
c      function is integrated from 0 to 1, and two integrals are added:
c           1/( (m-1)*(ell^3)*(t^(m/(m-1)))*
c               ( ((e/(a*ell))^3) * exp(-3/(t^(1/(m-1))))
c                 + (log( 1 + (zeta*a/e)*exp(1/(t^(1/(m-1)))) ))^m ) )
c      The function comes from the substitution:
c           k = (a/e)*exp(1/(t^(1/(m-1))))

      subroutine rolloff_integ_logpow (m)
      
      use rolloff_data
      implicit none
      
      real(8), intent(in) :: m
      
      real(8) :: a  ! Cutoff k value between the two integrals
      real(8) :: total_integ  ! Total integral so far
      
      real(8) :: prev_integ  ! Previous integral
      real(8) :: next_integ  ! Next integral
      real(8) :: s1  ! Sum of function values with coefficient 1
      real(8) :: s2  ! Sum of function values with coefficient 2
      real(8) :: s4  ! Sum of function values with coefficient 4
      integer :: n  ! Number of subintervals of [0,1]
      
      integer :: i  ! Current subinterval number
      real(8) :: k  ! Wavenumber
      real(8) :: t  ! Transformed wavenumber
      real(8) :: gksq  ! Calculated function value
      
      real(8), parameter :: tol = 1.0d-8  ! tolerance for convergence
      
      ! Initialize for total integral
      
      total_integ = 0.0d0
      
      ! Get the cutoff value
      ! First term puts cutoff about 8x the peak, 5x inflection point
      ! Second term puts cutoff about 2x where log becomes significant
      
      a = max(10.0d0/rolo_ell, 5.0d0/rolo_zeta)
      
      !=== Main part of function, integrated wrt k from 0 to a
      
      ! Initialize for adaptive integration
      
      prev_integ = 0.0d0
      next_integ = 0.0d0
      s2 = 0.0d0
      s4 = 0.0d0
      
      k = 0.0d0
      call rolloff_integ_logpow_k (gksq, k, m)
      s1 = gksq
      
      k = a
      call rolloff_integ_logpow_k (gksq, k, m)
      s1 = s1 + gksq
      
      ! Loop until maximum number of subintervals are considered
      
      n = 1
      
      do while (n < 1048576)  ! 1048576 = 2^20
        
        ! Double number of subintervals, and transfer new points
        
        n = n + n
        s2 = s2 + s4
        s4 = 0.0d0
        
        ! Loop over new points, and accumulate them
        
        do i = 1, n-1, 2
        
          k = a*dble(i)/dble(n)
          call rolloff_integ_logpow_k (gksq, k, m)
          
          s4 = s4 + gksq
        
        enddo
        
        ! Save previous integral, and calculate the new one
        
        prev_integ = next_integ
        next_integ = (s1 + 2.0d0*s2 + 4.0d0*s4)*a/dble(n*3)
        
        ! Check for convergence
        
        if (n > 256 .and.
     &      abs(next_integ - prev_integ) < abs(tol*next_integ)) then
          exit
        endif
      
      enddo  ! end loop on subinterval
        
      ! Add integral to total integral

      total_integ = total_integ + next_integ
      
      !=== Tail part of function, integrated wrt t from 0 to 1
      
      ! Initialize for adaptive integration
      
      prev_integ = 0.0d0
      next_integ = 0.0d0
      s2 = 0.0d0
      s4 = 0.0d0
      
      t = 0.0d0
      call rolloff_integ_logpow_t (gksq, t, m, a)
      s1 = gksq
      
      t = 1.0d0
      call rolloff_integ_logpow_t (gksq, t, m, a)
      s1 = s1 + gksq
      
      ! Loop until maximum number of subintervals are considered
      
      n = 1
      
      do while (n < 1048576)  ! 1048576 = 2^20
        
        ! Double number of subintervals, and transfer new points
        
        n = n + n
        s2 = s2 + s4
        s4 = 0.0d0
        
        ! Loop over new points, and accumulate them
        
        do i = 1, n-1, 2
        
          t = dble(i)/dble(n)
          call rolloff_integ_logpow_t (gksq, t, m, a)
          
          s4 = s4 + gksq
        
        enddo
        
        ! Save previous integral, and calculate the new one
        
        prev_integ = next_integ
        next_integ = (s1 + 2.0d0*s2 + 4.0d0*s4)/dble(n*3)
        
        ! Check for convergence
        
        if (n > 256 .and.
     &      abs(next_integ - prev_integ) < abs(tol*next_integ)) then
          exit
        endif
      
      enddo  ! end loop on subinterval
        
      ! Add integral to total integral

      total_integ = total_integ + next_integ
      
      ! Return the integral
      
      rolo_integ_ksq = total_integ
     
      return
      end subroutine rolloff_integ_logpow




c     rolloff_set_logsq - Set log-square rolloff formula.
c     Parameters:
c      corr_lambda = Correlation wavelength, in meters.
c      seis_b = Seismogenic wavelength constant, in meters.
c      seis_m = Seismogenic wavelength slope.
c      roll_lambda = Variable to receive rolloff wavelength, in meters.
c      seis_lambda = Variable to receive seismogeneic wavelength, in meters.
c     The parameters ell and zeta are calculated so that (k^2)*g(k)
c      reaches its maximum value at k = 2*pi/corr_lambda.  This means
c      the peak contribution to the stress perturbation occurs at the
c      correlation wavelength.  (Perhaps the correlation wavelength is
c      twice the desired perturbation size.)
c     The parameters are adjusted so that:
c        seis_lambda = seis_m*roll_lambda + seis_b
c      This allows the seismogenic wavelength to be fixed, or to vary as a
c      linear function of the rolloff wavelength.  (Perhaps the seismogenic
c      wavelength is twice the size of the largest earthquake or depth of
c      the seismogenic zone.)
c     The rolloff wavelength is the wavelength at which the power-law
c      decay begins, and the seismogenic wavelength is the wavelength at
c      which the logarithmic decay begins.

      subroutine rolloff_set_logsq (
     & corr_lambda, seis_b, seis_m, roll_lambda, seis_lambda)
      
      use math_constants
      use rolloff_data
      implicit none
      
      real(8), intent(in) :: corr_lambda
      real(8), intent(in) :: seis_b
      real(8), intent(in) :: seis_m
      real(8), intent(out) :: roll_lambda
      real(8), intent(out) :: seis_lambda
      
      real(8) :: ell_hi  ! hi value of ell
      real(8) :: ell_lo  ! lo value of ell
      real(8) :: k  ! wavenumber of peak
      logical :: f_search_hi  ! true if searching for high value
      real(8) :: ell_k  ! ell times k
      real(8) :: zeta_k  ! zeta times k
      real(8) :: log_term  ! logarithmic term
      real(8) :: deriv_num  ! negative numerator of derivative
      
      real(8), parameter :: ell_tol = 1.0d-12  ! relative tolerance
      
      ! Remember the selected formula
      
      rolo_form = rolo_form_logsq
      
      ! Remember the correlation wavelength
      
      rolo_corr_lambda = corr_lambda
      
      ! Convert correlation wavelength to wavenumber
      
      k = two_pi / corr_lambda
      
      ! Initialize hi and lo so first value is 1/k
      
      ell_lo = 0.0d0
      ell_hi = 2.0d0 / k
      
      ! Searching for hi value
      
      f_search_hi = .true.
      
      ! Loop until tolerance is met
      
      do while (ell_hi - ell_lo >= ell_hi * ell_tol)
        
        ! New value to check is midpoint of interval
        
        rolo_ell = 0.5d0 * (ell_hi + ell_lo)
        
        ! Calculate zeta value as linearly dependent on ell
        
        rolo_zeta = (seis_b / two_pi) + (seis_m * rolo_ell)
        
        ! Calculate the negative numerator of the derivative
        ! of (k^2) * g(k), which is monotonic increasing
        
        ell_k = rolo_ell * k
        zeta_k = rolo_zeta * k
        log_term = log (1.0d0 + zeta_k)
        deriv_num = ell_k*ell_k*ell_k*log_term*(
     &    0.5d0*log_term + zeta_k/(1.0d0 + zeta_k)) - 1.0d0
        
        ! If the negative derivative is positive ...
        
        if (deriv_num >= 0.0d0) then
          
          ! Current value is the new hi limit
          
          ell_hi = rolo_ell
          
          ! Remember we have seen a high value
          
          f_search_hi = .false.
        
        ! Otherwise, the negative derivative is negative ...
        
        else
          
          ! Current value is the new lo limit
          
          ell_lo = rolo_ell
          
          ! If searching for high value, double the hi limit
          
          if (f_search_hi) then
            ell_hi = ell_hi + ell_hi
          endif
        
        endif
      
      enddo  ! end loop until tolerance is met
      
      ! Return the resulting wavelengths
      
      roll_lambda = rolo_ell * two_pi
      seis_lambda = rolo_zeta * two_pi
      
      ! Compute the integral of (k^2)*g(k)
      
      call rolloff_integ_logpow (2.0d0)
      
      return
      end subroutine rolloff_set_logsq




c     rolloff_set_none - Set none rolloff formula.

      subroutine rolloff_set_none ()
      
      use rolloff_data
      implicit none
      
      ! Remember the selected formula
      
      rolo_form = rolo_form_none
      
      ! Fill unused values
      
      rolo_ell = 0.0d0
      rolo_zeta = 0.0d0
      rolo_integ_ksq = 0.0d0
      rolo_corr_lambda = 0.0d0
      
      return
      end subroutine rolloff_set_none




c     rolloff_read_param - Read parameters for rolloff formula.
c     Parameters:
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format for none rolloff formula:
c       Formula = rolo_form_none = 1
c       Continuum mean-square shear stress, in Pascal
c
c     File format for log-square rolloff formula:
c       Formula = rolo_form_logsq = 2
c       Correlation wavelength, in meters
c       Seismogenic wavelength constant, in meters
c       Seismogenic wavelength slope
c       Continuum mean-square shear stress, in Pascal

      subroutine rolloff_read_param (iunit, f_ok)
      
      use rolloff_data
      implicit none
      
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: form
      
      real(8) :: corr_lambda
      real(8) :: seis_b
      real(8) :: seis_m
      real(8) :: roll_lambda
      real(8) :: seis_lambda
      
      ! Read formula
      
      read(iunit,*) form
      
      ! Read selected formula
      
      select case (form)
      
      ! No rolloff
      
      case (rolo_form_none)
        call rolloff_set_none ()
      
      ! Log-square formula
      
      case (rolo_form_logsq)
        read(iunit,*) corr_lambda
        read(iunit,*) seis_b
        read(iunit,*) seis_m
        
        call rolloff_set_logsq (
     &    corr_lambda, seis_b, seis_m, roll_lambda, seis_lambda)
        
      ! Invalid formula
      
      case default
        f_ok = .false.
        return
      
      end select
      
      ! Common parameters
      
      read(iunit,*) rolo_tau_ms_cont
      f_ok = .true.
      
      return
      end subroutine rolloff_read_param








c----------------------------------------------------------------------
c
c     trexpd - Truncated exponential distribution.




c     trexpd_calc_H - Calculate the H factor.
c     Parameters:
c      H = Variable to received computed H factor.
c      g = Mean of truncated exponential distribution.
c     The H factor is defined to satisfy:
c           E_max = - (1/beta) * log(1-H)
c      where E_max is the maximum value of the truncated distribution,
c      and where beta is the decay exponent.  Note that 1/beta is the
c      mean of the untruncated distribution.
c     The g value is defined to be the mean of the truncated
c      distribution, expressed as a fraction of the mean of the
c      untruncated distribution.  So g ranges from 0 to 1, and the mean
c      of the truncated distribution is g/beta.
c     The formula relating g and H is:
c           1 + (1-H)log(1-H)/H = g
c      This formula is well-behaved at H=1, where g=1, and is also
c      well-behaved at H=0, where g ~ H/2.
c     The purpose of this subroutine is to solve the above equation
c      for H, given the value of g.  Define
c           f(H) = 1 + (1-H)log(1-H)/H
c           f'(H) = - (log(1-H)/H + 1)/H
c     The idea is to use Newton's method.  The main problem is that
c      f'(H) becomes infinite at H=1.  Note however that f'(1-C^2) is
c      about twice as large as f'(1-C) for C small.  So if we can
c      bracket H between 1-C and 1-C^2, then Newton's method ought to
c      work OK with starting value 1-C^2.
c     Another problem is that calculation of log(1-H) is inaccurate
c      when H is small.  We handle this by using a power series for H
c      when g is small.  The power series is also used to give starting
c      values for Newton's method in the region where f'(H) <= 1.  Note
c      that f''(H) is always positive, hence each Newton's iterate is
c      larger than the true root, so if we know g is not small then
c      Newton's method will never produce a small H.

      subroutine trexpd_calc_H (H, g)
      
      implicit none
      
      real(8), intent(out) :: H
      real(8), intent(in) :: g
      
      ! Power series expansion of H in powers of g.
      
      real(8), parameter ::
     &  Hg01 =  2.0000000000000000000000000000000d0  ! 2
      real(8), parameter ::
     &  Hg02 = -1.3333333333333333333333333333333d0  ! -4/3
      real(8), parameter ::
     &  Hg03 =  4.4444444444444444444444444444444d-1  ! 4/9
      real(8), parameter ::
     &  Hg04 = -1.1851851851851851851851851851852d-1  ! -16/135
      real(8), parameter ::
     &  Hg05 =  1.9753086419753086419753086419753d-2  ! 8/405
      real(8), parameter ::
     &  Hg06 = -5.6437389770723104056437389770723d-3  ! -16/2835
      
      ! Value of H at which f'(H) = 1, a few decimal places is OK.
      
      real(8), parameter :: H_slope_1 = 0.683803d0
      
      ! Value of g for H_slope_1, a few decimal places is OK.
      
      real(8), parameter :: g_slope_1 = 0.467586d0
      
      ! Relative tolerance for result, note this must be larger than
      ! the double-precision epsilon (typically 1e-16).  Note that
      ! 32 bits of precision is enough for random number generation.
      
      real(8), parameter :: tol = 1.0d-10
      
      ! Value of g small enough to be evaluated by power series alone.
      ! This should be about (tol)^(1/6) for degree-6 power series.
      
      real(8), parameter :: g_power = 0.0215d0
      
      ! Temporaries
      
      real(8) :: log_over_H  ! logarithm over H
      real(8) :: f  ! function value
      real(8) :: f_prime  ! derivative value
      real(8) :: delta_H  ! change in value of H
      
      real(8) :: S  ! S = 1-H
      
      integer :: state_S  ! State for finding S
      integer, parameter :: state_S_search = 1  ! Search for start value
      integer, parameter :: state_S_newton = 2  ! Need Newton's method
      integer, parameter :: state_S_zero = 3  ! S forced to zero
      
      !--- If g lies in region of slope <= 1 ...
      
      if (g <= g_slope_1) then
        
        ! Evaluate power series
        
        H = (((((Hg06*g+Hg05)*g+Hg04)*g+Hg03)*g+Hg02)*g+Hg01)*g
        
        ! If need to refine value with Newton's method
        
        if (g > g_power) then
          
          ! Loop until tolerance met
          
          do
            
            ! Evaluate function and derivative
          
            log_over_H = log(1.0d0 - H)/H
            f = 1.0d0 + (1.0d0 - H)*log_over_H
            f_prime = -(log_over_H + 1.0d0)/H
          
            ! Newton update
            
            delta_H = - (f - g)/f_prime
            H = H + delta_H
            
            ! Check tolerance
            
            if (abs(delta_H) <= abs(H * tol)) then
              exit
            endif
          
          enddo  ! end Newton's method
        
        endif
      
      !--- Otherwise, g lies in region of slope > 1 ...
      
      else
      
        ! Work in terms of S = 1-H
        
        S = 1 - H_slope_1
        
        ! Loop to search for starting value
        
        state_S = state_S_search
        
        do while (state_S == state_S_search)
        
          ! Square S, which roughly doubles the slope
          
          S = S*S
          
          ! Don't go smaller than the tolerance
          
          if (S <= tol) then
            S = tol
            state_S = state_S_zero
          endif
          
          ! If function value exceeds g, we found starting value
          
          if (1.0d0 + S*log(S)/(1.0d0 - S) >= g) then
            state_S = state_S_newton
          endif
        
        enddo  ! end loop to search for starting value
        
        ! If need to force S to zero ...
        
        if (state_S == state_S_zero) then
          
          ! Force S=0 which will give H=1, for g close to 1
          
          S = 0
          
        ! Otherwise, perform Newton's method ...
          
        else
        
          ! Loop until tolerance met
          
          do
            
            ! Evaluate function and derivative
          
            log_over_H = log(S)/(1.0d0 - S)
            f = 1.0d0 + S*log_over_H
            f_prime = -(log_over_H + 1.0d0)/(1.0d0 - S)
          
            ! Newton update
            
            delta_H = - (f - g)/f_prime
            S = S - delta_H
            
            ! Check tolerance
            
            if (abs(delta_H) <= abs((1.0d0 - S) * tol)) then
              exit
            endif
          
          enddo  ! end Newton's method
        
        endif
        
        ! Set value of H
        
        H = 1.0d0 - S
      
      endif
      
      return
      end subroutine trexpd_calc_H




c     trexpd_gen - Generate truncated exponential random number.
c     Parameters:
c      E = Variable to receive random number.
c      U_mean = Mean value of untruncated exponential distribution.
c      E_max = Maximum value of truncated exponential distribution.
c      g = Mean of truncated exponential distribution.
c     The U_mean is equal to 1/beta, where beta is the decay exponent.
c     The E_max value is defined to be the maximum value of the
c      truncated distribution, expressed as a fraction of the mean of
c      the untruncated distribution.  So the returned E satisfies
c      E <= E_max * U_mean.
c     The g value is defined to be the mean of the truncated
c      distribution, expressed as a fraction of the mean of the
c      untruncated distribution.  So g ranges from 0 to 1, and the mean
c      of the truncated distribution is g/beta == g * U_mean.
c     The function uses the lower of the two truncation limits implied
c      by E_max and g.  It also imposes a limit based on the assumed
c      32-bit resolution of the random number generator.  If the E_max
c      limit is not desired, set E_max = 100.0.  If the g limit is not
c      desired, set g = 1.0.
c     The returned value is
c           E = - U_mean * log(1 - H*U)
c      where U is uniformly distributed in [0,1).  The value H is as
c      described in trexpd_calc_H, except that we also impose
c           H <= 1 - exp(-E_max)
c      and
c           H <= 1 - 2^(-32) == 1 - exp(-32 * log(2))
c      The latter constraint reflects the 32-bit resolution of U, and
c      serves to prevent taking the log of zero.

      subroutine trexpd_gen (E, U_mean, E_max, g)
      implicit none
      
      real(8), intent(out) :: E
      real(8), intent(in) :: U_mean
      real(8), intent(in) :: E_max
      real(8), intent(in) :: g
      
      ! Random number function
      
      interface
        function genrand_real2()
          real(8) :: genrand_real2
        end function
      end interface
      
      ! Uniformly distributed random number in range [0,1)
      
      real(8) :: U
      
      ! Calculated H factor (see trexpd_calc_H)
      
      real(8) :: H
      
      ! Product H*U
      
      real(8) :: X
      
      ! Limit reflecting 32-bit resolution of random numbers
      
      real(8), parameter :: E_lim = 22.18d0  ! 32 * log(2)
      
      ! Coefficient of series expansion for -log(1-X)
      
      real(8), parameter ::
     &  L1 =  1.0000000000000000000000000000000d+0  ! 1/1
      real(8), parameter ::
     &  L2 =  5.0000000000000000000000000000000d-1  ! 1/2
      real(8), parameter ::
     &  L3 =  3.3333333333333333333333333333333d-1  ! 1/3
      real(8), parameter ::
     &  L4 =  2.5000000000000000000000000000000d-1  ! 1/4
      real(8), parameter ::
     &  L5 =  2.0000000000000000000000000000000d-1  ! 1/5
      real(8), parameter ::
     &  L6 =  1.6666666666666666666666666666667d-1  ! 1/6
      real(8), parameter ::
     &  L7 =  1.4285714285714285714285714285714d-1  ! 1/7
      real(8), parameter ::
     &  L8 =  1.2500000000000000000000000000000d-1  ! 1/8
      
      ! Calculate the H factor
      
      call trexpd_calc_H (H, g)
      
      ! Limit H to obtain requested maximum
      
      H = min(H, 1.0d0 - exp(- min(E_max, E_lim)))
      
      ! Get random number in the range [0,1)
      
      U = genrand_real2()
      
      ! Product
      
      X = H * U
      
      ! If X is not too small ...
      
      if (X >= 0.01d0) then
      
        ! Evaluate directly using log
      
        E = - U_mean * log(1 - X)
        
      ! Otherwise X is small ...
      
      else
      
        ! Evaluate using series expansion
        
        E = U_mean*(((((((L8*X+L7)*X+L6)*X+L5)*X+L4)*X+L3)*X+L2)*X+L1)*X
      
      endif
      
      return
      end subroutine trexpd_gen




c     trexpd_phase - Generate random phase.
c     Parameters:
c      C = Variable to receive cos(phase).
c      S = Variable to receive sin(phase).
c     Note:  If E is a truncated exponential random variable, then
c      C*sqrt(E) and S*sqrt(E) are a 2D truncated Gaussian random
c      variable.  In the special case where E is untruncated, they are
c      independent 1D Gaussian random variables.

      subroutine trexpd_phase (C, S)
      
      use math_constants
      implicit none
      
      real(8), intent(out) :: C
      real(8), intent(out) :: S
      
      ! Random number function
      
      interface
        function genrand_real2()
          real(8) :: genrand_real2
        end function
      end interface
      
      ! Uniformly distributed random number in range [0,1)
      
      real(8) :: U
      
      ! Phase
      
      real(8) :: phase
      
      ! Get random number in the range [0,1)
      
      U = genrand_real2()
      
      ! Calculate the phase
      
      phase = two_pi * U
      
      ! Return the cosine and sine
      
      C = cos(phase)
      S = sin(phase)
      
      return
      end subroutine trexpd_phase








c----------------------------------------------------------------------
c
c     ft - Fourier transform operations.




c     ft_data - Data module.
c
c     Fourier transform data is stored in a 3-dimensional array as:
c       z(0:1, 0:N1-1, 0:N0-1)
c     This means the storage order is row-by-row on the fault.
c     Direction 1 is along-strike, and direction 0 is along-dip.
c     In the spatial domain, the real part is the 1-component, and
c      the imaginary part is the 0-component of stress.
c     The frequency domain has integer coordinates y0, y1 with ranges
c           -N0/2 < y0 <= N0/2  and  -N1/2 < y1 <= N1/2
c     Wavenumbers k0, k1 are defined as
c           k0 = 2*pi*y0/L0  and  k1 = 2*pi*y1/L1
c     The Nyquist wavenumbers are thus pi*N0/L0 and pi*N1/L1.
c     We do not place any energy into the Nyquist modes, because we
c      want (+/-y0, +/-y1) to be four distinct modes if y0, y1 are
c      both nonzero, to ensure full symmetry.
c     The spatial domain has integer coordinates m0, m1 with ranges
c           0 <= m0 < N0  and  0 <= m1 < N1
c     Physical coordinates x0, x1 are defined as
c           x0 = m0*L0/N0  and   x1 = m1*L1/N1

      module ft_data
      implicit none
      
      ! Number of points in directions 0, 1
      
      integer, save :: ft_N0
      integer, save :: ft_N1
      
      ! Size of the domain in meters in directions 0, 1
      
      real(8), save :: ft_L0
      real(8), save :: ft_L1
      
      ! Stress rotation option
      
      integer, save :: ft_strrot
      
      ! Stress rotation = make slip horizontal
      
      integer, parameter :: ft_strrot_hor_slip = 1
      
      ! Stress rotation = make stress horizontal
      
      integer, parameter :: ft_strrot_hor_stress = 2
      
      ! Phase shift option
      
      integer, save :: ft_phshft
      
      ! Phase shift = none
      
      integer, parameter :: ft_phshft_none = 1
      
      ! Phase shift = real (sign change only)
      
      integer, parameter :: ft_phshft_real = 2
      
      ! Phase shift = complex (global complex phase shift)
      
      integer, parameter :: ft_phshft_cplx = 3
      
      end module ft_data




c     ft_cyclic_set - Set value, with cyclically-interpreted indexes.
c     Parameters:
c      z = Array for Fourier domain.
c      m1 = Index in 1-direction.
c      m0 = Index in 0-direction.
c      re_val = Real value to set.
c      im_val = Imaginary value to set.
c     The indexes are adjusted, cyclically, to lie in array index range.

      subroutine ft_cyclic_set (z, m1, m0, re_val, im_val)
      
      use math_constants
      use ft_data
      implicit none
      
      real(8), intent(inout) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      integer, intent(in) :: m1
      integer, intent(in) :: m0
      real(8), intent(in) :: re_val
      real(8), intent(in) :: im_val
      
      integer :: i1
      integer :: i0
      
      ! Adjust indexes to lie within range
      
      i1 = m1
      do while (i1 >= ft_N1)
        i1 = i1 - ft_N1
      enddo
      do while (i1 < 0)
        i1 = i1 + ft_N1
      enddo
      
      i0 = m0
      do while (i0 >= ft_N0)
        i0 = i0 - ft_N0
      enddo
      do while (i0 < 0)
        i0 = i0 + ft_N0
      enddo
      
      ! Set the value
      
      z(i_re, i1, i0) = re_val
      z(i_im, i1, i0) = im_val
      
      return
      end subroutine ft_cyclic_set




c     ft_cyclic_get - Get value, with cyclically-interpreted indexes.
c     Parameters:
c      z = Array for Fourier domain.
c      m1 = Index in 1-direction.
c      m0 = Index in 0-direction.
c      re_val = Variable to receive real value.
c      im_val = Variable to receive imaginary value.
c     The indexes are adjusted, cyclically, to lie in array index range.

      subroutine ft_cyclic_get (z, m1, m0, re_val, im_val)
      
      use math_constants
      use ft_data
      implicit none
      
      real(8), intent(in) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      integer, intent(in) :: m1
      integer, intent(in) :: m0
      real(8), intent(out) :: re_val
      real(8), intent(out) :: im_val
      
      integer :: i1
      integer :: i0
      
      ! Adjust indexes to lie within range
      
      i1 = m1
      do while (i1 >= ft_N1)
        i1 = i1 - ft_N1
      enddo
      do while (i1 < 0)
        i1 = i1 + ft_N1
      enddo
      
      i0 = m0
      do while (i0 >= ft_N0)
        i0 = i0 - ft_N0
      enddo
      do while (i0 < 0)
        i0 = i0 + ft_N0
      enddo
      
      ! Get the value
      
      re_val = z(i_re, i1, i0)
      im_val = z(i_im, i1, i0)
      
      return
      end subroutine ft_cyclic_get




c     ft_forward_fft - Perform forward fast Fourier transform.
c     Parameters:
c      z = Array for Fourier domain.
c     This function transforms from spatial to frequency domain.

      subroutine ft_forward_fft (z)
      
      use ft_data
      implicit none
      
      real(8), intent(inout) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      
      ! Array to hold extents of each dimension
      
      integer :: ndim
      integer :: nn(2)
      
      ! Sign flag, +1 = forward FFT, -1 = reverse FFT
      
      integer :: isign
      
      ! Set up extents
      
      ndim = 2
      nn(1) = ft_N1
      nn(2) = ft_N0
      
      ! Set up sign flag
      
      isign = 1
      
      ! Perform the Fourier transform
      
      call fft_ndim (z, nn, ndim, isign)
      
      return
      end subroutine ft_forward_fft




c     ft_reverse_fft - Perform reverse fast Fourier transform.
c     Parameters:
c      z = Array for Fourier domain.
c     This function transforms from frequency to spatial domain.

      subroutine ft_reverse_fft (z)
      
      use ft_data
      implicit none
      
      real(8), intent(inout) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      
      ! Array to hold extents of each dimension
      
      integer :: ndim
      integer :: nn(2)
      
      ! Sign flag, +1 = forward FFT, -1 = reverse FFT
      
      integer :: isign
      
      ! Set up extents
      
      ndim = 2
      nn(1) = ft_N1
      nn(2) = ft_N0
      
      ! Set up sign flag
      
      isign = -1
      
      ! Perform the Fourier transform
      
      call fft_ndim (z, nn, ndim, isign)
      
      return
      end subroutine ft_reverse_fft




c     ft_normalize_fft - Multiply by Fourier normalization factor.
c     Parameters:
c      z = Array for Fourier domain.
c     This function multiplies each element of the array by 1/(N0*N1).
c     Normalization makes forward followed by reverse FFT (or
c      vice-versa) yield the original value.
c     Normalization can be done in either spatial or frequency domain.

      subroutine ft_normalize_fft (z)
      
      use math_constants
      use ft_data
      implicit none
      
      real(8), intent(inout) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      
      ! Fourier domain coordinates
      
      integer :: m0
      integer :: m1
      
      ! Normalization factor for FFT
      
      real(8) :: fft_norm
      
      ! Apply normalization factor
      
      fft_norm = 1.0d0 / dble(ft_N0 * ft_N1)
      
      do m0 = 0, ft_N0-1
        do m1 = 0, ft_N1-1
          z(i_re, m1, m0) = z(i_re, m1, m0) * fft_norm
          z(i_im, m1, m0) = z(i_im, m1, m0) * fft_norm
        enddo
      enddo
      
      return
      end subroutine ft_normalize_fft




c     ft_fill_zero - Fill Fourier domain with zero.
c     Parameters:
c      z = Array for Fourier domain.
c     This function sets each element of the array to zero.
c     Zero-fill can be done in either spatial or frequency domain.

      subroutine ft_fill_zero (z)
      
      use math_constants
      use ft_data
      implicit none
      
      real(8), intent(out) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      
      ! Fourier domain coordinates
      
      integer :: m0
      integer :: m1
      
      ! Fill with zero
      
      do m0 = 0, ft_N0-1
        do m1 = 0, ft_N1-1
          z(i_re, m1, m0) = 0.0d0
          z(i_im, m1, m0) = 0.0d0
        enddo
      enddo
      
      return
      end subroutine ft_fill_zero




c     ft_copy - Copy Fourier domain.
c     Parameters:
c      dest = Destination array for Fourier domain.
c      src = Source array for Fourier domain.

      subroutine ft_copy (dest, src)
      
      use math_constants
      use ft_data
      implicit none
      
      real(8), intent(out) :: dest(0:1, 0:ft_N1-1, 0:ft_N0-1)
      real(8), intent(in) :: src(0:1, 0:ft_N1-1, 0:ft_N0-1)
      
      ! Fourier domain coordinates
      
      integer :: m0
      integer :: m1
      
      ! Copy
      
      do m0 = 0, ft_N0-1
        do m1 = 0, ft_N1-1
          dest(i_re, m1, m0) = src(i_re, m1, m0)
          dest(i_im, m1, m0) = src(i_im, m1, m0)
        enddo
      enddo
      
      return
      end subroutine ft_copy




c     ft_save_xreal - Save transformed real-valued function.
c     Parameters:
c      z = Array for Fourier domain.
c      h = Half-size array to save Fourier domain.
c     This function assumes z contains the Fourier transform of a real-
c      valued function, and hence that values at negative coordinates
c      are complex conjugates.  The Fourier domain is saved into an
c      array with half+1 as many rows.
c     Values at negative coordinates are averaged to yield the saved
c      value.

      subroutine ft_save_xreal (z, h)
      
      use math_constants
      use ft_data
      implicit none
      
      real(8), intent(in) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      real(8), intent(out) :: h(0:1, 0:ft_N1-1, 0:ft_N0/2)
      
      ! Fourier domain coordinates
      
      integer :: m0
      integer :: m1
      
      ! Negative Fourier domain coordinates
      
      integer :: neg_m0
      integer :: neg_m1
      
      ! Save values
      
      do m0 = 0, ft_N0/2
      
        if (m0 == 0) then
          neg_m0 = 0
        else
          neg_m0 = ft_N0 - m0
        endif
      
        do m1 = 0, ft_N1-1
      
          if (m1 == 0) then
            neg_m1 = 0
          else
            neg_m1 = ft_N1 - m1
          endif
        
          h(i_re, m1, m0)
     &      = 0.5d0 * (z(i_re, m1, m0) + z(i_re, neg_m1, neg_m0))
          h(i_im, m1, m0)
     &      = 0.5d0 * (z(i_im, m1, m0) - z(i_im, neg_m1, neg_m0))
          
        enddo
      enddo
      
      return
      end subroutine ft_save_xreal




c     ft_restore_xreal - Restore transformed real-valued function.
c     Parameters:
c      z = Array for Fourier domain.
c      h = Half-size array to save Fourier domain.
c     This function assumes z contains the Fourier transform of a real-
c      valued function, and hence that values at negative coordinates
c      are complex conjugates.  The Fourier domain is saved into an
c      array with half+1 as many rows.
c     Values at negative coordinates are averaged to yield the saved
c      value.

      subroutine ft_restore_xreal (z, h)
      
      use math_constants
      use ft_data
      implicit none
      
      real(8), intent(out) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      real(8), intent(in) :: h(0:1, 0:ft_N1-1, 0:ft_N0/2)
      
      ! Fourier domain coordinates
      
      integer :: m0
      integer :: m1
      
      ! Negative Fourier domain coordinates
      
      integer :: neg_m0
      integer :: neg_m1
      
      ! Fill with zero
      
      do m0 = 0, ft_N0-1
      
        if (m0 == 0) then
          neg_m0 = 0
        else
          neg_m0 = ft_N0 - m0
        endif
      
        do m1 = 0, ft_N1-1
      
          if (m1 == 0) then
            neg_m1 = 0
          else
            neg_m1 = ft_N1 - m1
          endif
          
          if (m0 <= ft_N0/2) then
        
            z(i_re, m1, m0) = h(i_re, m1, m0)
            z(i_im, m1, m0) = h(i_im, m1, m0)
            
          else
        
            z(i_re, m1, m0) = h(i_re, neg_m1, neg_m0)
            z(i_im, m1, m0) = - h(i_im, neg_m1, neg_m0)
            
          endif
          
        enddo
      enddo
      
      return
      end subroutine ft_restore_xreal




c     ft_gen_modes - Generate modes with random amplitude and phase.
c     Parameters:
c      z = Array for Fourier frequency domain.
c      tau_ms_mean = Variable to receive mean-square shear stress,
c                    mean value expected due to g-function and max k.
c      tau_ms_actual = Variable to receive mean-square shear stress,
c                      actual value.
c     We put energy only into modes that satisfy
c           (y0/(N0/2))^2 + (y1/(N1/2))^2 < 1
c      Note that this excludes the Nyquist modes.  If the node spacing
c      is the same in the two directions, this is a circle in the
c      k0,k1 plane.  The condition may be written in integer form as
c           lim0sq*(y0^2) + lim1sq*(y1^2) < limsq
c           limsq = max(N0/2, N1/2)^2
c           lim0sq = limsq/((N0/2)^2)
c           lim1sq = limsq/((N1/2)^2)
c     We also don't put energy into the DC mode.

      subroutine ft_gen_modes (z, tau_ms_mean, tau_ms_actual)
      
      use math_constants
      use ft_data
      implicit none
      
      real(8), intent(out) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      real(8), intent(out) :: tau_ms_mean
      real(8), intent(out) :: tau_ms_actual
      
      ! Wavenumbers and their squares
      
      real(8) :: k0
      real(8) :: k0sq
      real(8) :: k1
      real(8) :: k1sq
      real(8) :: k
      real(8) :: ksq
      
      ! Frequency domain coordinates
      
      integer :: y0
      integer :: y1
      
      ! Negative coordinates, shifted into range
      
      integer :: neg_y0
      integer :: neg_y1
      
      ! Absolute coordinates, in range 0 to Nyquist
      
      integer :: abs_y0
      integer :: abs_y1
      
      ! Low limit for y1
      
      integer :: y1_lo
      
      ! Nyquist coordinates
      
      integer :: half_N0
      integer :: half_N1
      
      ! Limits for k cutoff
      
      integer :: limsq
      integer :: lim0sq
      integer :: lim1sq
      
      ! Integral of (k^2)*g(k)
      
      real(8) :: integ_g_ksq
      
      ! Mean-square shear stress in the continuum
      
      real(8) :: tau_ms_cont
      
      ! Accumulators for shear stress
      
      real(8) :: tau_ms_mean_acc
      real(8) :: tau_ms_actual_acc
      
      ! Energy rolloff g function
      
      real(8) :: g
      
      ! Unconstrained expected total mean-square shear stress of pair
      
      real(8) :: U_mean
      
      ! Maximum energy of mode pair, as fraction of unconstrained energy
      ! (also number of standard deviations of max allowed energy)
      
      real(8), parameter :: E_max = 5.5d0  ! chop at cdf = 99.6%
      
      ! Randomly-generated total mean-square shear stress for pair
      
      real(8) :: tau_ms_pair
      
      ! Cosine and sine of randomly-generated phase
      
      real(8) :: C
      real(8) :: S
      
      ! Azimuthal terms for energy, and stresses in 0 and 1 directions
      
      real(8) :: azimuth_E
      real(8) :: azimuth_0
      real(8) :: azimuth_1
      
      ! Amplitudes of along-dip and along-strike stresses
      
      real(8) :: G0
      real(8) :: G1
      
      ! Calculate Nyquist coordinates
      
      half_N0 = ft_N0/2
      half_N1 = ft_N1/2
      
      ! Calculate limits for k cutoff
      
      limsq = max(half_N0, half_N1) 
      lim0sq = limsq / half_N0
      lim1sq = limsq / half_N1
      
      limsq = limsq * limsq
      lim0sq = lim0sq * lim0sq
      lim1sq = lim1sq * lim1sq
      
      ! Get integral of (k^2)*g(k)
      
      call rolloff_get_integ_ksq (integ_g_ksq)
      
      ! Get the continuum mean-square shear stress
      
      call rolloff_get_tau_ms_cont (tau_ms_cont)
      
      ! Zero out the Nyquist modes and the DC mode
      
      y0 = 0
      y1 = 0
      z(i_re, y1, y0) = 0.0d0
      z(i_im, y1, y0) = 0.0d0
      
      y0 = half_N0
      do y1 = 0, ft_N1-1
        z(i_re, y1, y0) = 0.0d0
        z(i_im, y1, y0) = 0.0d0
      enddo
      
      y1 = half_N1
      do y0 = 0, ft_N0-1
        z(i_re, y1, y0) = 0.0d0
        z(i_im, y1, y0) = 0.0d0
      enddo
      
      ! Initialize stress accumulation
      
      tau_ms_mean = 0.0d0
      tau_ms_actual = 0.0d0
      
      ! Loop over pairs of modes
      ! We need to loop over all pairs (y0,y1) and (-y0,-y1) except for
      ! the DC component and pairs that contain a Nyquist frequency
      
      do y0 = 0, ft_N0-1
      
        ! If Nyquist ...
      
        if (y0 == half_N0) then
        
          ! Skip Nyquist frequency
        
          cycle
        
        ! Otherwise, if below Nyquist ...
      
        else if (y0 < half_N0) then
        
          ! Start at 1, so we don't include the DC component, and don't
          ! double-count pairs of the form (y0,0) and (-y0,0)
        
          y1_lo = 1
        
        ! Otherwise, above Nyquist
      
        else
        
          ! Start at 0, to pick up pairs of the form (y0,0) and (-y0,0)
        
          y1_lo = 0
        
        endif
      
        ! Calculate the negative index
      
        if (y0 == 0) then
          neg_y0 = 0
        else
          neg_y0 = ft_N0 - y0
        endif
      
        ! Calculate the wavenumber and its square, and absolute index
      
        if (y0 <= half_N0) then
          k0 = dble(y0)*two_pi/ft_L0
          abs_y0 = y0
        else
          k0 = dble(y0 - ft_N0)*two_pi/ft_L0
          abs_y0 = ft_N0 - y0
        endif
        k0sq = k0*k0
      
        ! Initialize stress accumulation
      
        tau_ms_mean_acc = 0.0d0
        tau_ms_actual_acc = 0.0d0
      
        ! Inner loop
        
        do y1 = y1_lo, half_N1 - 1
      
          ! Calculate the negative index
      
          if (y1 == 0) then
            neg_y1 = 0
          else
            neg_y1 = ft_N1 - y1
          endif
      
          ! Calculate the wavenumber and its square, and absolute index
      
          if (y1 <= half_N1) then
            k1 = dble(y1)*two_pi/ft_L1
            abs_y1 = y1
          else
            k1 = dble(y1 - ft_N1)*two_pi/ft_L1
            abs_y1 = ft_N1 - y1
          endif
          k1sq = k1*k1
          
          ksq = k0sq + k1sq
          k = sqrt(ksq)
          
          ! If not within our defined k ellipse ...
          
          if (lim0sq*abs_y0*abs_y0 + lim1sq*abs_y1*abs_y1 >= limsq) then
            
            ! Zero stress for this pair of modes
            
            z(i_re,     y1,     y0) = 0.0d0
            z(i_im,     y1,     y0) = 0.0d0
          
            z(i_re, neg_y1, neg_y0) = 0.0d0
            z(i_im, neg_y1, neg_y0) = 0.0d0
            
            ! Go to next mode
            
            cycle
            
          ! Otherwise, within our k ellipse ...
            
          else
          
            ! Calculate energy rolloff for these wavenumbers
          
            call rolloff_calc_g (g, k0, k1)
          
            ! Unconstrained expected total mean-square shear stress
          
            U_mean = four_pi*k*tau_ms_cont/(ft_L0*ft_L1*integ_g_ksq)
          
            ! Generate random total mean-square shear stress for pair
          
            call trexpd_gen (tau_ms_pair, U_mean, E_max * g, g)
          
            ! Generate cosine and sine of random phase
          
            call trexpd_phase (C, S)
          
            ! Accumulate stresses
          
            tau_ms_mean_acc = tau_ms_mean_acc + (U_mean * g)
            tau_ms_actual_acc = tau_ms_actual_acc + tau_ms_pair
          
            ! Calculate stress amplitudes
      
            select case (ft_strrot)
            
            ! Stress amplitudes for horizontal slip
            
            case (ft_strrot_hor_slip)
          
              azimuth_E = (3.0d0*k1sq + ksq)/ksq
              azimuth_1 = (k1sq + ksq)/ksq
              azimuth_0 = (k0*k1)/ksq
          
              G1 = - sqrt(2.0d0*tau_ms_pair/azimuth_E) * azimuth_1
              G0 = - sqrt(2.0d0*tau_ms_pair/azimuth_E) * azimuth_0
            
            ! Stress amplitudes for horizontal stress
            
            case (ft_strrot_hor_stress)
          
              G1 = - sqrt(2.0d0*tau_ms_pair)
              G0 = 0.0d0
            
            end select
          
            ! Put modes into Fourier amplitudes, G1 real, G0 imaginary
            ! In complex numbers this operation is:
            !  z(y0, y1) = 0.5*(G1 + i*G0)*exp(i*phase)
            !  z(-y0, -y1) = 0.5*(G1 + i*G0)*exp(-i*phase)
        
            z(i_re,     y1,     y0) = 0.5d0*(  G1*C - G0*S)
            z(i_im,     y1,     y0) = 0.5d0*(  G1*S + G0*C)
          
            z(i_re, neg_y1, neg_y0) = 0.5d0*(  G1*C + G0*S)
            z(i_im, neg_y1, neg_y0) = 0.5d0*(- G1*S + G0*C)
          
          endif
        
        enddo  ! end inner loop
      
        ! Accumulate stresses from this row
      
        tau_ms_mean = tau_ms_mean + tau_ms_mean_acc
        tau_ms_actual = tau_ms_actual + tau_ms_actual_acc

      enddo  ! end loop over pairs of modes
      
      return
      end subroutine ft_gen_modes




c     ft_gen_sel_modes - Generate modes with random amplitude and phase,
c                        and select origin and phase to match template.
c     Parameters:
c      z = Array for Fourier frequency domain.
c      tau_ms_mean = Variable to receive mean-square shear stress,
c                    mean value expected due to g-function and max k.
c      tau_ms_actual = Variable to receive mean-square shear stress,
c                      actual value.
c      v = Array containing Fourier transform of selection template
c          (the contents of this array is destroyed).
c      org0 = Variable to receive template match location, coord 0.
c      org1 = Variable to receive template match location, coord 1.
c      sel_conv = Variable to receive selection template convolution.
c     We put energy only into modes that satisfy
c           (y0/(N0/2))^2 + (y1/(N1/2))^2 < 1
c      Note that this excludes the Nyquist modes.  If the node spacing
c      is the same in the two directions, this is a circle in the
c      k0,k1 plane.  The condition may be written in integer form as
c           lim0sq*(y0^2) + lim1sq*(y1^2) < limsq
c           limsq = max(N0/2, N1/2)^2
c           lim0sq = limsq/((N0/2)^2)
c           lim1sq = limsq/((N1/2)^2)
c     We also don't put energy into the DC mode.
c     The selection template is a real-valued table v(m0,m1).  (The v
c      array passed in to this function is the Fourier transform of v.)
c      We select an origin point (org0,org1) and a global phase theta
c      to maximize the real part of the convolution
c           sel_conv = Re SUM(m0,m1)(v(m0,m1)*z(org0+m0,org1+m1))
c      it being understood that v and z in the above equation refer to
c      their values in the Fourier spatial domain.  Note the phase
c      theta is chosen to make sel_conv positive.  The value of sel_conv
c      measures goodness-of-fit of the selection template.

      subroutine ft_gen_sel_modes (z, tau_ms_mean, tau_ms_actual,
     &                             v, org0, org1, sel_conv)
      
      use math_constants
      use ft_data
      implicit none
      
      real(8), intent(out) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      real(8), intent(out) :: tau_ms_mean
      real(8), intent(out) :: tau_ms_actual
      real(8), intent(out) :: v(0:1, 0:ft_N1-1, 0:ft_N0-1)
      integer, intent(out) :: org0
      integer, intent(out) :: org1
      real(8), intent(out) :: sel_conv
      
      ! Wavenumbers and their squares
      
      real(8) :: k0
      real(8) :: k0sq
      real(8) :: k1
      real(8) :: k1sq
      real(8) :: k
      real(8) :: ksq
      
      ! Frequency domain coordinates
      
      integer :: y0
      integer :: y1
      
      ! Negative coordinates, shifted into range
      
      integer :: neg_y0
      integer :: neg_y1
      
      ! Absolute coordinates, in range 0 to Nyquist
      
      integer :: abs_y0
      integer :: abs_y1
      
      ! Low limit for y1
      
      integer :: y1_lo
      
      ! Nyquist coordinates
      
      integer :: half_N0
      integer :: half_N1
      
      ! Limits for k cutoff
      
      integer :: limsq
      integer :: lim0sq
      integer :: lim1sq
      
      ! Integral of (k^2)*g(k)
      
      real(8) :: integ_g_ksq
      
      ! Mean-square shear stress in the continuum
      
      real(8) :: tau_ms_cont
      
      ! Accumulators for shear stress
      
      real(8) :: tau_ms_mean_acc
      real(8) :: tau_ms_actual_acc
      
      ! Energy rolloff g function
      
      real(8) :: g
      
      ! Unconstrained expected total mean-square shear stress of pair
      
      real(8) :: U_mean
      
      ! Maximum energy of mode pair, as fraction of unconstrained energy
      ! (also number of standard deviations of max allowed energy)
      
      real(8), parameter :: E_max = 5.5d0  ! chop at cdf = 99.6%
      
      ! Randomly-generated total mean-square shear stress for pair
      
      real(8) :: tau_ms_pair
      
      ! Cosine and sine of randomly-generated phase
      
      real(8) :: C
      real(8) :: S
      
      ! Azimuthal terms for energy, and stresses in 0 and 1 directions
      
      real(8) :: azimuth_E
      real(8) :: azimuth_0
      real(8) :: azimuth_1
      
      ! Amplitudes of along-dip and along-strike stresses
      
      real(8) :: G0
      real(8) :: G1
      
      ! Temporary for complex number
      
      real(8) :: re_tmp
      real(8) :: im_tmp
      
      ! New convolution value
      
      real(8) :: new_sel_conv
      
      ! Calculate Nyquist coordinates
      
      half_N0 = ft_N0/2
      half_N1 = ft_N1/2
      
      ! Calculate limits for k cutoff
      
      limsq = max(half_N0, half_N1) 
      lim0sq = limsq / half_N0
      lim1sq = limsq / half_N1
      
      limsq = limsq * limsq
      lim0sq = lim0sq * lim0sq
      lim1sq = lim1sq * lim1sq
      
      ! Get integral of (k^2)*g(k)
      
      call rolloff_get_integ_ksq (integ_g_ksq)
      
      ! Get the continuum mean-square shear stress
      
      call rolloff_get_tau_ms_cont (tau_ms_cont)
      
      ! Zero out the Nyquist modes and the DC mode
      
      y0 = 0
      y1 = 0
      z(i_re, y1, y0) = 0.0d0
      z(i_im, y1, y0) = 0.0d0
      v(i_re, y1, y0) = 0.0d0
      v(i_im, y1, y0) = 0.0d0
      
      y0 = half_N0
      do y1 = 0, ft_N1-1
        z(i_re, y1, y0) = 0.0d0
        z(i_im, y1, y0) = 0.0d0
        v(i_re, y1, y0) = 0.0d0
        v(i_im, y1, y0) = 0.0d0
      enddo
      
      y1 = half_N1
      do y0 = 0, ft_N0-1
        z(i_re, y1, y0) = 0.0d0
        z(i_im, y1, y0) = 0.0d0
        v(i_re, y1, y0) = 0.0d0
        v(i_im, y1, y0) = 0.0d0
      enddo
      
      !--- Stochastic stress generation ---
      
      ! Initialize stress accumulation
      
      tau_ms_mean = 0.0d0
      tau_ms_actual = 0.0d0
      
      ! Loop over pairs of modes
      ! We need to loop over all pairs (y0,y1) and (-y0,-y1) except for
      ! the DC component and pairs that contain a Nyquist frequency
      
      do y0 = 0, ft_N0-1
      
        ! If Nyquist ...
      
        if (y0 == half_N0) then
        
          ! Skip Nyquist frequency
        
          cycle
        
        ! Otherwise, if below Nyquist ...
      
        else if (y0 < half_N0) then
        
          ! Start at 1, so we don't include the DC component, and don't
          ! double-count pairs of the form (y0,0) and (-y0,0)
        
          y1_lo = 1
        
        ! Otherwise, above Nyquist
      
        else
        
          ! Start at 0, to pick up pairs of the form (y0,0) and (-y0,0)
        
          y1_lo = 0
        
        endif
      
        ! Calculate the negative index
      
        if (y0 == 0) then
          neg_y0 = 0
        else
          neg_y0 = ft_N0 - y0
        endif
      
        ! Calculate the wavenumber and its square, and absolute index
      
        if (y0 <= half_N0) then
          k0 = dble(y0)*two_pi/ft_L0
          abs_y0 = y0
        else
          k0 = dble(y0 - ft_N0)*two_pi/ft_L0
          abs_y0 = ft_N0 - y0
        endif
        k0sq = k0*k0
      
        ! Initialize stress accumulation
      
        tau_ms_mean_acc = 0.0d0
        tau_ms_actual_acc = 0.0d0
      
        ! Inner loop
        
        do y1 = y1_lo, half_N1 - 1
      
          ! Calculate the negative index
      
          if (y1 == 0) then
            neg_y1 = 0
          else
            neg_y1 = ft_N1 - y1
          endif
      
          ! Calculate the wavenumber and its square, and absolute index
      
          if (y1 <= half_N1) then
            k1 = dble(y1)*two_pi/ft_L1
            abs_y1 = y1
          else
            k1 = dble(y1 - ft_N1)*two_pi/ft_L1
            abs_y1 = ft_N1 - y1
          endif
          k1sq = k1*k1
          
          ksq = k0sq + k1sq
          k = sqrt(ksq)
          
          ! If not within our defined k ellipse ...
          
          if (lim0sq*abs_y0*abs_y0 + lim1sq*abs_y1*abs_y1 >= limsq) then
            
            ! Zero stress for this pair of modes
            
            z(i_re,     y1,     y0) = 0.0d0
            z(i_im,     y1,     y0) = 0.0d0
          
            z(i_re, neg_y1, neg_y0) = 0.0d0
            z(i_im, neg_y1, neg_y0) = 0.0d0
            
            v(i_re,     y1,     y0) = 0.0d0
            v(i_im,     y1,     y0) = 0.0d0
          
            v(i_re, neg_y1, neg_y0) = 0.0d0
            v(i_im, neg_y1, neg_y0) = 0.0d0
            
            ! Go to next mode
            
            cycle
            
          ! Otherwise, within our k ellipse ...
            
          else
          
            ! Calculate energy rolloff for these wavenumbers
          
            call rolloff_calc_g (g, k0, k1)
          
            ! Unconstrained expected total mean-square shear stress
          
            U_mean = four_pi*k*tau_ms_cont/(ft_L0*ft_L1*integ_g_ksq)
          
            ! Generate random total mean-square shear stress for pair
          
            call trexpd_gen (tau_ms_pair, U_mean, E_max * g, g)
          
            ! Generate cosine and sine of random phase
          
            call trexpd_phase (C, S)
          
            ! Accumulate stresses
          
            tau_ms_mean_acc = tau_ms_mean_acc + (U_mean * g)
            tau_ms_actual_acc = tau_ms_actual_acc + tau_ms_pair
          
            ! Calculate stress amplitudes
      
            select case (ft_strrot)
            
            ! Stress amplitudes for horizontal slip
            
            case (ft_strrot_hor_slip)
          
              azimuth_E = (3.0d0*k1sq + ksq)/ksq
              azimuth_1 = (k1sq + ksq)/ksq
              azimuth_0 = (k0*k1)/ksq
          
              G1 = - sqrt(2.0d0*tau_ms_pair/azimuth_E) * azimuth_1
              G0 = - sqrt(2.0d0*tau_ms_pair/azimuth_E) * azimuth_0
            
            ! Stress amplitudes for horizontal stress
            
            case (ft_strrot_hor_stress)
          
              G1 = - sqrt(2.0d0*tau_ms_pair)
              G0 = 0.0d0
            
            end select
          
            ! Put modes into Fourier amplitudes, G1 real, G0 imaginary
            ! In complex numbers this operation is:
            !  z(y0, y1) = 0.5*(G1 + i*G0)*exp(i*phase)
            !  z(-y0, -y1) = 0.5*(G1 + i*G0)*exp(-i*phase)
        
            z(i_re,     y1,     y0) = 0.5d0*(  G1*C - G0*S)
            z(i_im,     y1,     y0) = 0.5d0*(  G1*S + G0*C)
          
            z(i_re, neg_y1, neg_y0) = 0.5d0*(  G1*C + G0*S)
            z(i_im, neg_y1, neg_y0) = 0.5d0*(- G1*S + G0*C)
            
            ! Multiply selection template by complex transform of G1
            ! In complex numbers this operation is:
            !  v(y0, y1) = 0
            !  v(-y0, -y1) = conj(v(-y0, -y1)) * G1*exp(-i*phase)
        
            v(i_re,     y1,     y0) = 0.0d0
            v(i_im,     y1,     y0) = 0.0d0
            
            re_tmp = v(i_re, neg_y1, neg_y0)
            im_tmp = v(i_im, neg_y1, neg_y0)
          
            v(i_re, neg_y1, neg_y0) =   re_tmp*G1*C - im_tmp*G1*S
            v(i_im, neg_y1, neg_y0) = - re_tmp*G1*S - im_tmp*G1*C
          
          endif
        
        enddo  ! end inner loop
      
        ! Accumulate stresses from this row
      
        tau_ms_mean = tau_ms_mean + tau_ms_mean_acc
        tau_ms_actual = tau_ms_actual + tau_ms_actual_acc

      enddo  ! end loop over pairs of modes
      
      !--- Convolution to find optimum location and phase ---
      
      ! Reverse FFT to convolve template with along-strike stress
      ! (We should not call ft_normalize_fft here because our stresses
      ! already include the normalization factor.  The stresses are
      ! scaled to produce correct results with just ft_reverse_fft.)
      
      call ft_reverse_fft (v)
      !!call ft_normalize_fft (v)
      
      ! Search for maximum magnitude of the convolution
      
      y0 = 0
      y1 = 0
      
      org0 = y0
      org1 = y1
          
      re_tmp = v(i_re, y1, y0)
      im_tmp = v(i_im, y1, y0)
      
      select case (ft_phshft)
        case (ft_phshft_none)
          sel_conv = re_tmp
        case (ft_phshft_real)
          sel_conv = abs(re_tmp)
        case (ft_phshft_cplx)
          sel_conv = sqrt(re_tmp*re_tmp + im_tmp*im_tmp)
      end select
      
      do y0 = 0, ft_N0-1
        do y1 = 0, ft_N1-1
        
          ! Get magnitude of this element of convolution
          
          re_tmp = v(i_re, y1, y0)
          im_tmp = v(i_im, y1, y0)
      
          select case (ft_phshft)
            case (ft_phshft_none)
              new_sel_conv = re_tmp
            case (ft_phshft_real)
              new_sel_conv = abs(re_tmp)
            case (ft_phshft_cplx)
              new_sel_conv = sqrt(re_tmp*re_tmp + im_tmp*im_tmp)
          end select
          
          ! If new maximum, save location
          
          if (new_sel_conv > sel_conv) then
          
            org0 = y0
            org1 = y1
            
            sel_conv = new_sel_conv
            
          endif
        
        enddo
      enddo
      
      ! For the location of the maximum, get the magnitude and phase
      
      y0 = org0
      y1 = org1
          
      re_tmp = v(i_re, y1, y0)
      im_tmp = v(i_im, y1, y0)
      
      select case (ft_phshft)
        case (ft_phshft_none)
          sel_conv = re_tmp
          im_tmp = 0.0d0
        case (ft_phshft_real)
          sel_conv = abs(re_tmp)
          im_tmp = 0.0d0
        case (ft_phshft_cplx)
          sel_conv = sqrt(re_tmp*re_tmp + im_tmp*im_tmp)
      end select
      
      C = re_tmp / sel_conv  ! cos(theta)
      S = im_tmp / sel_conv  ! sin(theta)
      
      !--- Adjust stresses for optimum phase ---
      
      ! Loop over pairs of modes
      ! We need to loop over all pairs (y0,y1) and (-y0,-y1) except for
      ! the DC component and pairs that contain a Nyquist frequency
      
      do y0 = 0, ft_N0-1
      
        ! If Nyquist ...
      
        if (y0 == half_N0) then
        
          ! Skip Nyquist frequency
        
          cycle
        
        ! Otherwise, if below Nyquist ...
      
        else if (y0 < half_N0) then
        
          ! Start at 1, so we don't include the DC component, and don't
          ! double-count pairs of the form (y0,0) and (-y0,0)
        
          y1_lo = 1
        
        ! Otherwise, above Nyquist
      
        else
        
          ! Start at 0, to pick up pairs of the form (y0,0) and (-y0,0)
        
          y1_lo = 0
        
        endif
      
        ! Calculate the negative index
      
        if (y0 == 0) then
          neg_y0 = 0
        else
          neg_y0 = ft_N0 - y0
        endif
      
        ! Inner loop
        
        do y1 = y1_lo, half_N1 - 1
      
          ! Calculate the negative index
      
          if (y1 == 0) then
            neg_y1 = 0
          else
            neg_y1 = ft_N1 - y1
          endif
          
          ! Apply optimum phase to Fourier amplitudes
          ! In complex numbers this operation is:
          !  z(y0, y1) = z(y0, y1)*exp(i*theta)
          !  z(-y0, -y1) = z(-y0, -y1)*exp(-i*theta)
        
          re_tmp = z(i_re,     y1,     y0)
          im_tmp = z(i_im,     y1,     y0)
        
          z(i_re,     y1,     y0) =   re_tmp*C - im_tmp*S
          z(i_im,     y1,     y0) =   re_tmp*S + im_tmp*C
          
          re_tmp = z(i_re, neg_y1, neg_y0)
          im_tmp = z(i_im, neg_y1, neg_y0)
          
          z(i_re, neg_y1, neg_y0) =   re_tmp*C + im_tmp*S
          z(i_im, neg_y1, neg_y0) = - re_tmp*S + im_tmp*C

        enddo  ! end inner loop

      enddo  ! end loop over pairs of modes
      
      return
      end subroutine ft_gen_sel_modes








c----------------------------------------------------------------------
c
c     fault - Fault plane operations.




c     fault_data - Data module.
c
c     Fault data is stored in a 3-dimensional array as:
c       w(0:1, 0:cell_count1, 0:cell_count0)
c     This means the storage order is row-by-row on the fault.
c     Direction 1 is along-strike, and direction 0 is along-dip
c      (assuming a strike-slip fault).
c     In the fault domain, the real part is the 1-component, and
c      the imaginary part is the 0-component of stress.
c     Fault nodes have integer coordinates node0, node1 with ranges
c           0 <= node0 < cell_count0  and  0 <= node1 < cell_count1
c     Physical coordinates x0, x1 are defined as
c           x0 = node0*cell_size0  and   x1 = node1*cell_size1

      module fault_data
      implicit none
      
      ! Number of cells in directions 0, 1
      
      integer, save :: cell_count0
      integer, save :: cell_count1
      
      ! Size of a cell in meters in directions 0, 1
      
      real(8), save :: cell_size0
      real(8), save :: cell_size1
      
      ! Number of fft points per cell in directions 0, 1
      
      integer, save :: cell_stride0
      integer, save :: cell_stride1
      
      ! Radius of smoothing kernel, in fft points, in directions 0, 1
      ! 1 = no smoothing, cell_stride = FEM consistent stress
      
      integer, save :: smooth_stride0
      integer, save :: smooth_stride1
      
      ! Smoothing kernel
      ! Dimension:  smooth_kernel(-(smooth_stride1-1):smooth_stride1-1,
      !                           -(smooth_stride0-1):smooth_stride0-1)
      
      real(8), allocatable, save :: smooth_kernel(:,:)
      
      ! Scale factors to apply to real and imaginary parts
      
      real(8), save :: smooth_scale_re
      real(8), save :: smooth_scale_im
      
      ! Fault type: true for strike-slip, false for dip_slip
      
      logical, save :: f_strike_slip
      
      end module fault_data




c     fault_make_smooth_kernel - Create the smoothing kernel.
c     Assumes that smooth_stride0, smooth_stride1 have been set.

      subroutine fault_make_smooth_kernel ()
      
      use fault_data
      implicit none
      
      integer :: si0
      integer :: si1
      real(8) :: weight
      
      ! Construct the smoothing kernel, bilinear
      
      if (allocated(smooth_kernel)) then
        deallocate (smooth_kernel)
      endif
      
      allocate (smooth_kernel(-(smooth_stride1-1):smooth_stride1-1,
     &                        -(smooth_stride0-1):smooth_stride0-1))
      
      weight = 0.0d0
      
      do si0 = -(smooth_stride0-1), smooth_stride0-1
        do si1 = -(smooth_stride1-1), smooth_stride1-1
          smooth_kernel(si1, si0) = 
     &        dble(smooth_stride1 - abs(si1))
     &      * dble(smooth_stride0 - abs(si0))
     
          weight = weight + smooth_kernel(si1, si0)
        enddo
      enddo
      
      do si0 = -(smooth_stride0-1), smooth_stride0-1
        do si1 = -(smooth_stride1-1), smooth_stride1-1
          smooth_kernel(si1, si0) = 
     &      smooth_kernel(si1, si0) / weight
        enddo
      enddo
      
      return
      end subroutine fault_make_smooth_kernel




c     fault_ft_read_param - Read parameters for fault and fft.
c     Parameters:
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       cell_count0, cell_count 1
c       cell_size0, cell_size1
c       cell_stride0, cell_stride1
c       smooth_stride0, smooth_stride1
c       f_strike_slip (0 = dip slip, 1 = strike slip)
c       min_N0, min_N1

      subroutine fault_ft_read_param (iunit, f_ok)
      
      use ft_data
      use fault_data
      implicit none
      
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: form
      
      integer :: iflag
      integer :: i
      integer :: min_N0
      integer :: min_N1
      
      ! Cell count
      
      read(iunit,*) cell_count0, cell_count1
      
      if (cell_count0 < 1 .or. cell_count1 < 1) then
        f_ok = .false.
        return
      endif
      
      ! Cell size
      
      read(iunit,*) cell_size0, cell_size1
      
      if (cell_size0 <= 0.0d0 .or. cell_size1 <= 0.0d0) then
        f_ok = .false.
        return
      endif
      
      ! Cell stride
      
      read(iunit,*) cell_stride0, cell_stride1
      
      if (cell_stride0 < 1 .or. cell_stride1 < 1) then
        f_ok = .false.
        return
      endif
      
      ! Smoothing kernel
      
      read(iunit,*) smooth_stride0, smooth_stride1
      
      if (smooth_stride0 < 1 .or. smooth_stride1 < 1) then
        f_ok = .false.
        return
      endif
      
      call fault_make_smooth_kernel ()
      
      ! Smoothing scale
      
      read(iunit,*) smooth_scale_re, smooth_scale_im
      
      ! Fault type flag
      
      read(iunit,*) iflag
      select case (iflag)
        case (0)
          f_strike_slip = .false.
        case (1)
          f_strike_slip = .true.
        case default
          f_ok = .false.
          return
      end select
      
      ! FFT domain number of points
      
      read(iunit,*) min_N0, min_N1
      
      if (min_N0 < 0 .or. min_N1 < 0) then
        f_ok = .false.
        return
      endif
      
      min_N0 = max(min_N0, 2*cell_count0*cell_stride0)
      min_N1 = max(min_N1, 2*cell_count1*cell_stride1)
      
      ft_N0 = 2
      do while (ft_N0 < min_N0)
        ft_N0 = ft_N0 * 2
      enddo
      
      ft_N1 = 2
      do while (ft_N1 < min_N1)
        ft_N1 = ft_N1 * 2
      enddo
      
      ! FFT domain size
      
      ft_L0 = dble(ft_N0) * cell_size0 / dble(cell_stride0)
      ft_L1 = dble(ft_N1) * cell_size1 / dble(cell_stride1)
      
      ! FFT stress rotation option
      
      read(iunit,*) ft_strrot
      
      select case (ft_strrot)
        case (ft_strrot_hor_slip)
        case (ft_strrot_hor_stress)
        case default
          f_ok = .false.
          return
      end select
      
      ! FFT phase shift option
      
      read(iunit,*) ft_phshft
      
      select case (ft_phshft)
        case (ft_phshft_none)
        case (ft_phshft_real)
        case (ft_phshft_cplx)
        case default
          f_ok = .false.
          return
      end select
     
      ! Success
        
      f_ok = .true.
      
      return
      end subroutine fault_ft_read_param




c     fault_write_table_re - Write table containing real values.
c     Parameters:
c      iunit = Fortran unit number.
c      w = Array containing fault data.
c     Data is written in row-major order.
c     The ordering takes into account strike-slip vs dip-slip.

      subroutine fault_write_table_re (iunit, w)
      
      use math_constants
      use fault_data
      implicit none
      
      integer, intent(in) :: iunit
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Do index
      
      integer :: i
      
      ! If strike-slip ...
      
      if (f_strike_slip) then
      
        do node0 = 0, cell_count0
          do node1 = 0, cell_count1, 8
            write(iunit,'(8(1x,1pe15.6))')
     &       (w(i_re, i, node0),
     &          i = node1, min(node1+7,cell_count1))
          enddo
        enddo
      
      ! Otherwise, dip-slip ...
      
      else
      
        do node1 = 0, cell_count1
          do node0 = 0, cell_count0, 8
            write(iunit,'(8(1x,1pe15.6))')
     &       (w(i_re, node1, i),
     &          i = node0, min(node0+7,cell_count0))
          enddo
        enddo
      
      endif
     
      return
      end subroutine fault_write_table_re




c     fault_write_table_im - Write table containing imaginary values.
c     Parameters:
c      iunit = Fortran unit number.
c      w = Array containing fault data.
c     Data is written in row-major order.
c     The ordering takes into account strike-slip vs dip-slip.

      subroutine fault_write_table_im (iunit, w)
      
      use math_constants
      use fault_data
      implicit none
      
      integer, intent(in) :: iunit
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Do index
      
      integer :: i
      
      ! If strike-slip ...
      
      if (f_strike_slip) then
      
        do node0 = 0, cell_count0
          do node1 = 0, cell_count1, 8
            write(iunit,'(8(1x,1pe15.6))')
     &       (w(i_im, i, node0),
     &          i = node1, min(node1+7,cell_count1))
          enddo
        enddo
      
      ! Otherwise, dip-slip ...
      
      else
      
        do node1 = 0, cell_count1
          do node0 = 0, cell_count0, 8
            write(iunit,'(8(1x,1pe15.6))')
     &       (w(i_im, node1, i),
     &          i = node0, min(node0+7,cell_count0))
          enddo
        enddo
      
      endif
     
      return
      end subroutine fault_write_table_im




c     fault_extract_from_ft - Extract fault data from Fourier data.
c     Parameters:
c      w = Array to receive extracted fault data.
c      z = Array containing Fourier spatial domain.
c      org0 = Coordinate 0 of upper left corner of fault in Fourier domain.
c      org1 = Coordinate 1 of upper left corner of fault in Fourier domain.
c      scale = Scale factor applied to all data.

      subroutine fault_extract_from_ft (w, z, org0, org1, scale)
      
      use math_constants
      use ft_data
      use fault_data
      implicit none
      
      real(8), intent(out) :: w(0:1, 0:cell_count1, 0:cell_count0)
      real(8), intent(in) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      integer, intent(in) :: org0
      integer, intent(in) :: org1
      real(8), intent(in) :: scale
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Index of corresponding point in Fourier domain
      
      integer :: m0
      integer :: m1
      
      ! Real and imaginary values, and accumulators
      
      real(8) :: re_val
      real(8) :: im_val
      
      real(8) :: re_val_acc
      real(8) :: im_val_acc
      
      ! Index for smoothing kernel
      
      integer :: si0
      integer :: si1
      
      ! Loop over nodes within fault
      
      do node0 = 0, cell_count0 
        do node1 = 0, cell_count1
          
          ! Get corresponding index into Fourier domain
          
          m0 = node0*cell_stride0 + org0
          m1 = node1*cell_stride1 + org1
          
          ! Initialize accumulators over smoothing
          
          re_val_acc = 0.0d0
          im_val_acc = 0.0d0
          
          ! Loop over entries in the smoothing kernel
          
          do si0 = -(smooth_stride0-1), smooth_stride0-1
            do si1 = -(smooth_stride1-1), smooth_stride1-1
            
              ! Get value from Fourier domain
              
              call ft_cyclic_get (z, m1+si1, m0+si0, re_val, im_val)
              
              ! Add value times smoothing kernel to accumulators
              
              re_val_acc = re_val_acc + re_val*smooth_kernel(si1, si0)
              im_val_acc = im_val_acc + im_val*smooth_kernel(si1, si0)
          
            enddo
          enddo
          
          ! Save scaled accumulated value
          
          w(i_re, node1, node0) = re_val_acc * scale * smooth_scale_re
          w(i_im, node1, node0) = im_val_acc * scale * smooth_scale_im
          
        enddo
      enddo
      
      return
      end subroutine fault_extract_from_ft




c     fault_extract_rot_from_ft - Extract rotated fault data from Fourier data.
c     Parameters:
c      w = Array to receive extracted fault data.
c      z = Array containing Fourier spatial domain.
c      org0 = Coordinate 0 of upper left corner of fault in Fourier domain.
c      org1 = Coordinate 1 of upper left corner of fault in Fourier domain.
c      scale = Scale factor applied to all data.
c     The fault data is rotated 180 degrees from the Fourier data.  Note that
c      org0 and org1 give the upper left corner of the rotated fault.

      subroutine fault_extract_rot_from_ft (w, z, org0, org1, scale)
      
      use math_constants
      use ft_data
      use fault_data
      implicit none
      
      real(8), intent(out) :: w(0:1, 0:cell_count1, 0:cell_count0)
      real(8), intent(in) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      integer, intent(in) :: org0
      integer, intent(in) :: org1
      real(8), intent(in) :: scale
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Index of corresponding point in Fourier domain
      
      integer :: m0
      integer :: m1
      
      ! Real and imaginary values, and accumulators
      
      real(8) :: re_val
      real(8) :: im_val
      
      real(8) :: re_val_acc
      real(8) :: im_val_acc
      
      ! Index for smoothing kernel
      
      integer :: si0
      integer :: si1
      
      ! Loop over nodes within fault
      
      do node0 = 0, cell_count0 
        do node1 = 0, cell_count1
          
          ! Get corresponding index into Fourier domain
          
          m0 = (cell_count0 - node0)*cell_stride0 + org0
          m1 = (cell_count1 - node1)*cell_stride1 + org1
          
          ! Initialize accumulators over smoothing
          
          re_val_acc = 0.0d0
          im_val_acc = 0.0d0
          
          ! Loop over entries in the smoothing kernel
          
          do si0 = -(smooth_stride0-1), smooth_stride0-1
            do si1 = -(smooth_stride1-1), smooth_stride1-1
            
              ! Get value from Fourier domain
              
              call ft_cyclic_get (z, m1+si1, m0+si0, re_val, im_val)
              
              ! Add value times smoothing kernel to accumulators
              
              re_val_acc = re_val_acc + re_val*smooth_kernel(si1, si0)
              im_val_acc = im_val_acc + im_val*smooth_kernel(si1, si0)
          
            enddo
          enddo
          
          ! Save scaled accumulated value
          
          w(i_re, node1, node0) = re_val_acc * scale * smooth_scale_re
          w(i_im, node1, node0) = im_val_acc * scale * smooth_scale_im
          
        enddo
      enddo
      
      return
      end subroutine fault_extract_rot_from_ft








c----------------------------------------------------------------------
c
c     stpl - Selection template functions.




c     stpl_data - Data module.

      module stpl_data
      
      use value2d_include
      implicit none
      
      ! Template value
      
      type(value2d), pointer, save :: stpl_value
      
      ! Flag is true to force zero DC value
      
      logical, save :: stpl_f_zero_dc
      
      end module stpl_data




c     stpl_read_param - Read parameters for selection template.
c     Parameters:
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       2D template value [see value2d]
c       Flag to force zero DC value (0 = false, 1 = true)

      subroutine stpl_read_param (iunit, f_ok)
      
      use stpl_data
      implicit none
      
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: iflag
      
      ! Read template value
      
      call value2d_new (stpl_value, iunit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
        
      ! Zero DC flag
        
      read(iunit,*) iflag
      select case (iflag)
        case (0)
          stpl_f_zero_dc = .false.
        case (1)
          stpl_f_zero_dc = .true.
        case default
          f_ok = .false.
          return
      end select
      
      return
      end subroutine stpl_read_param




c     stpl_force_zero_dc - Force template to have zero DC component.
c     Parameters:
c      v = Array for selection template.
c     Note: The array is in the spatial domain.
c     Note: This function multiplies all negative values by a scale
c      factor, selected to yield the zero DC component.

      subroutine stpl_force_zero_dc (v)
      
      use math_constants
      use ft_data
      implicit none
      
      real(8), intent(inout) :: v(0:1, 0:ft_N1-1, 0:ft_N0-1)
      
      ! Fourier spatial domain coordinates
      
      integer :: m0
      integer :: m1
      
      ! Accumulators for positive and negative values
      
      real(8) :: pos_acc
      real(8) :: neg_acc
      
      ! Scale factor
      
      real(8) :: scale
      
      ! Accumulate sums of all positive and negative terms
      
      pos_acc = 0.0d0
      neg_acc = 0.0d0
      
      do m0 = 0, ft_N0-1
        do m1 = 0, ft_N1-1
          if (v(i_re, m1, m0) < 0.0d0) then
            neg_acc = neg_acc - v(i_re, m1, m0)
          else
            pos_acc = pos_acc + v(i_re, m1, m0)
          endif
        enddo
      enddo
      
      ! Apply required scale factor to all negative terms
      
      scale = pos_acc / neg_acc
      
      do m0 = 0, ft_N0-1
        do m1 = 0, ft_N1-1
          if (v(i_re, m1, m0) < 0.0d0) then
            v(i_re, m1, m0) = v(i_re, m1, m0) * scale
          endif
        enddo
      enddo
      
      return
      end subroutine stpl_force_zero_dc




c     stpl_calc_v - Calculate selection template.
c     Parameters:
c      v = Array for Fourier transform of selection template.
c     Note: The returned array is in the frequency domain.

      subroutine stpl_calc_v (v)
      
      use math_constants
      use ft_data
      use stpl_data
      implicit none
      
      real(8), intent(out) :: v(0:1, 0:ft_N1-1, 0:ft_N0-1)
      
      ! Nyquist coordinates
      
      integer :: half_N0
      integer :: half_N1
      
      ! Fourier point spacing
      
      real(8) :: step_h0
      real(8) :: step_h1
      
      ! Fourier spatial domain coordinates
      
      integer :: m0
      integer :: m1
      
      ! Fourier spatial coordinates in meters
      
      real(8) :: x0
      real(8) :: x1
      
      ! Calculate Nyquist coordinates
      
      half_N0 = ft_N0/2
      half_N1 = ft_N1/2
      
      ! Calculate Fourier point spacing
      
      step_h0 = ft_L0 / dble(ft_N0)
      step_h1 = ft_L1 / dble(ft_N1)
      
      ! Initialize array to all zero
      
      call ft_fill_zero (v)
      
      ! Loop over Fourier spatial domain
      
      do m0 = 0, ft_N0-1
      
        ! Get 0-centered coordinate in meters
        
        if (m0 <= half_N0) then
          x0 = step_h0 * dble(m0)
        else
          x0 = step_h0 * dble(m0 - ft_N0)
        endif
        
        ! Inner loop
        
        do m1 = 0, ft_N1-1
          
          ! Get 0-centered coordinate in meters
          
          if (m1 <= half_N1) then
            x1 = step_h1 * dble(m1)
          else
            x1 = step_h1 * dble(m1 - ft_N1)
          endif
          
          ! Get value as real part of Fourier coefficient
          
          call value2d_eval (stpl_value, v(i_re, m1, m0), x0, x1)
        
        enddo
      enddo
      
      ! If zero DC component is desired, force it
      
      if (stpl_f_zero_dc) then
        call stpl_force_zero_dc (v)
      endif
      
      ! Transform into frequency domain
      
      call ft_forward_fft (v)
      
      return
      end subroutine stpl_calc_v




c     stpl_test_plot - Test selection template.
c     Parameters:
c      in_unit = Unit from which to read parameters.
c      f_ok = Variable receives true if success, false if error.
c     The test consists of generating a selection template,
c      selecting the portion that lies within the fault surface,
c      then writing it out as a table that can be plotted.
c
c     Parameter file format:
c      Output filename
c      Fault and Fourier parameters
c      Selection template parameters

      subroutine stpl_test_plot (in_unit, f_ok)
      
      use io_unit_data
      use math_constants
      use ft_data
      use fault_data
      implicit none
      
      integer, intent(in) :: in_unit
      logical, intent(out) :: f_ok
      
      ! Fourier domain for template
      
      real(8), allocatable :: v(:,:,:)
      
      ! Fault plane
      
      real(8), allocatable :: w(:,:,:)
      
      ! Output filename

      character(max_path) :: out_filename
      
      ! Output unit
      
      integer :: out_unit
      
      ! Read the output filename

      read(in_unit,*) out_filename
      
      ! Read the fault and Fourier parameters
      
      call fault_ft_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the selection template parameters
      
      call stpl_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Allocate Fourier domain for template
      
      allocate(v(0:1, 0:ft_N1-1, 0:ft_N0-1))
      
      ! Allocate fault plane
      
      allocate(w(0:1, 0:cell_count1, 0:cell_count0))
      
      ! Calculate selection template
      
      call stpl_calc_v (v)
      
      ! Transform back to spatial domain
      
      call ft_reverse_fft (v)
      call ft_normalize_fft (v)
      
      ! Extract into fault surface
      
      call fault_extract_from_ft (w, v, 0, 0, 1.0d0)
      
      ! Write into table
      
      call open_unit_w (out_unit, out_filename)
      
      call fault_write_table_re (out_unit, w)
      
      call close_unit (out_unit)
      
      ! Deallocate arrays
      
      deallocate(w)
      deallocate(v)
      
      ! Success
        
      f_ok = .true.
      
      return
      end subroutine stpl_test_plot








c----------------------------------------------------------------------
c
c     vsel - Selector for part of complex value.




c     vsel_data - Data module.

      module vsel_data
      implicit none
      
      ! Return real part of complex value
      
      integer, parameter :: vsel_real = 0
      
      ! Return imaginary part of complex value
      
      integer, parameter :: vsel_imaginary = 1
      
      ! Return magnitude of complex value
      
      integer, parameter :: vsel_magnitude = 2
      
      contains
      
      
      
      
c     vsel_get - Get selected part of complex value
c     Parameters:
c      vsel = Selection.
c      v_re = Real part of complex value.
c      v_im = Imaginary part of complex value.
c     Returns:
c      Selected part of complex value.

      function vsel_get (vsel, v_re, v_im)
      implicit none
      
      real(8) :: vsel_get
      
      integer, intent(in) :: vsel
      real(8), intent(in) :: v_re
      real(8), intent(in) :: v_im
      
      ! Get selected part
      
      select case (vsel)
      case (vsel_real)
        vsel_get = v_re
      case (vsel_imaginary)
        vsel_get = v_im
      case (vsel_magnitude)
        vsel_get = sqrt(v_re*v_re + v_im*v_im)
      case default
        vsel_get = 0.0d0
      end select
      
      return
      end function vsel_get
      
      
      
      
      end module vsel_data








c----------------------------------------------------------------------
c
c     rotpl - Rotation template functions.




c     rotpl_data - Data module.

      module rotpl_data
      
      use value2d_include
      implicit none
      
      ! Flag is true to enable rotation
      
      logical, save :: rotpl_f_enable

      ! Displacement of rotated fault surface, in directions 0, 1
      ! This is measured in fault cells (not fft points), and is the
      !  value that must be added to the upper left corner of the
      !  unrotated fault surface to get the upper left corner of the
      !  rotated fault surface.
      
      integer, save :: rotpl_disp0
      integer, save :: rotpl_disp1
      
      ! Template value
      
      type(value2d), pointer, save :: rotpl_value
      
      end module rotpl_data




c     rotpl_read_param - Read parameters for rotation template.
c     Parameters:
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       Flag to enable rotation (0 = false, 1 = true)
c       2D template value [see value2d] [Not present if flag == 0]

      subroutine rotpl_read_param (iunit, f_ok)
      
      use rotpl_data
      implicit none
      
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: iflag
        
      ! Rotation enable flag
        
      read(iunit,*) iflag
      select case (iflag)
        case (0)
          rotpl_f_enable = .false.
        case (1)
          rotpl_f_enable = .true.
        case default
          f_ok = .false.
          return
      end select
      
      ! Read rotation displacement
      
      if (rotpl_f_enable) then
      
        read(iunit,*) rotpl_disp0, rotpl_disp1
      
      else

        rotpl_disp0 = 0
        rotpl_disp1 = 0
        
      endif
      
      ! Read template value
      
      if (rotpl_f_enable) then
      
        call value2d_new (rotpl_value, iunit, f_ok)
        
        if (.not. f_ok) then
          return
        endif
      
      else
      
        rotpl_value => null()
        
      endif
      
      return
      end subroutine rotpl_read_param




c     rotpl_calc_rot_org - Calculate origin of rotated fault plane.
c     Parameters:
c      org0 = Coordinate 0 of upper left corner of fault in Fourier domain.
c      org1 = Coordinate 1 of upper left corner of fault in Fourier domain.
c      rot_org0 = Receives coordinate 0 of upper left corner of rotated fault in Fourier domain.
c      rot_org1 = Receives coordinate 1 of upper left corner of rotated fault in Fourier domain.
c     The returned coordinates may be outside the ranges 0 to ft_N0-1, and 0 to ft_N1-1.
c      When used, they should be brought into range, for example with ft_cyclic_get. 

      subroutine rotpl_calc_rot_org (org0, org1, rot_org0, rot_org1)
      
      use fault_data
      use rotpl_data
      implicit none

      integer, intent(in) :: org0
      integer, intent(in) :: org1
      integer, intent(out) :: rot_org0
      integer, intent(out) :: rot_org1

      ! Apply displacement, converting from fault cells to fft points

      rot_org0 = rotpl_disp0*cell_stride0 + org0
      rot_org1 = rotpl_disp1*cell_stride1 + org1

      return
      end subroutine rotpl_calc_rot_org




c     rotpl_do_rotation - Rotation 180 degrees if desired.
c     Parameters:
c      w = Array containing fault data.
c      vsel = Value selection, vsel_xxxx.
c      rot_result = Result of operation.
c     Result is 0 if no rotation performed, 1 if rotated 180 degrees.

      subroutine rotpl_do_rotation (w, vsel, rot_result)
      
      use math_constants
      use fault_data
      use vsel_data
      use rotpl_data
      implicit none
      
      real(8), intent(inout) :: w(0:1, 0:cell_count1, 0:cell_count0)
      integer, intent(in) :: vsel
      integer, intent(out) :: rot_result
      
      ! Values of stress and template
      
      real(8) :: v
      real(8) :: tpl
      
      ! Accumulators for 0 and 180 degree rotations
      
      real(8) :: acc_0
      real(8) :: acc_0_row
      
      real(8) :: acc_180
      real(8) :: acc_180_row
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Rotated node index
      
      integer :: rot_node0
      integer :: rot_node1
      
      ! Node location in meters
      
      real(8) :: x0
      real(8) :: x1
      
      ! Temporary for swapping
      
      real(8) :: tmp_re
      real(8) :: tmp_im
      
      ! Upper limit for swapping loop
      
      integer :: node1_hi
      
      ! If not enabled, just return
      
      if (.not. rotpl_f_enable) then
        rot_result = 0;
        return
      endif
      
      ! Accumulate product of template with 0 and 180 rotated values
      
      acc_0 = 0.0d0
      acc_180 = 0.0d0
      
      do node0 = 0, cell_count0
      
        rot_node0 = cell_count0 - node0
        x0 = cell_size0 * dble(node0)
        
        acc_0_row = 0.0d0
        acc_180_row = 0.0d0
         
        do node1 = 0, cell_count1
          
          rot_node1 = cell_count1 - node1
          x1 = cell_size1 * dble(node1)
          
          ! Template value at this location
          
          call value2d_eval (rotpl_value, tpl, x0, x1)
          
          ! Stress value at 0 degree rotation
        
          v = vsel_get (vsel, w(i_re, node1, node0),
     &                        w(i_im, node1, node0))
        
          acc_0_row = acc_0_row + (v * tpl)
          
          ! Stress value at 180 degree rotation
        
          v = vsel_get (vsel, w(i_re, rot_node1, rot_node0),
     &                        w(i_im, rot_node1, rot_node0))
        
          acc_180_row = acc_180_row + (v * tpl)
        
        enddo
        
        acc_0 = acc_0 + acc_0_row
        acc_180 = acc_180 + acc_180_row
        
      enddo
      
      ! If no benefit to rotating, just return
      
      if (acc_0 >= acc_180) then
        rot_result = 0;
        return
      endif
      
      ! Rotate fault surface 180 degrees
      
      do node0 = 0, cell_count0 / 2
      
        rot_node0 = cell_count0 - node0
      
        if (node0 == rot_node0) then
          node1_hi = (cell_count1 - 1) / 2
        else
          node1_hi = cell_count1
        endif
         
        do node1 = 0, node1_hi
          
          rot_node1 = cell_count1 - node1
          
          ! Swap values at rotated locations
          
          tmp_re = w(i_re, node1, node0)
          tmp_im = w(i_im, node1, node0)
          
          w(i_re, node1, node0) = w(i_re, rot_node1, rot_node0)
          w(i_im, node1, node0) = w(i_im, rot_node1, rot_node0)
          
          w(i_re, rot_node1, rot_node0) = tmp_re
          w(i_im, rot_node1, rot_node0) = tmp_im
          
        enddo
      enddo
      
      rot_result = 1
      
      return
      end subroutine rotpl_do_rotation




c     rotpl_do_rotation_2 - Rotation 180 degrees if desired.
c     Parameters:
c      w_0 = Array containing unrotated fault data.
c      w_180 = Array containing rotated fault data.
c      vsel = Value selection, vsel_xxxx.
c      rot_result = Result of operation.
c     Result is 0 if no rotation performed, 1 if rotated 180 degrees.
c     If rotation is enabled, the function tests if the rotated data
c      fits the template better than the unrotated data.  If so, then
c      the contents of w_0 and w_180 are swapped.

      subroutine rotpl_do_rotation_2 (w_0, w_180, vsel, rot_result)
      
      use math_constants
      use fault_data
      use vsel_data
      use rotpl_data
      implicit none
      
      real(8), intent(inout) :: w_0(0:1, 0:cell_count1, 0:cell_count0)
      real(8), intent(inout) :: w_180(0:1, 0:cell_count1, 0:cell_count0)
      integer, intent(in) :: vsel
      integer, intent(out) :: rot_result
      
      ! Values of stress and template
      
      real(8) :: v
      real(8) :: tpl
      
      ! Accumulators for 0 and 180 degree rotations
      
      real(8) :: acc_0
      real(8) :: acc_0_row
      
      real(8) :: acc_180
      real(8) :: acc_180_row
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Node location in meters
      
      real(8) :: x0
      real(8) :: x1
      
      ! Temporary for swapping
      
      real(8) :: tmp_re
      real(8) :: tmp_im
      
      ! If not enabled, just return
      
      if (.not. rotpl_f_enable) then
        rot_result = 0;
        return
      endif
      
      ! Accumulate product of template with 0 and 180 rotated values
      
      acc_0 = 0.0d0
      acc_180 = 0.0d0
      
      do node0 = 0, cell_count0
      
        x0 = cell_size0 * dble(node0)
        
        acc_0_row = 0.0d0
        acc_180_row = 0.0d0
         
        do node1 = 0, cell_count1
          
          x1 = cell_size1 * dble(node1)
          
          ! Template value at this location
          
          call value2d_eval (rotpl_value, tpl, x0, x1)
          
          ! Stress value at 0 degree rotation
        
          v = vsel_get (vsel, w_0(i_re, node1, node0),
     &                        w_0(i_im, node1, node0))
        
          acc_0_row = acc_0_row + (v * tpl)
          
          ! Stress value at 180 degree rotation
        
          v = vsel_get (vsel, w_180(i_re, node1, node0),
     &                        w_180(i_im, node1, node0))
        
          acc_180_row = acc_180_row + (v * tpl)
        
        enddo
        
        acc_0 = acc_0 + acc_0_row
        acc_180 = acc_180 + acc_180_row
        
      enddo
      
      ! If no benefit to rotating, just return
      
      if (acc_0 >= acc_180) then
        rot_result = 0;
        return
      endif
      
      ! Rotate fault surface 180 degrees, by swapping the arrays

      do node0 = 0, cell_count0
        do node1 = 0, cell_count1
          
          ! Swap values at rotated and unrotated locations
          
          tmp_re = w_0(i_re, node1, node0)
          tmp_im = w_0(i_im, node1, node0)
          
          w_0(i_re, node1, node0) = w_180(i_re, node1, node0)
          w_0(i_im, node1, node0) = w_180(i_im, node1, node0)
          
          w_180(i_re, node1, node0) = tmp_re
          w_180(i_im, node1, node0) = tmp_im
          
        enddo
      enddo
      
      rot_result = 1
      
      return
      end subroutine rotpl_do_rotation_2




c     rotpl_test_plot - Test rotation template.
c     Parameters:
c      in_unit = Unit from which to read parameters.
c      f_ok = Variable receives true if success, false if error.
c     The test consists of generating a rotation template,
c      then writing it out as a table that can be plotted.
c
c     Parameter file format:
c      Output filename
c      Fault and Fourier parameters
c      Rotation template parameters

      subroutine rotpl_test_plot (in_unit, f_ok)
      
      use io_unit_data
      use math_constants
      use fault_data
      use rotpl_data
      implicit none
      
      integer, intent(in) :: in_unit
      logical, intent(out) :: f_ok
      
      ! Fault plane
      
      real(8), allocatable :: w(:,:,:)
      
      ! Output filename

      character(max_path) :: out_filename
      
      ! Output unit
      
      integer :: out_unit
      
      ! Value
      
      real(8) :: tpl
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Node location in meters
      
      real(8) :: x0
      real(8) :: x1
      
      ! Read the output filename

      read(in_unit,*) out_filename
      
      ! Read the fault and Fourier parameters
      
      call fault_ft_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the rotation template parameters
      
      call rotpl_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Allocate fault plane
      
      allocate(w(0:1, 0:cell_count1, 0:cell_count0))
      
      ! Fill fault plane with template values
      
      do node0 = 0, cell_count0
      
        x0 = cell_size0 * dble(node0)
         
        do node1 = 0, cell_count1
          
          x1 = cell_size1 * dble(node1)
          
          ! Template value at this location
          
          call value2d_eval (rotpl_value, tpl, x0, x1)
          
          w(i_re, node1, node0) = tpl
          w(i_im, node1, node0) = 0.0d0
          
        enddo
      enddo
      
      ! Write into table
      
      call open_unit_w (out_unit, out_filename)
      
      call fault_write_table_re (out_unit, w)
      
      call close_unit (out_unit)
      
      ! Deallocate arrays
      
      deallocate(w)
      
      ! Success
        
      f_ok = .true.
      
      return
      end subroutine rotpl_test_plot








c----------------------------------------------------------------------
c
c     fzsc - Fault zone scaling information.




c     fzsc_data - Data module.

      module fzsc_data
      implicit none
      
      ! Scaling information
      
      type :: fzsc_info
      
        ! Minimum stress
        
        real(8) :: str_min
        
        ! Maximum stress
        
        real(8) :: str_max
        
        ! Mean stress
        
        real(8) :: str_mean
        
        ! Standard deviation of stress
        
        real(8) :: str_stddev
        
        ! Median of stress
        
        real(8) :: str_median
      
      end type fzsc_info
      
      contains
      


      
c     fzsc_info_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
      
      subroutine fzsc_info_new (this)
      implicit none
      
      type(fzsc_info), pointer :: this  ! out
      
      ! Allocate the object
      
      allocate(this)
      
      return
      end subroutine fzsc_info_new
      
      
      
      
c     fzsc_info_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine fzsc_info_delete (this)
      implicit none
      
      type(fzsc_info), pointer :: this  ! inout
      
      ! If non-null ...
      
      if (associated(this)) then
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine fzsc_info_delete
      
      
      
      
c     fzsc_info_clear - Clear all fields to zero.
c     Parameters:
c      this = Pointer.
      
      subroutine fzsc_info_clear (this)
      implicit none
      
      type(fzsc_info), pointer :: this  ! out
      
      ! Clear fields
      
      this%str_min = 0.0d0
      this%str_max = 0.0d0
      this%str_mean = 0.0d0
      this%str_stddev = 0.0d0
      this%str_median = 0.0d0
      
      return
      end subroutine fzsc_info_clear
      
      
      
      
c     fzsc_info_write - Write scaling info to a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
c
c     File format (one line):
c      str_min, str_max, str_mean, str_stddev, str_median
      
      subroutine fzsc_info_write (this, iunit)
      implicit none
      
      type(fzsc_info), pointer :: this  ! in
      integer, intent(in) :: iunit
      
      ! Write info
      
      write(iunit,'(5(1x,1pe25.16))')
     &  this%str_min,
     &  this%str_max,
     &  this%str_mean,
     &  this%str_stddev,
     &  this%str_median
      
      return
      end subroutine fzsc_info_write
      
      
      
      
c     fzsc_info_read - Read scaling info from a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
c
c     File format (one line):
c      str_min, str_max, str_mean, str_stddev, str_median
      
      subroutine fzsc_info_read (this, iunit)
      implicit none
      
      type(fzsc_info), pointer :: this  ! out
      integer, intent(in) :: iunit
      
      ! Write info
      
      read(iunit,*)
     &  this%str_min,
     &  this%str_max,
     &  this%str_mean,
     &  this%str_stddev,
     &  this%str_median
      
      return
      end subroutine fzsc_info_read
      
      
      
      
c     fzsc_info_dot - Dot product of scaling info.
c     Parameters:
c      this = Pointer.
c      other = Pointer to other object.
c     Returns:
c      Dot product.
c     This allows an fzsc_info structure to function as a set of weights
c      for selecting and weighting individual fields.
      
      function fzsc_info_dot (this, other)
      implicit none
      
      real(8) :: fzsc_info_dot
      
      type(fzsc_info), pointer :: this  ! in
      type(fzsc_info), pointer :: other  ! in
      
      ! Calculate dot product
      
      fzsc_info_dot
     &  = (this%str_min * other%str_min)
     &  + (this%str_max * other%str_max)
     &  + (this%str_mean * other%str_mean)
     &  + (this%str_stddev * other%str_stddev)
     &  + (this%str_median * other%str_median)
      
      return
      end function fzsc_info_dot
      
      
      
      
c     fzsc_info_copy - Copy scaling info from other structure.
c     Parameters:
c      this = Pointer.
c      other = Pointer to other object.
      
      subroutine fzsc_info_copy (this, other)
      implicit none
      
      type(fzsc_info), pointer :: this  ! out
      type(fzsc_info), pointer :: other  ! in
      
      ! Copy fields
      
      this%str_min = other%str_min
      this%str_max = other%str_max
      this%str_mean = other%str_mean
      this%str_stddev = other%str_stddev
      this%str_median = other%str_median
      
      return
      end subroutine fzsc_info_copy
      
      
      
      
c     fzsc_info_add - Add scaling info from other structure.
c     Parameters:
c      this = Pointer.
c      other = Pointer to other object.
      
      subroutine fzsc_info_add (this, other)
      implicit none
      
      type(fzsc_info), pointer :: this  ! inout
      type(fzsc_info), pointer :: other  ! in
      
      ! Add fields
      
      this%str_min = this%str_min + other%str_min
      this%str_max = this%str_max + other%str_max
      this%str_mean = this%str_mean + other%str_mean
      this%str_stddev = this%str_stddev + other%str_stddev
      this%str_median = this%str_median + other%str_median
      
      return
      end subroutine fzsc_info_add
      
      
      
      
c     fzsc_info_scale - Apply scale factor.
c     Parameters:
c      this = Pointer.
c      other = Pointer to other object.
      
      subroutine fzsc_info_scale (this, scale)
      implicit none
      
      type(fzsc_info), pointer :: this  ! inout
      real(8), intent(in) :: scale
      
      ! Scale fields
      
      this%str_min = this%str_min * scale
      this%str_max = this%str_max * scale
      this%str_mean = this%str_mean * scale
      this%str_stddev = this%str_stddev * scale
      this%str_median = this%str_median * scale
      
      return
      end subroutine fzsc_info_scale
      
      
      
      
      end module fzsc_data








c----------------------------------------------------------------------
c
c     sstat - Stress statistics.




c     sstat_data - Data module.

      module sstat_data
      implicit none
      
      ! Range for main fault, in cells, direction 0 (0 to cell_count0)
      ! For stress distribution, along-dip (0 to cell_count_d)
      
      integer, save :: sstat_range_0_lo
      integer, save :: sstat_range_0_hi
      
      ! Range for main fault, in cells, direction 1 (0 to cell_count1)
      ! For stress distribution, along-strike (0 to cell_count_s)
      
      integer, save :: sstat_range_1_lo
      integer, save :: sstat_range_1_hi
      
      ! Range for transition zone, in cells, dir 0 (0 to cell_count0)
      ! For stress distribution, along-dip (0 to cell_count_d)
      
      integer, save :: sstat_inner_0_lo
      integer, save :: sstat_inner_0_hi
      
      ! Range for transition zone, in cells, dir 1 (0 to cell_count1)
      ! For stress distribution, along-strike (0 to cell_count_s)
      
      integer, save :: sstat_inner_1_lo
      integer, save :: sstat_inner_1_hi
      
      ! Range for stopping zone, in cells, dir 0 (0 to cell_count0)
      ! For stress distribution, along-dip (0 to cell_count_d)
      
      integer, save :: sstat_outer_0_lo
      integer, save :: sstat_outer_0_hi
      
      ! Range for stopping zone, in cells, dir 1 (0 to cell_count1)
      ! For stress distribution, along-strike (0 to cell_count_s)
      
      integer, save :: sstat_outer_1_lo
      integer, save :: sstat_outer_1_hi
      
      ! Number of bins to use for histograms
      ! A value of 0 disables histograms
      
      integer, save :: sstat_bin_count
      
      ! Statistics results
      
      type :: sstat_result
      
        ! Number of nodes
        
        integer :: str_nodes
      
        ! Minimum
        
        real(8) :: str_min
        
        ! Maximum
        
        real(8) :: str_max
        
        ! Mean
        
        real(8) :: str_mean
        
        ! Standard deviation
        
        real(8) :: str_stddev
        
        ! Median
        
        real(8) :: str_median
        
        ! Histogram
        
        integer, pointer :: str_hist(:)
        
        ! Mean, measured in bins above minimum
        
        real(8) :: str_mean_bins
        
        ! Standard deviation, measured in bins
        
        real(8) :: str_stddev_bins
        
        ! Median, measured in bins
        
        real(8) :: str_median_bins
        
        ! Gaussian fit parameters
        
        real(8) :: str_gauss_mean   ! mean
        real(8) :: str_gauss_sdev   ! standard deviation
        real(8) :: str_gauss_err    ! sum of absolute errors
        
        ! Cauchy fit parameters
        
        real(8) :: str_cauchy_loc   ! location
        real(8) :: str_cauchy_scale ! scale
        real(8) :: str_cauchy_err   ! sum of absolute errors
        real(8) :: str_cauchy_rel   ! error divided by Gaussian error
      
      end type sstat_result
      
      contains
      


      
c     sstat_result_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
      
      subroutine sstat_result_new (this)
      implicit none
      
      type(sstat_result), pointer :: this
      
      ! Allocate the object
      
      allocate(this)
      this%str_hist => null()
      
      ! Allocate histogram bins
      
      if (sstat_bin_count > 0) then
        allocate(this%str_hist(sstat_bin_count))
      endif
      
      return
      end subroutine sstat_result_new
      
      
      
      
c     sstat_result_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine sstat_result_delete (this)
      implicit none
      
      type(sstat_result), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
      
        ! Deallocate our held objects
        
        if (associated(this%str_hist)) then
          deallocate(this%str_hist)
        endif
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine sstat_result_delete
      
      
      
      
c     sstat_result_clear - Clear all fields to zero.
c     Parameters:
c      this = Pointer.
      
      subroutine sstat_result_clear (this)
      implicit none
      
      type(sstat_result), pointer :: this
      
      integer :: i
      
      ! Clear fields
      
      this%str_nodes = 0
      this%str_min = 0.0d0
      this%str_max = 0.0d0
      this%str_mean = 0.0d0
      this%str_stddev = 0.0d0
      this%str_median = 0.0d0
      
      if (sstat_bin_count > 0) then
        do i = 1, sstat_bin_count 
          this%str_hist(i) = 0
        enddo
      endif
      
      this%str_mean_bins = 0.0d0
      this%str_stddev_bins = 0.0d0
      this%str_median_bins = 0.0d0
      
      this%str_gauss_mean = 0.0d0
      this%str_gauss_sdev = 0.0d0
      this%str_gauss_err = 0.0d0
      
      this%str_cauchy_loc = 0.0d0
      this%str_cauchy_scale = 0.0d0
      this%str_cauchy_err = 0.0d0
      this%str_cauchy_rel = 0.0d0
      
      return
      end subroutine sstat_result_clear
      
      
      
      
c     sstat_result_write - Write results to a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
      
      subroutine sstat_result_write (this, iunit)
      implicit none
      
      type(sstat_result), pointer :: this
      integer, intent(in) :: iunit
      
      integer :: i
      
      ! Write results
      
      if (this%str_nodes > 999999999) then
        write(iunit,'(A,I16)') 'str_nodes = ',this%str_nodes
      else
        write(iunit,'(A,I9)') 'str_nodes = ',this%str_nodes
      endif
      write(iunit,'(A,1pe15.6)') 'str_min = ',this%str_min
      write(iunit,'(A,1pe15.6)') 'str_max = ',this%str_max
      write(iunit,'(A,1pe15.6)') 'str_mean = ',this%str_mean
      write(iunit,'(A,1pe15.6)') 'str_stddev = ',this%str_stddev
      write(iunit,'(A,1pe15.6)') 'str_median = ',this%str_median
      
      write(iunit,*)
      
      ! Write histogram results
      
      if (sstat_bin_count > 0) then
      
        do i = 1, sstat_bin_count
          if (this%str_nodes > 999999999) then
            write(iunit,'(A,I3,A,I16)')
     &        'str_hist(',i,') = ',this%str_hist(i)
          else
            write(iunit,'(A,I3,A,I9)')
     &        'str_hist(',i,') = ',this%str_hist(i)
          endif
        enddo
        
        write(iunit,*)
      
        write(iunit,'(A,1pe15.6)')
     &    'str_mean_bins = ',this%str_mean_bins
        write(iunit,'(A,1pe15.6)')
     &    'str_stddev_bins = ',this%str_stddev_bins
        write(iunit,'(A,1pe15.6)')
     &    'str_median_bins = ',this%str_median_bins
      
        write(iunit,*)
        
        write(iunit,'(A,1pe15.6)')
     &    'str_gauss_mean = ',this%str_gauss_mean
        write(iunit,'(A,1pe15.6)')
     &    'str_gauss_sdev = ',this%str_gauss_sdev
        write(iunit,'(A,1pe15.6)')
     &    'str_gauss_err = ',this%str_gauss_err
        
        write(iunit,*)
        
        write(iunit,'(A,1pe15.6)')
     &    'str_cauchy_loc = ',this%str_cauchy_loc
        write(iunit,'(A,1pe15.6)')
     &    'str_cauchy_scale = ',this%str_cauchy_scale
        write(iunit,'(A,1pe15.6)')
     &    'str_cauchy_err = ',this%str_cauchy_err
        write(iunit,'(A,1pe15.6)')
     &    'str_cauchy_rel = ',this%str_cauchy_rel
        
        write(iunit,*)
      
      endif
      
      return
      end subroutine sstat_result_write
      
      
      
      
c     sstat_result_save_scaling - Save scaling information.
c     Parameters:
c      this = Pointer.
c      fzsc = Pointer to scaling info structure.
      
      subroutine sstat_result_save_scaling (this, fzsc)
      
      use fzsc_data
      implicit none
      
      type(sstat_result), pointer :: this
      type(fzsc_info), pointer :: fzsc
      
      ! Copy scaling info
      
      if (associated(fzsc)) then
      
        fzsc%str_min = this%str_min
        fzsc%str_max = this%str_max
        fzsc%str_mean = this%str_mean
        fzsc%str_stddev = this%str_stddev
        fzsc%str_median = this%str_median
        
      endif

      return
      end subroutine sstat_result_save_scaling
      
      
      
      
c     sstat_result_calc - Calculate results.
c     Parameters:
c      this = Pointer.
c      w = Array containing fault data.
c      vsel = Value selection, vsel_xxxx.
c      range_0_lo = Range of cells in direction 0, low.
c      range_0_hi = Range of cells in direction 0, high.
c      range_1_lo = Range of cells in direction 1, low.
c      range_1_hi = Range of cells in direction 1, high.
      
      subroutine sstat_result_calc (
     &  this, w, vsel, range_0_lo, range_0_hi, range_1_lo, range_1_hi)
      
      use math_constants
      use fault_data
      use vsel_data
      implicit none
      
      type(sstat_result), pointer :: this
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      integer, intent(in) :: vsel
      integer, intent(in) :: range_0_lo
      integer, intent(in) :: range_0_hi
      integer, intent(in) :: range_1_lo
      integer, intent(in) :: range_1_hi
      
      ! Value
      
      real(8) :: v
      
      ! Accumulator
      
      real(8) :: acc
      real(8) :: acc_row
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Bin index
      
      integer :: i
      
      ! Histogram bin width
      
      real(8) :: hist_delta
      
      ! Clear all fields
      
      call sstat_result_clear (this)
      
      ! Get the total number of nodes
      
      this%str_nodes = (range_0_hi + 1 - range_0_lo)
     &                   * (range_1_hi + 1 - range_1_lo)
      
      if (this%str_nodes <= 0) then
        return
      endif
      
      ! Minimum and maximum
      
      v = vsel_get (vsel, w(i_re, range_1_lo, range_0_lo),
     &                    w(i_im, range_1_lo, range_0_lo))
      
      this%str_min = v
      this%str_max = v
      
      do node0 = range_0_lo, range_0_hi 
        do node1 = range_1_lo, range_1_hi
        
          v = vsel_get (vsel, w(i_re, node1, node0),
     &                        w(i_im, node1, node0))
        
          if (v < this%str_min) then
            this%str_min = v
          endif
        
          if (v > this%str_max) then
            this%str_max = v
          endif
        
        enddo
      enddo
      
      ! Mean
      
      acc = 0.0d0
      
      do node0 = range_0_lo, range_0_hi
      
        acc_row = 0.0d0
         
        do node1 = range_1_lo, range_1_hi
        
          v = vsel_get (vsel, w(i_re, node1, node0),
     &                        w(i_im, node1, node0))
        
          acc_row = acc_row + v
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_mean = acc / dble(this%str_nodes)
      
      ! Standard deviation
      
      acc = 0.0d0
      
      do node0 = range_0_lo, range_0_hi
      
        acc_row = 0.0d0
         
        do node1 = range_1_lo, range_1_hi
        
          v = vsel_get (vsel, w(i_re, node1, node0),
     &                        w(i_im, node1, node0))
          
          v = v - this%str_mean
          acc_row = acc_row + (v*v)
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_stddev = sqrt(acc / dble(max(1, this%str_nodes - 1)))
      
      ! If doing histograms ...
      
      if (sstat_bin_count > 0) then
      
        ! Histogram
      
        do i = 1, sstat_bin_count
          this%str_hist(i) = 0
        enddo
      
        hist_delta = (this%str_max - this%str_min)/dble(sstat_bin_count)
      
        do node0 = range_0_lo, range_0_hi 
          do node1 = range_1_lo, range_1_hi
        
            v = vsel_get (vsel, w(i_re, node1, node0),
     &                          w(i_im, node1, node0))
          
            i = ceiling((v - this%str_min)/hist_delta)
            i = max(i, 1)
            i = min(i, sstat_bin_count)
            this%str_hist(i) = this%str_hist(i) + 1
        
          enddo
        enddo
      
        ! Finish histogram calculations
      
        call sstat_result_finish_hist (this)
      
      endif
      
      return
      end subroutine sstat_result_calc
      
      
      
      
c     sstat_result_calc_ext - Calculate results external to range.
c     Parameters:
c      this = Pointer.
c      w = Array containing fault data.
c      vsel = Value selection, vsel_xxxx.
c      range_0_lo = Range of cells in direction 0, low.
c      range_0_hi = Range of cells in direction 0, high.
c      range_1_lo = Range of cells in direction 1, low.
c      range_1_hi = Range of cells in direction 1, high.
      
      subroutine sstat_result_calc_ext (
     &  this, w, vsel, range_0_lo, range_0_hi, range_1_lo, range_1_hi)
      
      use math_constants
      use fault_data
      use vsel_data
      implicit none
      
      type(sstat_result), pointer :: this
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      integer, intent(in) :: vsel
      integer, intent(in) :: range_0_lo
      integer, intent(in) :: range_0_hi
      integer, intent(in) :: range_1_lo
      integer, intent(in) :: range_1_hi
      
      ! Value
      
      real(8) :: v
      
      ! Accumulator
      
      real(8) :: acc
      real(8) :: acc_row
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Bin index
      
      integer :: i
      
      ! Histogram bin width
      
      real(8) :: hist_delta
      
      ! Clear all fields
      
      call sstat_result_clear (this)
      
      ! Get the total number of nodes
      
      this%str_nodes = (cell_count0 + 1) * (cell_count1 + 1)
     &  - (range_0_hi + 1 - range_0_lo) * (range_1_hi + 1 - range_1_lo)
      
      if (this%str_nodes <= 0) then
        return
      endif
      
      ! Minimum and maximum
      
      if (range_0_lo > 0 .or. range_1_lo > 0) then
        node0 = 0
        node1 = 0
      else
        node0 = cell_count0
        node1 = cell_count1
      endif
      
      v = vsel_get (vsel, w(i_re, node1, node0),
     &                    w(i_im, node1, node0))
      
      this%str_min = v
      this%str_max = v
      
      do node0 = 0, cell_count0 
        do node1 = 0, cell_count1
          if (node0 >= range_0_lo .and. node0 <= range_0_hi
     &        .and. node1 >= range_1_lo .and. node1 <= range_1_hi) then
            cycle
          endif
      
          v = vsel_get (vsel, w(i_re, node1, node0),
     &                        w(i_im, node1, node0))
        
          if (v < this%str_min) then
            this%str_min = v
          endif
        
          if (v > this%str_max) then
            this%str_max = v
          endif
        
        enddo
      enddo
      
      ! Mean
      
      acc = 0.0d0
      
      do node0 = 0, cell_count0 
      
        acc_row = 0.0d0
         
        do node1 = 0, cell_count1
          if (node0 >= range_0_lo .and. node0 <= range_0_hi
     &        .and. node1 >= range_1_lo .and. node1 <= range_1_hi) then
            cycle
          endif
      
          v = vsel_get (vsel, w(i_re, node1, node0),
     &                        w(i_im, node1, node0))
        
          acc_row = acc_row + v
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_mean = acc / dble(this%str_nodes)
      
      ! Standard deviation
      
      acc = 0.0d0
      
      do node0 = 0, cell_count0 
      
        acc_row = 0.0d0
         
        do node1 = 0, cell_count1
          if (node0 >= range_0_lo .and. node0 <= range_0_hi
     &        .and. node1 >= range_1_lo .and. node1 <= range_1_hi) then
            cycle
          endif
      
          v = vsel_get (vsel, w(i_re, node1, node0),
     &                        w(i_im, node1, node0))
          
          v = v - this%str_mean
          acc_row = acc_row + (v*v)
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_stddev = sqrt(acc / dble(max(1, this%str_nodes - 1)))
      
      ! If doing histograms ...
      
      if (sstat_bin_count > 0) then
      
        ! Histogram
      
        do i = 1, sstat_bin_count
          this%str_hist(i) = 0
        enddo
      
        hist_delta = (this%str_max - this%str_min)/dble(sstat_bin_count)
      
        do node0 = 0, cell_count0 
          do node1 = 0, cell_count1
            if (node0 >= range_0_lo .and. node0 <= range_0_hi
     &         .and. node1 >= range_1_lo .and. node1 <= range_1_hi) then
              cycle
            endif
      
            v = vsel_get (vsel, w(i_re, node1, node0),
     &                          w(i_im, node1, node0))
          
            i = ceiling((v - this%str_min)/hist_delta)
            i = max(i, 1)
            i = min(i, sstat_bin_count)
            this%str_hist(i) = this%str_hist(i) + 1
        
          enddo
        enddo
      
        ! Finish histogram calculations
      
        call sstat_result_finish_hist (this)
      
      endif
      
      return
      end subroutine sstat_result_calc_ext
      
      
      
      
c     sstat_result_calc_ann - Calculate results in annular range.
c     Parameters:
c      this = Pointer.
c      w = Array containing fault data.
c      vsel = Value selection, vsel_xxxx.
c      inner_0_lo = Inner range of cells in direction 0, low.
c      inner_0_hi = Inner range of cells in direction 0, high.
c      inner_1_lo = Inner range of cells in direction 1, low.
c      inner_1_hi = Inner range of cells in direction 1, high.
c      outer_0_lo = Outer range of cells in direction 0, low.
c      outer_0_hi = Outer range of cells in direction 0, high.
c      outer_1_lo = Outer range of cells in direction 1, low.
c      outer_1_hi = Outer range of cells in direction 1, high.
c     This function operates on nodes that lie in the outer range, but
c      do not lie in the inner ranges.  Ranges are inclusive.
      
      subroutine sstat_result_calc_ann (
     &  this, w, vsel, inner_0_lo, inner_0_hi, inner_1_lo, inner_1_hi,
     &                 outer_0_lo, outer_0_hi, outer_1_lo, outer_1_hi)
      
      use math_constants
      use fault_data
      use vsel_data
      implicit none
      
      type(sstat_result), pointer :: this
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      integer, intent(in) :: vsel
      integer, intent(in) :: inner_0_lo
      integer, intent(in) :: inner_0_hi
      integer, intent(in) :: inner_1_lo
      integer, intent(in) :: inner_1_hi
      integer, intent(in) :: outer_0_lo
      integer, intent(in) :: outer_0_hi
      integer, intent(in) :: outer_1_lo
      integer, intent(in) :: outer_1_hi
      
      ! Value
      
      real(8) :: v
      
      ! Accumulator
      
      real(8) :: acc
      real(8) :: acc_row
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Bin index
      
      integer :: i
      
      ! Histogram bin width
      
      real(8) :: hist_delta
      
      ! Clear all fields
      
      call sstat_result_clear (this)
      
      ! Get the total number of nodes
      
      this%str_nodes
     &  = (outer_0_hi + 1 - outer_0_lo) * (outer_1_hi + 1 - outer_1_lo)
     &  - (inner_0_hi + 1 - inner_0_lo) * (inner_1_hi + 1 - inner_1_lo)
      
      if (this%str_nodes <= 0) then
        return
      endif
      
      ! Minimum and maximum
      
      if (inner_0_lo > outer_0_lo .or. inner_1_lo > outer_1_lo) then
        node0 = outer_0_lo
        node1 = outer_1_lo
      else
        node0 = outer_0_hi
        node1 = outer_1_hi
      endif
      
      v = vsel_get (vsel, w(i_re, node1, node0),
     &                    w(i_im, node1, node0))
      
      this%str_min = v
      this%str_max = v
      
      do node0 = outer_0_lo, outer_0_hi 
        do node1 = outer_1_lo, outer_1_hi
          if (node0 >= inner_0_lo .and. node0 <= inner_0_hi
     &        .and. node1 >= inner_1_lo .and. node1 <= inner_1_hi) then
            cycle
          endif
      
          v = vsel_get (vsel, w(i_re, node1, node0),
     &                        w(i_im, node1, node0))
        
          if (v < this%str_min) then
            this%str_min = v
          endif
        
          if (v > this%str_max) then
            this%str_max = v
          endif
        
        enddo
      enddo
      
      ! Mean
      
      acc = 0.0d0
      
      do node0 = outer_0_lo, outer_0_hi 
      
        acc_row = 0.0d0
         
        do node1 = outer_1_lo, outer_1_hi
          if (node0 >= inner_0_lo .and. node0 <= inner_0_hi
     &        .and. node1 >= inner_1_lo .and. node1 <= inner_1_hi) then
            cycle
          endif
      
          v = vsel_get (vsel, w(i_re, node1, node0),
     &                        w(i_im, node1, node0))
        
          acc_row = acc_row + v
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_mean = acc / dble(this%str_nodes)
      
      ! Standard deviation
      
      acc = 0.0d0
      
      do node0 = outer_0_lo, outer_0_hi 
      
        acc_row = 0.0d0
         
        do node1 = outer_1_lo, outer_1_hi
          if (node0 >= inner_0_lo .and. node0 <= inner_0_hi
     &        .and. node1 >= inner_1_lo .and. node1 <= inner_1_hi) then
            cycle
          endif
      
          v = vsel_get (vsel, w(i_re, node1, node0),
     &                        w(i_im, node1, node0))
          
          v = v - this%str_mean
          acc_row = acc_row + (v*v)
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_stddev = sqrt(acc / dble(max(1, this%str_nodes - 1)))
      
      ! If doing histograms ...
      
      if (sstat_bin_count > 0) then
      
        ! Histogram
      
        do i = 1, sstat_bin_count
          this%str_hist(i) = 0
        enddo
      
        hist_delta = (this%str_max - this%str_min)/dble(sstat_bin_count)
      
        do node0 = outer_0_lo, outer_0_hi 
          do node1 = outer_1_lo, outer_1_hi
            if (node0 >= inner_0_lo .and. node0 <= inner_0_hi
     &         .and. node1 >= inner_1_lo .and. node1 <= inner_1_hi) then
              cycle
            endif
      
            v = vsel_get (vsel, w(i_re, node1, node0),
     &                          w(i_im, node1, node0))
          
            i = ceiling((v - this%str_min)/hist_delta)
            i = max(i, 1)
            i = min(i, sstat_bin_count)
            this%str_hist(i) = this%str_hist(i) + 1
        
          enddo
        enddo
      
        ! Finish histogram calculations
      
        call sstat_result_finish_hist (this)
      
      endif
      
      return
      end subroutine sstat_result_calc_ann
      
      
      
      
c     sstat_result_calc_ft - Calculate results for Fourier domain.
c     Parameters:
c      this = Pointer.
c      z = Array containing Fourier spatial domain.
c      vsel = Value selection, vsel_xxxx.
c      scale = Scale factor applied to all data.
      
      subroutine sstat_result_calc_ft (this, z, vsel, scale)
      
      use math_constants
      use ft_data
      use vsel_data
      implicit none
      
      type(sstat_result), pointer :: this
      real(8), intent(in) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      integer, intent(in) :: vsel
      real(8), intent(in) :: scale
      
      ! Ranges
      
      integer :: range_0_lo
      integer :: range_0_hi
      integer :: range_1_lo
      integer :: range_1_hi
      
      ! Value
      
      real(8) :: v
      
      ! Accumulator
      
      real(8) :: acc
      real(8) :: acc_row
      
      ! Node index within fault
      
      integer :: node0
      integer :: node1
      
      ! Bin index
      
      integer :: i
      
      ! Histogram bin width
      
      real(8) :: hist_delta
      
      ! Set ranges
      
      range_0_lo = 0
      range_0_hi = ft_N0-1
      range_1_lo = 0
      range_1_hi = ft_N1-1
      
      ! Clear all fields
      
      call sstat_result_clear (this)
      
      ! Get the total number of nodes
      
      this%str_nodes = (range_0_hi + 1 - range_0_lo)
     &                   * (range_1_hi + 1 - range_1_lo)
      
      if (this%str_nodes <= 0) then
        return
      endif
      
      ! Minimum and maximum
      
      v = vsel_get (vsel, z(i_re, range_1_lo, range_0_lo) * scale,
     &                    z(i_im, range_1_lo, range_0_lo) * scale)
      
      this%str_min = v
      this%str_max = v
      
      do node0 = range_0_lo, range_0_hi 
        do node1 = range_1_lo, range_1_hi
          
          v = vsel_get (vsel, z(i_re, node1, node0) * scale,
     &                        z(i_im, node1, node0) * scale)
        
          if (v < this%str_min) then
            this%str_min = v
          endif
        
          if (v > this%str_max) then
            this%str_max = v
          endif
        
        enddo
      enddo
      
      ! Mean
      
      acc = 0.0d0
      
      do node0 = range_0_lo, range_0_hi
      
        acc_row = 0.0d0
         
        do node1 = range_1_lo, range_1_hi
          
          v = vsel_get (vsel, z(i_re, node1, node0) * scale,
     &                        z(i_im, node1, node0) * scale)
        
          acc_row = acc_row + v
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_mean = acc / dble(this%str_nodes)
      
      ! Standard deviation
      
      acc = 0.0d0
      
      do node0 = range_0_lo, range_0_hi
      
        acc_row = 0.0d0
         
        do node1 = range_1_lo, range_1_hi
          
          v = vsel_get (vsel, z(i_re, node1, node0) * scale,
     &                        z(i_im, node1, node0) * scale)
          
          v = v - this%str_mean
          acc_row = acc_row + (v*v)
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_stddev = sqrt(acc / dble(max(1, this%str_nodes - 1)))
      
      ! If doing histograms ...
      
      if (sstat_bin_count > 0) then
      
        ! Histogram
      
        do i = 1, sstat_bin_count
          this%str_hist(i) = 0
        enddo
      
        hist_delta = (this%str_max - this%str_min)/dble(sstat_bin_count)
      
        do node0 = range_0_lo, range_0_hi 
          do node1 = range_1_lo, range_1_hi
          
            v = vsel_get (vsel, z(i_re, node1, node0) * scale,
     &                          z(i_im, node1, node0) * scale)
          
            i = ceiling((v - this%str_min)/hist_delta)
            i = max(i, 1)
            i = min(i, sstat_bin_count)
            this%str_hist(i) = this%str_hist(i) + 1
        
          enddo
        enddo
      
        ! Finish histogram calculations
      
        call sstat_result_finish_hist (this)
      
      endif
      
      return
      end subroutine sstat_result_calc_ft
      
      
      
      
c     sstat_result_calc_sd - Calculate results for stress distribution.
c     Parameters:
c      this = Pointer.
c      sd = Array containing stress distribution.
c      cell_count_d = Total number of cells along-dip.
c      cell_count_s = Total number of cells along-strike.
c      range_d_lo = Range of cells along-dip, low.
c      range_d_hi = Range of cells along-dip, high.
c      range_s_lo = Range of cells along-strike, low.
c      range_s_hi = Range of cells along-strike, high.
      
      subroutine sstat_result_calc_sd (
     &  this, sd, cell_count_d, cell_count_s,
     &  range_d_lo, range_d_hi, range_s_lo, range_s_hi)
      
      implicit none
      
      type(sstat_result), pointer :: this
      integer, intent(in) :: cell_count_d
      integer, intent(in) :: cell_count_s
      real(8), intent(in) :: sd(0:cell_count_s, 0:cell_count_d)
      integer, intent(in) :: range_d_lo
      integer, intent(in) :: range_d_hi
      integer, intent(in) :: range_s_lo
      integer, intent(in) :: range_s_hi
      
      ! Value
      
      real(8) :: v
      
      ! Accumulator
      
      real(8) :: acc
      real(8) :: acc_row
      
      ! Node index within fault
      
      integer :: node_d
      integer :: node_s
      
      ! Bin index
      
      integer :: i
      
      ! Histogram bin width
      
      real(8) :: hist_delta
      
      ! Clear all fields
      
      call sstat_result_clear (this)
      
      ! Get the total number of nodes
      
      this%str_nodes = (range_d_hi + 1 - range_d_lo)
     &                   * (range_s_hi + 1 - range_s_lo)
      
      if (this%str_nodes <= 0) then
        return
      endif
      
      ! Minimum and maximum
      
      v = sd(range_s_lo, range_d_lo)
      
      this%str_min = v
      this%str_max = v
      
      do node_d = range_d_lo, range_d_hi 
        do node_s = range_s_lo, range_s_hi
        
          v = sd(node_s, node_d)
        
          if (v < this%str_min) then
            this%str_min = v
          endif
        
          if (v > this%str_max) then
            this%str_max = v
          endif
        
        enddo
      enddo
      
      ! Mean
      
      acc = 0.0d0
      
      do node_d = range_d_lo, range_d_hi
      
        acc_row = 0.0d0
         
        do node_s = range_s_lo, range_s_hi
        
          v = sd(node_s, node_d)
        
          acc_row = acc_row + v
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_mean = acc / dble(this%str_nodes)
      
      ! Standard deviation
      
      acc = 0.0d0
      
      do node_d = range_d_lo, range_d_hi
      
        acc_row = 0.0d0
         
        do node_s = range_s_lo, range_s_hi
        
          v = sd(node_s, node_d)
          
          v = v - this%str_mean
          acc_row = acc_row + (v*v)
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_stddev = sqrt(acc / dble(max(1, this%str_nodes - 1)))
      
      ! If doing histograms ...
      
      if (sstat_bin_count > 0) then
      
        ! Histogram
      
        do i = 1, sstat_bin_count
          this%str_hist(i) = 0
        enddo
      
        hist_delta = (this%str_max - this%str_min)/dble(sstat_bin_count)
      
        do node_d = range_d_lo, range_d_hi 
          do node_s = range_s_lo, range_s_hi
        
            v = sd(node_s, node_d)
          
            i = ceiling((v - this%str_min)/hist_delta)
            i = max(i, 1)
            i = min(i, sstat_bin_count)
            this%str_hist(i) = this%str_hist(i) + 1
        
          enddo
        enddo
      
        ! Finish histogram calculations
      
        call sstat_result_finish_hist (this)
      
      endif
      
      return
      end subroutine sstat_result_calc_sd
      
      
      
      
c     sstat_result_calc_sd_ann - Calculate results in annular range,
c                                for a stress distribution.
c     Parameters:
c      this = Pointer.
c      sd = Array containing stress distribution.
c      cell_count_d = Total number of cells along-dip.
c      cell_count_s = Total number of cells along-strike.
c      inner_d_lo = Inner range of cells along-dip, low.
c      inner_d_hi = Inner range of cells along-dip, high.
c      inner_s_lo = Inner range of cells along-strike, low.
c      inner_s_hi = Inner range of cells along-strike, high.
c      outer_d_lo = Outer range of cells along-dip, low.
c      outer_d_hi = Outer range of cells along-dip, high.
c      outer_s_lo = Outer range of cells along-strike, low.
c      outer_s_hi = Outer range of cells along-strike, high.
c     This function operates on nodes that lie in the outer range, but
c      do not lie in the inner ranges.  Ranges are inclusive.
      
      subroutine sstat_result_calc_sd_ann (
     &  this, sd, cell_count_d, cell_count_s,
     &  inner_d_lo, inner_d_hi, inner_s_lo, inner_s_hi,
     &  outer_d_lo, outer_d_hi, outer_s_lo, outer_s_hi)
      
      implicit none
      
      type(sstat_result), pointer :: this
      integer, intent(in) :: cell_count_d
      integer, intent(in) :: cell_count_s
      real(8), intent(in) :: sd(0:cell_count_s, 0:cell_count_d)
      integer, intent(in) :: inner_d_lo
      integer, intent(in) :: inner_d_hi
      integer, intent(in) :: inner_s_lo
      integer, intent(in) :: inner_s_hi
      integer, intent(in) :: outer_d_lo
      integer, intent(in) :: outer_d_hi
      integer, intent(in) :: outer_s_lo
      integer, intent(in) :: outer_s_hi
      
      ! Value
      
      real(8) :: v
      
      ! Accumulator
      
      real(8) :: acc
      real(8) :: acc_row
      
      ! Node index within fault
      
      integer :: node_d
      integer :: node_s
      
      ! Bin index
      
      integer :: i
      
      ! Histogram bin width
      
      real(8) :: hist_delta
      
      ! Clear all fields
      
      call sstat_result_clear (this)
      
      ! Get the total number of nodes
      
      this%str_nodes
     &  = (outer_d_hi + 1 - outer_d_lo) * (outer_s_hi + 1 - outer_s_lo)
     &  - (inner_d_hi + 1 - inner_d_lo) * (inner_s_hi + 1 - inner_s_lo)
      
      if (this%str_nodes <= 0) then
        return
      endif
      
      ! Minimum and maximum
      
      if (inner_d_lo > outer_d_lo .or. inner_s_lo > outer_s_lo) then
        node_d = outer_d_lo
        node_s = outer_s_lo
      else
        node_d = outer_d_hi
        node_s = outer_s_hi
      endif
      
      v = sd(node_s, node_d)
      
      this%str_min = v
      this%str_max = v
      
      do node_d = outer_d_lo, outer_d_hi 
        do node_s = outer_s_lo, outer_s_hi
          if (node_d >= inner_d_lo .and. node_d <= inner_d_hi
     &       .and. node_s >= inner_s_lo .and. node_s <= inner_s_hi) then
            cycle
          endif
      
          v = sd(node_s, node_d)
        
          if (v < this%str_min) then
            this%str_min = v
          endif
        
          if (v > this%str_max) then
            this%str_max = v
          endif
        
        enddo
      enddo
      
      ! Mean
      
      acc = 0.0d0
      
      do node_d = outer_d_lo, outer_d_hi 
      
        acc_row = 0.0d0
         
        do node_s = outer_s_lo, outer_s_hi
          if (node_d >= inner_d_lo .and. node_d <= inner_d_hi
     &       .and. node_s >= inner_s_lo .and. node_s <= inner_s_hi) then
            cycle
          endif
      
          v = sd(node_s, node_d)
        
          acc_row = acc_row + v
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_mean = acc / dble(this%str_nodes)
      
      ! Standard deviation
      
      acc = 0.0d0
      
      do node_d = outer_d_lo, outer_d_hi 
      
        acc_row = 0.0d0
         
        do node_s = outer_s_lo, outer_s_hi
          if (node_d >= inner_d_lo .and. node_d <= inner_d_hi
     &       .and. node_s >= inner_s_lo .and. node_s <= inner_s_hi) then
            cycle
          endif
      
          v = sd(node_s, node_d)
          
          v = v - this%str_mean
          acc_row = acc_row + (v*v)
        
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      this%str_stddev = sqrt(acc / dble(max(1, this%str_nodes - 1)))
      
      ! If doing histograms ...
      
      if (sstat_bin_count > 0) then
      
        ! Histogram
      
        do i = 1, sstat_bin_count
          this%str_hist(i) = 0
        enddo
      
        hist_delta = (this%str_max - this%str_min)/dble(sstat_bin_count)
      
        do node_d = outer_d_lo, outer_d_hi 
          do node_s = outer_s_lo, outer_s_hi
            if (node_d >= inner_d_lo .and. node_d <= inner_d_hi
     &       .and. node_s >= inner_s_lo .and. node_s <= inner_s_hi) then
              cycle
            endif
      
            v = sd(node_s, node_d)
          
            i = ceiling((v - this%str_min)/hist_delta)
            i = max(i, 1)
            i = min(i, sstat_bin_count)
            this%str_hist(i) = this%str_hist(i) + 1
        
          enddo
        enddo
      
        ! Finish histogram calculations
      
        call sstat_result_finish_hist (this)
      
      endif
      
      return
      end subroutine sstat_result_calc_sd_ann
      
      
      
      
c     sstat_result_finish_hist - Finish histogram calculations.
c     Parameters:
c      this = Pointer.
      
      subroutine sstat_result_finish_hist (this)
      
      implicit none
      
      type(sstat_result), pointer :: this
      
      ! Histogram bin width
      
      real(8) :: hist_delta
      
      ! Bin index
      
      integer :: i
      
      ! Running total of bin counts
      
      integer :: acc_new
      integer :: acc_old
      
      ! Get histogram bin width
      
      hist_delta = (this%str_max - this%str_min)/dble(sstat_bin_count)
    
      ! Mean in bins
    
      this%str_mean_bins = (this%str_mean - this%str_min)/hist_delta
    
      ! Standard deviation in bins
    
      this%str_stddev_bins = this%str_stddev/hist_delta
      
      ! Loop searching for bin that contains median
      
      acc_new = 0
      
      do i = 1, sstat_bin_count
        
        ! Update accumulators
      
        acc_old = acc_new
        acc_new = acc_new + this%str_hist(i)
        
        ! If we have reached the median ...
        
        if ((acc_new * 2) >= this%str_nodes) then
          
          ! Interpolate to get median in bins
          
          this%str_median_bins = dble(i - 1)
     &       + ( dble(this%str_nodes - (acc_old * 2))
     &             / dble((acc_new * 2) - (acc_old * 2)) )
          
          ! Now get estimated median
          
          this%str_median =
     &      (this%str_median_bins * hist_delta) + this%str_min
          
          ! Stop scan
          
          exit
        
        endif
        
      enddo
    
      ! Fit distributions
    
      call sstat_result_fit_gaussian (this)
      
      return
      end subroutine sstat_result_finish_hist
      
      
      
      
c     sstat_result_fit_gaussian - Fit distributions to histogram.
c     Parameters:
c      this = Pointer.
      
      subroutine sstat_result_fit_gaussian (this)
      
      implicit none
      
      type(sstat_result), pointer :: this
      
      ! Histogram in floating-point
      
      real(8), allocatable :: ahist(:)
      
      ! Trial probability density function.

      real(8), allocatable :: tpdf(:)
      
      ! Trial Gaussian parameters
      
      real(8) :: gauss_mean
      real(8) :: gauss_sdev
      integer :: i_gauss_mean
      integer :: i_gauss_sdev
      
      ! Trial Cauchy parameters
      
      real(8) :: cauchy_loc
      real(8) :: cauchy_scale
      integer :: i_cauchy_loc
      integer :: i_cauchy_scale
      
      ! Smallest pdf we will consider
      
      real(8) :: smpdf
      
      ! Current error
      
      real(8) :: err
      
      ! Parameter resolution in bins is 1/i_pres
      
      integer :: i_pres
      
      ! Indexes into bins
      
      integer :: i
      integer :: j
      
      ! Temporaries
      
      real(8) :: u
      real(8) :: v
      
      ! Allocate our work arrays
      
      allocate(ahist(sstat_bin_count))
      allocate(tpdf(sstat_bin_count))
      
      ! Convert histogram to floating-point
      
      do i = 1, sstat_bin_count
        ahist(i) = dble(this%str_hist(i))
      enddo
      
      ! Fit Gaussian to histogram
      
      this%str_gauss_mean = -1.0d0
      this%str_gauss_sdev = -1.0d0
      
      this%str_gauss_err = 0.0d0  ! initialize to error for zero pdf
      do i = 1, sstat_bin_count
        this%str_gauss_err = this%str_gauss_err + ahist(i)
      enddo
      
      smpdf = 0.01d0 / dble(this%str_nodes)
      
      if (sstat_bin_count <= 50) then
        i_pres = 10   ! Parameter resolution
      else if (sstat_bin_count <= 100) then
        i_pres = 5    ! Parameter resolution
      else if (sstat_bin_count <= 250) then
        i_pres = 2    ! Parameter resolution
      else
        i_pres = 1    ! Parameter resolution
      endif
      
      ! Loop over possible means, at interval of 1/i_pres bin
      
      do i_gauss_mean = i_pres, sstat_bin_count*i_pres
        gauss_mean = dble(i_gauss_mean) / dble(i_pres)
        
        ! Loop over possible std deviations, at 1/i_pres bin intervals
        
        do i_gauss_sdev = 1, sstat_bin_count*i_pres
          gauss_sdev = dble(i_gauss_sdev) / dble(i_pres)
          
          ! Calculate probability distribution
          
          do i = 1, sstat_bin_count
            u = (dble(i) - gauss_mean)/gauss_sdev
            tpdf(i) = exp(-0.5d0*u*u)
          enddo
          
          ! Total error is the sum of absolute values of per-bin errors
          ! (The optimum multiplier for the pdf always makes an exact
          ! match at one of the bins, so we can find the optimum
          ! multiplier by trying a match at each bin.)
          
          do i = 1, sstat_bin_count
            if (tpdf(i) >= smpdf) then
              v = ahist(i)/tpdf(i)
              err = 0.0d0
              do j = 1, sstat_bin_count
                err = err + abs(v*tpdf(j) - ahist(j))
              enddo
              if (err < this%str_gauss_err) then
                this%str_gauss_err = err
                this%str_gauss_mean = gauss_mean
                this%str_gauss_sdev = gauss_sdev
              endif
            endif
          enddo
          
        enddo
        
      enddo
      
      ! Fit Cauchy to histogram
      
      this%str_cauchy_loc = -1.0d0
      this%str_cauchy_scale = -1.0d0
      
      this%str_cauchy_err = 0.0d0  ! initialize to error for zero pdf
      do i = 1, sstat_bin_count
        this%str_cauchy_err = this%str_cauchy_err + ahist(i)
      enddo
      
      smpdf = 0.01d0 / dble(this%str_nodes)
      
      if (sstat_bin_count <= 50) then
        i_pres = 10   ! Parameter resolution
      else if (sstat_bin_count <= 100) then
        i_pres = 5    ! Parameter resolution
      else if (sstat_bin_count <= 250) then
        i_pres = 2    ! Parameter resolution
      else
        i_pres = 1    ! Parameter resolution
      endif
      
      ! Loop over possible locations, at interval of 1/i_pres bin
      
      do i_cauchy_loc = i_pres, sstat_bin_count*i_pres
        cauchy_loc = dble(i_cauchy_loc) / dble(i_pres)
        
        ! Loop over possible scale factors, at 1/i_pres bin intervals
        
        do i_cauchy_scale = 1, sstat_bin_count*i_pres
          cauchy_scale = dble(i_cauchy_scale) / dble(i_pres)
          
          ! Calculate probability distribution
          
          do i = 1, sstat_bin_count
            u = (dble(i) - cauchy_loc)/cauchy_scale
            tpdf(i) = 1.0d0 / (1.0d0 + u*u)
          enddo
          
          ! Total error is the sum of absolute values of per-bin errors
          ! (The optimum multiplier for the pdf always makes an exact
          ! match at one of the bins, so we can find the optimum
          ! multiplier by trying a match at each bin.)
          
          do i = 1, sstat_bin_count
            if (tpdf(i) >= smpdf) then
              v = ahist(i)/tpdf(i)
              err = 0.0d0
              do j = 1, sstat_bin_count
                err = err + abs(v*tpdf(j) - ahist(j))
              enddo
              if (err < this%str_cauchy_err) then
                this%str_cauchy_err = err
                this%str_cauchy_loc = cauchy_loc
                this%str_cauchy_scale = cauchy_scale
              endif
            endif
          enddo
          
        enddo
        
      enddo
      
      ! Cauchy error relative to gauss error
      
      this%str_cauchy_rel = this%str_cauchy_err / this%str_gauss_err
      
      ! Deallocate our work arrays
      
      deallocate(ahist)
      deallocate(tpdf)
      
      return
      end subroutine sstat_result_fit_gaussian
      
      
      
      
c     sstat_result_calc_entire - Calculate results for entire fault.
c     Parameters:
c      this = Pointer.
c      w = Array containing fault data.
c      vsel = Value selection, vsel_xxxx.
      
      subroutine sstat_result_calc_entire (this, w, vsel)
      
      use fault_data
      implicit none
      
      type(sstat_result), pointer :: this
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      integer, intent(in) :: vsel
      
      call sstat_result_calc (
     &  this, w, vsel, 0, cell_count0, 0, cell_count1)
      
      return
      end subroutine sstat_result_calc_entire
      
      
      
      
c     sstat_result_calc_selected - Calculate results for selected range.
c     Parameters:
c      this = Pointer.
c      w = Array containing fault data.
c      vsel = Value selection, vsel_xxxx.
      
      subroutine sstat_result_calc_selected (this, w, vsel)
      
      use fault_data
      implicit none
      
      type(sstat_result), pointer :: this
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      integer, intent(in) :: vsel
      
      call sstat_result_calc (
     &  this, w, vsel,
     &  sstat_range_0_lo, sstat_range_0_hi,
     &  sstat_range_1_lo, sstat_range_1_hi)
      
      return
      end subroutine sstat_result_calc_selected
      
      
      
      
c     sstat_result_calc_selected_ext - Calculate results external to selected range.
c     Parameters:
c      this = Pointer.
c      w = Array containing fault data.
c      vsel = Value selection, vsel_xxxx.
      
      subroutine sstat_result_calc_selected_ext (this, w, vsel)
      
      use fault_data
      implicit none
      
      type(sstat_result), pointer :: this
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      integer, intent(in) :: vsel
      
      call sstat_result_calc_ext (
     &  this, w, vsel,
     &  sstat_range_0_lo, sstat_range_0_hi,
     &  sstat_range_1_lo, sstat_range_1_hi)
      
      return
      end subroutine sstat_result_calc_selected_ext
      
      
      
      
c     sstat_result_calc_selected_ann - Calculate results for annular stop zone.
c     Parameters:
c      this = Pointer.
c      w = Array containing fault data.
c      vsel = Value selection, vsel_xxxx.
      
      subroutine sstat_result_calc_selected_ann (this, w, vsel)
      
      use fault_data
      implicit none
      
      type(sstat_result), pointer :: this
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      integer, intent(in) :: vsel
      
      call sstat_result_calc_ann (
     &  this, w, vsel,
     &  sstat_inner_0_lo, sstat_inner_0_hi,
     &  sstat_inner_1_lo, sstat_inner_1_hi,
     &  sstat_outer_0_lo, sstat_outer_0_hi,
     &  sstat_outer_1_lo, sstat_outer_1_hi)
      
      return
      end subroutine sstat_result_calc_selected_ann
      
      
      
      
c     sstat_result_calc_sd_entire - Calculate results for entire fault
c                                   for a stress distribution.
c     Parameters:
c      this = Pointer.
c      sd = Array containing stress distribution.
c      cell_count_d = Total number of cells along-dip.
c      cell_count_s = Total number of cells along-strike.
      
      subroutine sstat_result_calc_sd_entire (
     &  this, sd, cell_count_d, cell_count_s)
      
      implicit none
      
      type(sstat_result), pointer :: this
      integer, intent(in) :: cell_count_d
      integer, intent(in) :: cell_count_s
      real(8), intent(in) :: sd(0:cell_count_s, 0:cell_count_d)
      
      call sstat_result_calc_sd (
     &  this, sd, cell_count_d, cell_count_s,
     &  0, cell_count_d, 0, cell_count_s)
      
      return
      end subroutine sstat_result_calc_sd_entire
      
      
      
      
c     sstat_result_calc_sd_selected - Calculate results for selected range
c                                     for a stress distribution.
c     Parameters:
c      this = Pointer.
c      sd = Array containing stress distribution.
c      cell_count_d = Total number of cells along-dip.
c      cell_count_s = Total number of cells along-strike.
      
      subroutine sstat_result_calc_sd_selected (
     &  this, sd, cell_count_d, cell_count_s)
      
      implicit none
      
      type(sstat_result), pointer :: this
      integer, intent(in) :: cell_count_d
      integer, intent(in) :: cell_count_s
      real(8), intent(in) :: sd(0:cell_count_s, 0:cell_count_d)
      
      call sstat_result_calc_sd (
     &  this, sd, cell_count_d, cell_count_s,
     &  sstat_range_0_lo, sstat_range_0_hi,
     &  sstat_range_1_lo, sstat_range_1_hi)
      
      return
      end subroutine sstat_result_calc_sd_selected
      
      
      
      
c     sstat_result_calc_sd_selected_ann - Calculate results for annular stop zone
c                                         for a stress distribution.
c     Parameters:
c      this = Pointer.
c      sd = Array containing stress distribution.
c      cell_count_d = Total number of cells along-dip.
c      cell_count_s = Total number of cells along-strike.
      
      subroutine sstat_result_calc_sd_selected_ann (
     &  this, sd, cell_count_d, cell_count_s)
      
      implicit none
      
      type(sstat_result), pointer :: this
      integer, intent(in) :: cell_count_d
      integer, intent(in) :: cell_count_s
      real(8), intent(in) :: sd(0:cell_count_s, 0:cell_count_d)
      
      call sstat_result_calc_sd_ann (
     &  this, sd, cell_count_d, cell_count_s,
     &  sstat_inner_0_lo, sstat_inner_0_hi,
     &  sstat_inner_1_lo, sstat_inner_1_hi,
     &  sstat_outer_0_lo, sstat_outer_0_hi,
     &  sstat_outer_1_lo, sstat_outer_1_hi)
      
      return
      end subroutine sstat_result_calc_sd_selected_ann
      
      
      
      
c     sstat_result_calc_sd_selected_ext - Calculate results external to selected range
c                                         for a stress distribution.
c     Parameters:
c      this = Pointer.
c      sd = Array containing stress distribution.
c      cell_count_d = Total number of cells along-dip.
c      cell_count_s = Total number of cells along-strike.
      
      subroutine sstat_result_calc_sd_selected_ext (
     &  this, sd, cell_count_d, cell_count_s)
      
      implicit none
      
      type(sstat_result), pointer :: this
      integer, intent(in) :: cell_count_d
      integer, intent(in) :: cell_count_s
      real(8), intent(in) :: sd(0:cell_count_s, 0:cell_count_d)
      
      call sstat_result_calc_sd_ann (
     &  this, sd, cell_count_d, cell_count_s,
     &  sstat_range_0_lo, sstat_range_0_hi,
     &  sstat_range_1_lo, sstat_range_1_hi,
     &  0, cell_count_d, 0, cell_count_s)
      
      return
      end subroutine sstat_result_calc_sd_selected_ext
      
      
      
      
      end module sstat_data




c     sstat_read_param - Read parameters for statistics.
c     Parameters:
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       range_0_lo, range_0_hi
c       range_1_lo, range_1_hi
c       Number of bins in histograms

      subroutine sstat_read_param (iunit, f_ok)
      
      use sstat_data
      implicit none
      
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      integer :: iflag
      
      ! Main fault range in direction 0
        
      read(iunit,*) sstat_range_0_lo, sstat_range_0_hi
      
      if (sstat_range_0_lo > sstat_range_0_hi) then
        f_ok = .false.
        return
      endif
      
      ! Main fault range in direction 1
        
      read(iunit,*) sstat_range_1_lo, sstat_range_1_hi
      
      if (sstat_range_1_lo > sstat_range_1_hi) then
        f_ok = .false.
        return
      endif
      
      ! Threshold zone range in direction 0
        
      read(iunit,*) sstat_inner_0_lo, sstat_inner_0_hi
      
      if (sstat_inner_0_lo > sstat_inner_0_hi
     &    .or. sstat_inner_0_lo > sstat_range_0_lo
     &    .or. sstat_inner_0_hi < sstat_range_0_hi) then
        f_ok = .false.
        return
      endif
      
      ! Threshold zone range in direction 1
        
      read(iunit,*) sstat_inner_1_lo, sstat_inner_1_hi
      
      if (sstat_inner_1_lo > sstat_inner_1_hi
     &    .or. sstat_inner_1_lo > sstat_range_1_lo
     &    .or. sstat_inner_1_hi < sstat_range_1_hi) then
        f_ok = .false.
        return
      endif
      
      ! Stopping zone range in direction 0
        
      read(iunit,*) sstat_outer_0_lo, sstat_outer_0_hi
      
      if (sstat_outer_0_lo > sstat_outer_0_hi
     &    .or. sstat_outer_0_lo > sstat_inner_0_lo
     &    .or. sstat_outer_0_hi < sstat_inner_0_hi) then
        f_ok = .false.
        return
      endif
      
      ! Stopping zone range in direction 1
        
      read(iunit,*) sstat_outer_1_lo, sstat_outer_1_hi
      
      if (sstat_outer_1_lo > sstat_outer_1_hi
     &    .or. sstat_outer_1_lo > sstat_inner_1_lo
     &    .or. sstat_outer_1_hi < sstat_inner_1_hi) then
        f_ok = .false.
        return
      endif
      
      ! Number of histogram bins
        
      read(iunit,*) sstat_bin_count
      
      if (sstat_bin_count < 0) then
        f_ok = .false.
        return
      endif
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine sstat_read_param




c     sstat_include - Module to hold client routines, provide interfaces.

      module sstat_include
      implicit none
      
      contains




c     sstat_calc_and_write - Calculate and write statistics.
c     Parameters:
c      iunit = Fortran unit number.
c      w = Array containing fault data.

      subroutine sstat_calc_and_write (iunit, w)
      
      use fault_data
      use sstat_data
      use vsel_data
      implicit none
      
      integer, intent(in) :: iunit
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      
      type(sstat_result), pointer :: s_result
      
      ! Allocate result structure
      
      call sstat_result_new (s_result)
      
      ! Results for entire fault
      
      write(iunit,'(A)') '***** Statistics for entire fault'
      write(iunit,*)
      
      call sstat_result_calc_entire (s_result, w, vsel_real)
      call sstat_result_write (s_result, iunit)
      
      ! Results for selected range
      
      write(iunit,'(A)') '***** Statistics for selected range'
      write(iunit,*)
      
      call sstat_result_calc_selected (s_result, w, vsel_real)
      call sstat_result_write (s_result, iunit)
      
      ! Results external to selected range
      
      write(iunit,'(A)') '***** Statistics external to selected range'
      write(iunit,*)
      
      call sstat_result_calc_selected_ext (s_result, w, vsel_real)
      call sstat_result_write (s_result, iunit)
      
      ! Deallocate result structure
      
      call sstat_result_delete (s_result)
      
      return
      end subroutine sstat_calc_and_write




c     sstat_calc_and_write_2 - Calculate and write statistics, ver 2.
c     Parameters:
c      iunit = Fortran unit number.
c      w = Array containing fault data.
c      entire_fzsc = Structure to receive entire fault scaling info.
c      main_fzsc = Structure to receive main fault scaling info.
c      stop_fzsc = Structure to receive stopping zone scaling info.

      subroutine sstat_calc_and_write_2 (
     &  iunit, w, entire_fzsc, main_fzsc, stop_fzsc)
      
      use fault_data
      use sstat_data
      use vsel_data
      use fzsc_data
      implicit none
      
      integer, intent(in) :: iunit
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      type(fzsc_info), pointer :: entire_fzsc  ! out
      type(fzsc_info), pointer :: main_fzsc  ! out
      type(fzsc_info), pointer :: stop_fzsc  ! out
      
      type(sstat_result), pointer :: s_result
      
      ! Allocate result structure
      
      call sstat_result_new (s_result)
      
      ! Results for entire fault
      
      write(iunit,'(A)') '***** Statistics for entire fault'
      write(iunit,*)
      
      call sstat_result_calc_entire (s_result, w, vsel_real)
      call sstat_result_write (s_result, iunit)
      
      call sstat_result_save_scaling (s_result, entire_fzsc)
      
      ! Results for selected main fault
      
      write(iunit,'(A)') '***** Statistics for main fault'
      write(iunit,*)
      
      call sstat_result_calc_selected (s_result, w, vsel_real)
      call sstat_result_write (s_result, iunit)
      
      call sstat_result_save_scaling (s_result, main_fzsc)
      
      ! Results for annular stopping zone
      
      write(iunit,'(A)') '***** Statistics for stopping zone'
      write(iunit,*)
      
      call sstat_result_calc_selected_ann (s_result, w, vsel_real)
      call sstat_result_write (s_result, iunit)
      
      call sstat_result_save_scaling (s_result, stop_fzsc)
      
      ! Deallocate result structure
      
      call sstat_result_delete (s_result)
      
      return
      end subroutine sstat_calc_and_write_2




c     sstat_calc_and_write_ft - Calculate and write statistics for Fourier domain.
c     Parameters:
c      iunit = Fortran unit number.
c      z = Array containing Fourier spatial domain.
c      scale = Scale factor applied to all data.

      subroutine sstat_calc_and_write_ft (iunit, z, scale)
      
      use ft_data
      use sstat_data
      use vsel_data
      implicit none
      
      integer, intent(in) :: iunit
      real(8), intent(in) :: z(0:1, 0:ft_N1-1, 0:ft_N0-1)
      real(8), intent(in) :: scale
      
      type(sstat_result), pointer :: s_result
      
      ! Allocate result structure
      
      call sstat_result_new (s_result)
      
      ! Results for entire fault
      
      write(iunit,'(A)') '***** Statistics for Fourier domain'
      write(iunit,*)
      
      call sstat_result_calc_ft (s_result, z, vsel_real, scale)
      call sstat_result_write (s_result, iunit)
      
      ! Deallocate result structure
      
      call sstat_result_delete (s_result)
      
      return
      end subroutine sstat_calc_and_write_ft




c     sstat_calc_sd_and_write - Calculate and write statistics.
c                               for a stress distribution.
c     Parameters:
c      iunit = Fortran unit number.
c      sd = Array containing stress distribution.
c      cell_count_d = Total number of cells along-dip.
c      cell_count_s = Total number of cells along-strike.

      subroutine sstat_calc_sd_and_write (
     &  iunit, sd, cell_count_d, cell_count_s)
      
      use sstat_data
      implicit none
      
      integer, intent(in) :: iunit
      integer, intent(in) :: cell_count_d
      integer, intent(in) :: cell_count_s
      real(8), intent(in) :: sd(0:cell_count_s, 0:cell_count_d)
      
      type(sstat_result), pointer :: s_result
      
      ! Allocate result structure
      
      call sstat_result_new (s_result)
      
      ! Results for entire fault
      
      write(iunit,'(A)') '***** Statistics for entire fault'
      write(iunit,*)
      
      call sstat_result_calc_sd_entire (
     &  s_result, sd, cell_count_d, cell_count_s)
      call sstat_result_write (s_result, iunit)
      
      ! Results for selected range
      
      write(iunit,'(A)') '***** Statistics for selected range'
      write(iunit,*)
      
      call sstat_result_calc_sd_selected (
     &  s_result, sd, cell_count_d, cell_count_s)
      call sstat_result_write (s_result, iunit)
      
      ! Results external to selected range
      
      write(iunit,'(A)') '***** Statistics external to selected range'
      write(iunit,*)
      
      call sstat_result_calc_sd_selected_ext (
     &  s_result, sd, cell_count_d, cell_count_s)
      call sstat_result_write (s_result, iunit)
      
      ! Deallocate result structure
      
      call sstat_result_delete (s_result)
      
      return
      end subroutine sstat_calc_sd_and_write




c     sstat_calc_sd_and_write_2 - Calculate and write statistics, ver 2,
c                                 for a stress distribution.
c     Parameters:
c      iunit = Fortran unit number.
c      sd = Array containing stress distribution.
c      cell_count_d = Total number of cells along-dip.
c      cell_count_s = Total number of cells along-strike.
c      entire_fzsc = Structure to receive entire fault scaling info.
c      main_fzsc = Structure to receive main fault scaling info.
c      stop_fzsc = Structure to receive stopping zone scaling info.

      subroutine sstat_calc_sd_and_write_2 (
     &  iunit, sd, cell_count_d, cell_count_s,
     &  entire_fzsc, main_fzsc, stop_fzsc)
      
      use sstat_data
      use fzsc_data
      implicit none
      
      integer, intent(in) :: iunit
      integer, intent(in) :: cell_count_d
      integer, intent(in) :: cell_count_s
      real(8), intent(in) :: sd(0:cell_count_s, 0:cell_count_d)
      type(fzsc_info), pointer :: entire_fzsc  ! out
      type(fzsc_info), pointer :: main_fzsc  ! out
      type(fzsc_info), pointer :: stop_fzsc  ! out
      
      type(sstat_result), pointer :: s_result
      
      ! Allocate result structure
      
      call sstat_result_new (s_result)
      
      ! Results for entire fault
      
      write(iunit,'(A)') '***** Statistics for entire fault'
      write(iunit,*)
      
      call sstat_result_calc_sd_entire (
     &  s_result, sd, cell_count_d, cell_count_s)
      call sstat_result_write (s_result, iunit)
      
      call sstat_result_save_scaling (s_result, entire_fzsc)
      
      ! Results for selected main fault
      
      write(iunit,'(A)') '***** Statistics for main fault'
      write(iunit,*)
      
      call sstat_result_calc_sd_selected (
     &  s_result, sd, cell_count_d, cell_count_s)
      call sstat_result_write (s_result, iunit)
      
      call sstat_result_save_scaling (s_result, main_fzsc)
      
      ! Results for annular stopping zone
      
      write(iunit,'(A)') '***** Statistics for stopping zone'
      write(iunit,*)
      
      call sstat_result_calc_sd_selected_ann (
     &  s_result, sd, cell_count_d, cell_count_s)
      call sstat_result_write (s_result, iunit)
      
      call sstat_result_save_scaling (s_result, stop_fzsc)
      
      ! Deallocate result structure
      
      call sstat_result_delete (s_result)
      
      return
      end subroutine sstat_calc_sd_and_write_2
      
      
      
      
      end module sstat_include








c----------------------------------------------------------------------
c
c     rawsdist - Raw stress distribution.




c     rawsdist_data - Data module.

      module rawsdist_data
      
      use fzsc_data
      implicit none
      
      ! Raw stress distribution holder
      
      type :: rawsdist_holder
        
        ! Linked list pointer for use by clients
        
        type(rawsdist_holder), pointer :: next
        
        ! Realization and sub-realization
        
        integer :: realiz
        integer :: sub_realiz
      
        ! Number of cells along strike, dip
        
        integer :: cell_count_s
        integer :: cell_count_d
        
        ! Size of a cell in meters along strike, dip
        
        real(8) :: cell_size_s
        real(8) :: cell_size_d
        
        ! Low and high coordinates along strike, dip
        ! Note these are for plotting, include a cell-size pixel around
        !  each node, and hence are one cell larger than the fault in
        !  each direction.
        
        real(8) :: coord_s_lo
        real(8) :: coord_s_hi
        
        real(8) :: coord_d_lo
        real(8) :: coord_d_hi
        
        ! True if stress is along strike
        
        logical :: f_along_strike
      
        ! Hypocenter location along strike, dip; or -1 if unknown
        
        integer :: hypo_s
        integer :: hypo_d
        
        ! Hypocenter radius along strike, dip
        
        real(8) :: hypo_radius_s
        real(8) :: hypo_radius_d
        
        ! Selection template convolution
        
        real(8) :: sel_conv
        
        ! Statistics for entire fault
        
        type(fzsc_info), pointer :: entire_fzsc
        
        ! Statistics for main fault
        
        type(fzsc_info), pointer :: main_fzsc
        
        ! Statistics for stopping zone
        
        type(fzsc_info), pointer :: stop_fzsc
        
        ! Stress distribution
        ! sd(0:cell_count_s, 0:cell_count_d)
        
        real(8), pointer :: sd(:,:)
      
      end type rawsdist_holder
      
      contains
      


      
c     rawsdist_holder_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
      
      subroutine rawsdist_holder_new (this)
      implicit none
      
      type(rawsdist_holder), pointer :: this
      
      ! Allocate the object
      
      allocate(this)
      this%next => null()
      this%entire_fzsc => null()
      this%main_fzsc => null()
      this%stop_fzsc => null()
      this%sd => null()
      
      call fzsc_info_new (this%entire_fzsc)
      call fzsc_info_new (this%main_fzsc)
      call fzsc_info_new (this%stop_fzsc)
      
      return
      end subroutine rawsdist_holder_new
      
      
      
      
c     rawsdist_holder_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine rawsdist_holder_delete (this)
      implicit none
      
      type(rawsdist_holder), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
      
        ! Deallocate our held objects
        
        if (associated(this%stop_fzsc)) then
          call fzsc_info_delete (this%stop_fzsc)
        endif
        
        if (associated(this%main_fzsc)) then
          call fzsc_info_delete (this%main_fzsc)
        endif
        
        if (associated(this%entire_fzsc)) then
          call fzsc_info_delete (this%entire_fzsc)
        endif
        
        if (associated(this%sd)) then
          deallocate(this%sd)
        endif
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine rawsdist_holder_delete
      
      
      
      
c     rawsdist_holder_clear - Clear all fields to zero.
c     Parameters:
c      this = Pointer.
      
      subroutine rawsdist_holder_clear (this)
      implicit none
      
      type(rawsdist_holder), pointer :: this
      
      integer :: i
      
      ! Clear fields
      
      this%realiz = -1
      this%sub_realiz = -1
      
      this%cell_count_s = 0
      this%cell_count_d = 0
      
      this%cell_size_s = 0.0d0
      this%cell_size_d = 0.0d0
      
      this%coord_s_lo = 0.0d0
      this%coord_s_hi = 0.0d0
      
      this%coord_d_lo = 0.0d0
      this%coord_d_hi = 0.0d0
      
      this%f_along_strike = .true.
      
      this%hypo_s = -1
      this%hypo_d = -1
      
      this%hypo_radius_s = 0.0d0
      this%hypo_radius_d = 0.0d0
      
      this%sel_conv = 0.0d0
      
      call fzsc_info_clear (this%entire_fzsc)
      call fzsc_info_clear (this%main_fzsc)
      call fzsc_info_clear (this%stop_fzsc)
      
      ! Free the stress distribution
        
      if (associated(this%sd)) then
        deallocate(this%sd)
      endif
      this%sd => null()
      
      return
      end subroutine rawsdist_holder_clear
      
      
      
      
c     rawsdist_holder_write - Write to a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
c
c     File format:
c      cell_count_d+1, cell_count_s+1, coord_s_lo, coord_s_hi, coord_d_lo, coord_d_hi
c      Distribution in row-major order
c      realiz, sub_realiz
c      cell_size_s, cell_size_d
c      f_along_strike as an integer (0 = false, 1 = true)
c      hypo_s, hypo_d
c      hypo_radius_s, hypo_radius_d
c      sel_conv
c      entire_fzsc  [see fzsc_info_write]
c      main_fzsc  [see fzsc_info_write]
c      stop_fzsc  [see fzsc_info_write]
c
c     The distribution is written so that there is a line break at the
c      end of each row.  There are also line breaks within rows.
      
      subroutine rawsdist_holder_write (this, iunit)
      implicit none
      
      type(rawsdist_holder), pointer :: this
      integer, intent(in) :: iunit
      
      integer :: i
      integer :: node_s
      integer :: node_d
      integer :: iflag
      
      ! Write header line
      
      write(iunit,'(2(1x,I9),4(1x,1pe25.16))')
     &  this%cell_count_d + 1,
     &  this%cell_count_s + 1,
     &  this%coord_s_lo,
     &  this%coord_s_hi,
     &  this%coord_d_lo,
     &  this%coord_d_hi
      
      ! Write the stress distribution, by rows
      
      do node_d = 0, this%cell_count_d
        do node_s = 0, this%cell_count_s, 4
          write(iunit,'(4(1x,1pe25.16))')
     &     (this%sd(i, node_d),
     &        i = node_s, min(node_s+3, this%cell_count_s))
        enddo
      enddo
      
      ! Write realization
      
      write(iunit,'(2(1x,I9))')
     &  this%realiz,
     &  this%sub_realiz
      
      ! Write cell size
      
      write(iunit,'(2(1x,1pe25.16))')
     &  this%cell_size_s,
     &  this%cell_size_d
      
      ! Write direction flag
      
      if (this%f_along_strike) then
        iflag = 1
      else
        iflag = 0
      endif
      
      write(iunit,'(1x,I9)')
     &  iflag
      
      ! Write hypocenter location
      
      write(iunit,'(2(1x,I9))')
     &  this%hypo_s,
     &  this%hypo_d
      
      ! Write hypocenter radius
      
      write(iunit,'(2(1x,1pe25.16))')
     &  this%hypo_radius_s,
     &  this%hypo_radius_d
      
      ! Write selection template convolution
      
      write(iunit,'(1x,1pe25.16)')
     &  this%sel_conv
      
      ! Write statistics
      
      call fzsc_info_write (this%entire_fzsc, iunit)
      call fzsc_info_write (this%main_fzsc, iunit)
      call fzsc_info_write (this%stop_fzsc, iunit)
      
      return
      end subroutine rawsdist_holder_write
      
      
      
      
c     rawsdist_holder_read - Read from a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
c
c     File format:
c      cell_count_d+1, cell_count_s+1, coord_s_lo, coord_s_hi, coord_d_lo, coord_d_hi
c      Distribution in row-major order
c      realiz, sub_realiz
c      cell_size_s, cell_size_d
c      f_along_strike as an integer (0 = false, 1 = true)
c      hypo_s, hypo_d
c      hypo_radius_s, hypo_radius_d
c      sel_conv
c      entire_fzsc  [see fzsc_info_read]
c      main_fzsc  [see fzsc_info_read]
c      stop_fzsc  [see fzsc_info_read]
c
c     The file is assumed to be written by rawsdist_holder_write.
      
      subroutine rawsdist_holder_read (this, iunit)
      implicit none
      
      type(rawsdist_holder), pointer :: this
      integer, intent(in) :: iunit
      
      integer :: i
      integer :: node_s
      integer :: node_d
      integer :: iflag
      
      ! Clear the structure
      
      call rawsdist_holder_clear (this)
      
      ! Read header line
      
      read(iunit,*)
     &  this%cell_count_d,
     &  this%cell_count_s,
     &  this%coord_s_lo,
     &  this%coord_s_hi,
     &  this%coord_d_lo,
     &  this%coord_d_hi
      
      this%cell_count_d = this%cell_count_d - 1
      this%cell_count_s = this%cell_count_s - 1
      
      ! Allocate the stress distribution
      
      allocate(this%sd(0:this%cell_count_s, 0:this%cell_count_d))
      
      ! Read the stress distribution, by rows
      
      do node_d = 0, this%cell_count_d
        do node_s = 0, this%cell_count_s, 4
          read(iunit,*)
     &     (this%sd(i, node_d),
     &        i = node_s, min(node_s+3, this%cell_count_s))
        enddo
      enddo
      
      ! Read realization
      
      read(iunit,*)
     &  this%realiz,
     &  this%sub_realiz
      
      ! Read cell size
      
      read(iunit,*)
     &  this%cell_size_s,
     &  this%cell_size_d
      
      ! Read direction flag
      
      read(iunit,*)
     &  iflag
      
      if (iflag == 0) then
        this%f_along_strike = .false.
      else
        this%f_along_strike = .true.
      endif
      
      ! Read hypocenter location
      
      read(iunit,*)
     &  this%hypo_s,
     &  this%hypo_d
      
      ! Read hypocenter radius
      
      read(iunit,*)
     &  this%hypo_radius_s,
     &  this%hypo_radius_d
      
      ! Read selection template convolution
      
      read(iunit,*)
     &  this%sel_conv
      
      ! Read statistics
      
      call fzsc_info_read (this%entire_fzsc, iunit)
      call fzsc_info_read (this%main_fzsc, iunit)
      call fzsc_info_read (this%stop_fzsc, iunit)
      
      return
      end subroutine rawsdist_holder_read
      
      
      
      
c     rawsdist_holder_fill - Fill with generator results.
c     Parameters:
c      this = Pointer.
c      sub_realiz = Sub-realization number.
c      w = Array containing fault data.
c      vsel = Value selection, vsel_xxxx.
c      sel_conv = Selection template convolution value.
c      entire_fzsc = Structure containing entire fault scaling info.
c      main_fzsc = Structure containing main fault scaling info.
c      stop_fzsc = Structure containing stopping zone scaling info.
      
      subroutine rawsdist_holder_fill (
     &  this, sub_realiz, w, vsel, sel_conv,
     &  entire_fzsc, main_fzsc, stop_fzsc)
      
      use math_constants
      use fault_data
      use vsel_data
      implicit none
      
      type(rawsdist_holder), pointer :: this
      integer, intent(in) :: sub_realiz
      real(8), intent(in) :: w(0:1, 0:cell_count1, 0:cell_count0)
      integer, intent(in) :: vsel
      real(8), intent(in) :: sel_conv
      type(fzsc_info), pointer :: entire_fzsc  ! in
      type(fzsc_info), pointer :: main_fzsc  ! in
      type(fzsc_info), pointer :: stop_fzsc  ! in
      
      integer :: realiz
      integer :: node_s
      integer :: node_d
      real(8) :: v
      
      ! Clear the structure
      
      call rawsdist_holder_clear (this)
      
      ! Fill realization
      
      call rgen_get_realiz (realiz)
      
      this%realiz = realiz
      this%sub_realiz = sub_realiz
      
      ! Fill cell count, size, direction
      
      if (f_strike_slip) then
      
        this%cell_count_s = cell_count1
        this%cell_count_d = cell_count0
        
        this%cell_size_s = cell_size1
        this%cell_size_d = cell_size0
      
        this%f_along_strike = .true.
      
      else
      
        this%cell_count_s = cell_count0
        this%cell_count_d = cell_count1
        
        this%cell_size_s = cell_size0
        this%cell_size_d = cell_size1
      
        this%f_along_strike = .false.
      
      endif
      
      ! Fill coordinate ranges, origin at top center of fault
      
      this%coord_s_lo
     &  = -0.5d0 * dble(this%cell_count_s + 1) * this%cell_size_s
      this%coord_s_hi
     &  =  0.5d0 * dble(this%cell_count_s + 1) * this%cell_size_s
      
      this%coord_d_lo = 0.0d0
      this%coord_d_hi = dble(this%cell_count_d + 1) * this%cell_size_d
      
      ! Allocate the stress distribution
      
      allocate(this%sd(0:this%cell_count_s, 0:this%cell_count_d))
      
      ! Fill the stress distribution
      
      do node_d = 0, this%cell_count_d
        do node_s = 0, this%cell_count_s
        
          if (f_strike_slip) then
            v = vsel_get (vsel, w(i_re, node_s, node_d),
     &                          w(i_im, node_s, node_d))
          else
            v = vsel_get (vsel, w(i_re, node_d, node_s),
     &                          w(i_im, node_d, node_s))
          endif
          
          this%sd(node_s, node_d) = v
       
        enddo
      enddo
      
      ! Fill selection template convolution
      
      this%sel_conv = sel_conv
      
      ! Fill statistics
      
      call fzsc_info_copy (this%entire_fzsc, entire_fzsc)
      call fzsc_info_copy (this%main_fzsc, main_fzsc)
      call fzsc_info_copy (this%stop_fzsc, stop_fzsc)
      
      return
      end subroutine rawsdist_holder_fill
      
      
      
      
c     rsdh_recalc_stats_and_write - Recalculate and write statistics.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
      
      subroutine rsdh_recalc_stats_and_write (
     &  this, iunit)
      
      use sstat_include
      implicit none
      
      type(rawsdist_holder), pointer :: this
      integer, intent(in) :: iunit
      
      ! Calculate statistics and write
      
      call sstat_calc_sd_and_write_2 (
     &  iunit, this%sd, this%cell_count_d, this%cell_count_s,
     &  this%entire_fzsc, this%main_fzsc, this%stop_fzsc)
      
      return
      end subroutine rsdh_recalc_stats_and_write
      
      
      
      
c     rsdh_scale_linear - Apply linear scaling.
c     Parameters:
c      this = Pointer.
c      scale = Scale factor applied to each stress.
c      offset = Offset added to each scaled stress.
      
      subroutine rsdh_scale_linear (this, scale, offset)
      
      implicit none
      
      type(rawsdist_holder), pointer :: this
      real(8), intent(in) :: scale
      real(8), intent(in) :: offset
      
      ! Node position
      
      integer :: node_s
      integer :: node_d
      
      ! Apply the scaling
      
      do node_d = 0, this%cell_count_d
        do node_s = 0, this%cell_count_s
        
          this%sd(node_s, node_d) =
     &      (this%sd(node_s, node_d) * scale) + offset
       
        enddo
      enddo
      
      return
      end subroutine rsdh_scale_linear
      
      
      
      
c     rsdh_scale_min - Apply minimum scaling.
c     Parameters:
c      this = Pointer.
c      pval = Pointer to 2D value.
c     The 2D value pval is applied as a minimum.
      
      subroutine rsdh_scale_min (this, pval)
      
      use value2d_include
      implicit none
      
      type(rawsdist_holder), pointer :: this
      type(value2d), pointer :: pval
      
      ! Node position
      
      integer :: node_s
      integer :: node_d
      
      ! Node coordinates
      
      real(8) :: xs
      real(8) :: xd
      
      ! 2D value
      
      real(8) :: v
      
      ! Apply the scaling
      
      do node_d = 0, this%cell_count_d
      
        xd = dble(node_d) * this%cell_size_d
        
        do node_s = 0, this%cell_count_s
      
          xs = dble(node_s) * this%cell_size_s
          
          call value2d_eval (pval, v, xd, xs)
        
          this%sd(node_s, node_d) =
     &      max(this%sd(node_s, node_d), v)
       
        enddo
      enddo
      
      return
      end subroutine rsdh_scale_min
      
      
      
      
c     rsdh_scale_max - Apply maximum scaling.
c     Parameters:
c      this = Pointer.
c      pval = Pointer to 2D value.
c     The 2D value pval is applied as a maximum.
      
      subroutine rsdh_scale_max (this, pval)
      
      use value2d_include
      implicit none
      
      type(rawsdist_holder), pointer :: this
      type(value2d), pointer :: pval
      
      ! Node position
      
      integer :: node_s
      integer :: node_d
      
      ! Node coordinates
      
      real(8) :: xs
      real(8) :: xd
      
      ! 2D value
      
      real(8) :: v
      
      ! Apply the scaling
      
      do node_d = 0, this%cell_count_d
      
        xd = dble(node_d) * this%cell_size_d
        
        do node_s = 0, this%cell_count_s
      
          xs = dble(node_s) * this%cell_size_s
          
          call value2d_eval (pval, v, xd, xs)
        
          this%sd(node_s, node_d) =
     &      min(this%sd(node_s, node_d), v)
       
        enddo
      enddo
      
      return
      end subroutine rsdh_scale_max
      
      
      
      
      end module rawsdist_data








c----------------------------------------------------------------------
c
c     sdlist - List of stress distributions.




c     sdlist_data - Data module.

      module sdlist_data
      
      use rawsdist_data
      use fzsc_data
      implicit none
      
      ! Head and tail pointers for list
      
      type(rawsdist_holder), pointer, save :: sdlist_head
      type(rawsdist_holder), pointer, save :: sdlist_tail
      
      ! Number of entries on list
      
      integer, save :: sdlist_count
      
      ! Average statistics for entire fault
      
      type(fzsc_info), pointer, save :: sdlist_entire_fzsc
      
      ! Average statistics for main fault
      
      type(fzsc_info), pointer, save :: sdlist_main_fzsc
      
      ! Average statistics for stopping zone
      
      type(fzsc_info), pointer, save :: sdlist_stop_fzsc
      
      end module sdlist_data




c     sdlist_read_param - Read parameters for stress distribution list.
c     Parameters:
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       Realization number lo, hi
c       Filename base, postfix

      subroutine sdlist_read_param (iunit, f_ok)
      
      use io_unit_data
      use sdlist_data
      implicit none
      
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      ! Realization number range
      
      integer :: realiz_lo
      integer :: realiz_hi
      
      ! Base and postfix filenames
      
      character(max_path) :: basename
      character(max_path) :: postname
      
      ! Allocated fixed structures
      
      call fzsc_info_new (sdlist_entire_fzsc)
      call fzsc_info_new (sdlist_main_fzsc)
      call fzsc_info_new (sdlist_stop_fzsc)
      
      ! Read realization number range
        
      read(iunit,*) realiz_lo, realiz_hi
      
      ! Read base and postfix filenames
      
      read(iunit,*) basename, postname
      
      ! Read the stress distributions
      
      call sdlist_read_list (
     &       realiz_lo, realiz_hi, basename, postname, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Calculate average statistics
      
      call sdlist_calc_average_stat ()
      
      return
      end subroutine sdlist_read_param




c     sdlist_read_list - Read list of stress distributions.
c     Parameters:
c      realiz_lo = Realization number, lo.
c      realiz_hi = Realization number, hi.
c      basename = Base filename, used as a prefix.
c      postname = Postfix filename to use.
c      f_ok = Variable receives true if success, false if error.

      subroutine sdlist_read_list (
     &  realiz_lo, realiz_hi, basename, postname, f_ok)
      
      use io_unit_data
      use sdlist_data
      implicit none
      
      integer, intent(in) :: realiz_lo
      integer, intent(in) :: realiz_hi
      character(*), intent(in)  :: basename
      character(*), intent(in) :: postname
      logical, intent(out) :: f_ok
      
      ! Current realization and sub-realization
      
      integer :: realiz
      integer :: sub_realiz
      
      ! Filename construction
      
      character(max_path) :: filename
      logical :: f_exists
      
      ! File unit
      
      integer :: iunit
      
      ! New stress distribution being read
      
      type(rawsdist_holder), pointer :: p_rawsd
      
      ! Initialize to empty list
      
      sdlist_head => null()
      sdlist_tail => null()
      sdlist_count = 0
      
      ! Loop over realizations
      
      do realiz = realiz_lo, realiz_hi
      
        ! Loop over possible sub-realizations
        
        do sub_realiz = 0, 999
        
          ! Make filename for sub-realization
          
          call make_realiz_filename (
     &         filename, basename, realiz, sub_realiz, postname)
        
          ! Check if this file exists, go to next realization if not
          
          inquire (file=filename, exist=f_exists)
          
          if (.not. f_exists) then
            exit
          endif
          
          ! Open file for reading
          
          call open_unit_r (iunit, filename)
          
          ! Allocate a new stress distribution holder, and add to list
          
          call rawsdist_holder_new (p_rawsd)
          
          if (sdlist_count == 0) then
            sdlist_head => p_rawsd
          else
            sdlist_tail%next => p_rawsd
          endif
          
          sdlist_tail => p_rawsd
          p_rawsd%next => null()
          sdlist_count = sdlist_count + 1
          
          ! Read the stress distribution from the file
          
          call rawsdist_holder_read (p_rawsd, iunit)
          
          ! Close file
          
          call close_unit (iunit)
        
        enddo  ! end loop over possible sub-realizations
      
      enddo  ! end loop over realizations
      
      ! If found no stress distributions, error
      
      if (sdlist_count == 0) then
        f_ok = .false.
        return
      endif
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine sdlist_read_list




c     sdlist_write_list - Write list of stress distributions.
c     Parameters:
c      basename = Base filename, used as a prefix.
c      postname = Postfix filename to use.

      subroutine sdlist_write_list (basename, postname)
      
      use io_unit_data
      use sdlist_data
      implicit none
      
      character(*), intent(in)  :: basename
      character(*), intent(in) :: postname
      
      ! Current realization and sub-realization
      
      integer :: realiz
      integer :: sub_realiz
      
      ! Filename construction
      
      character(max_path) :: filename
      
      ! File unit
      
      integer :: iunit
      
      ! List pointer
      
      type(rawsdist_holder), pointer :: p_rawsd
      
      ! Loop over list elements
      
      p_rawsd => sdlist_head
      
      do while (associated(p_rawsd))
        
        ! Make filename for sub-realization
        
        realiz = p_rawsd%realiz
        sub_realiz = p_rawsd%sub_realiz
        
        call make_realiz_filename (
     &       filename, basename, realiz, sub_realiz, postname)
        
        ! Open file for writing
        
        call open_unit_w (iunit, filename)
        
        ! Write the stress distribution to the file
        
        call rawsdist_holder_write (p_rawsd, iunit)
        
        ! Close file
        
        call close_unit (iunit)
        
        ! Next element in list
        
        p_rawsd => p_rawsd%next
        
      enddo  ! end loop over list elements
      
      return
      end subroutine sdlist_write_list




c     sdlist_calc_average_stat - Calculate average statistics for list.
c     Parameters:
c      realiz_lo = Realization number, lo.
c      realiz_hi = Realization number, hi.
c      basename = Base filename, used as a prefix.
c      postname = Postfix filename to use.
c      f_ok = Variable receives true if success, false if error.

      subroutine sdlist_calc_average_stat ()
      
      use sdlist_data
      implicit none
      
      ! List pointer
      
      type(rawsdist_holder), pointer :: p_rawsd
      
      ! Initialize accumulators to zero
      
      call fzsc_info_clear (sdlist_entire_fzsc)
      call fzsc_info_clear (sdlist_main_fzsc)
      call fzsc_info_clear (sdlist_stop_fzsc)
      
      ! Loop over list elements
      
      p_rawsd => sdlist_head
      
      do while (associated(p_rawsd))
        
        ! Accumulate statistics for this element
        
        call fzsc_info_add (sdlist_entire_fzsc, p_rawsd%entire_fzsc)
        call fzsc_info_add (sdlist_main_fzsc, p_rawsd%main_fzsc)
        call fzsc_info_add (sdlist_stop_fzsc, p_rawsd%stop_fzsc)
      
        ! Next element in list
        
        p_rawsd => p_rawsd%next
        
      enddo  ! end loop over list elements
      
      ! Finish computing averages
        
      call fzsc_info_scale(sdlist_entire_fzsc, 1.0d0/dble(sdlist_count))
      call fzsc_info_scale (sdlist_main_fzsc, 1.0d0/dble(sdlist_count))
      call fzsc_info_scale (sdlist_stop_fzsc, 1.0d0/dble(sdlist_count))
      
      return
      end subroutine sdlist_calc_average_stat








c----------------------------------------------------------------------
c
c     sifmt - Stress input file format.




c     sifmt_data - Data module.

      module sifmt_data
      implicit none
      
      
      
      ! Record 1: Realization
      
      type :: sifmt_realiz
        
        ! Realization and sub-realization
        
        integer :: realiz
        integer :: sub_realiz
      
      end type sifmt_realiz
      
      
      
      ! Record 2: Fault size
      
      type :: sifmt_faultsize
        
        ! Number of cells along strike, dip
        
        integer :: cell_count_s
        integer :: cell_count_d
        
        ! Size of entire fault in meters along strike, dip
        
        real(8) :: fault_size_s
        real(8) :: fault_size_d
      
      end type sifmt_faultsize
      
      
      
      ! Record 3: Hypocenter
      
      type :: sifmt_hypo
      
        ! Hypocenter location along strike, dip
        
        integer :: hypo_s
        integer :: hypo_d
      
        ! Hypocenter coordinates in meters along strike, dip
        
        real(8) :: coord_hypo_s
        real(8) :: coord_hypo_d
      
        ! Nucleation zone radius in meters along strike, dip
        
        real(8) :: hypo_radius_s
        real(8) :: hypo_radius_d
      
      end type sifmt_hypo
      
      
      
      ! Record 4: Friction
      
      type :: sifmt_friction
      
        ! Node location along strike, dip
        
        integer :: node_s
        integer :: node_d
      
        ! Node coordinates in meters along strike, dip
        
        real(8) :: coord_node_s
        real(8) :: coord_node_d
      
        ! Normal stress in Pascal
        
        real(8) :: normal_stress
      
        ! Shear stress in Pascal along strike, dip
        ! shear_stress_s is positive for right lateral slip
        ! shear_stress_d is positive for downward slip
        
        real(8) :: shear_stress_s
        real(8) :: shear_stress_d
      
        ! Ratio of shear/normal stress along strike, dip
        
        real(8) :: stress_ratio_s
        real(8) :: stress_ratio_d
      
        ! Static coefficient of friction
        
        real(8) :: sta_coeff
      
        ! Dynamic coefficient of friction
        
        real(8) :: dyn_coeff
      
        ! Slip-weakening critical distance in meters
        
        real(8) :: sw_crit
      
        ! Cohesion in Pascal
        
        real(8) :: cohesion
      
        ! Time of forced rupture in seconds, or 1.0d9 if none
        
        real(8) :: rupture_time
      
      end type sifmt_friction
      
      ! Value to use for no forced rupture time
      
      real(8), parameter :: sifmt_no_rupture_time = 1.0d9
      
      contains
      


      
c     sifmt_realiz_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
      
      subroutine sifmt_realiz_new (this)
      implicit none
      
      type(sifmt_realiz), pointer :: this
      
      ! Allocate the object
      
      allocate(this)
      
      return
      end subroutine sifmt_realiz_new
      
      
      
      
c     sifmt_realiz_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine sifmt_realiz_delete (this)
      implicit none
      
      type(sifmt_realiz), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine sifmt_realiz_delete
      
      
      
      
c     sifmt_realiz_write - Write to a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
      
      subroutine sifmt_realiz_write (this, iunit)
      implicit none
      
      type(sifmt_realiz), pointer :: this
      integer, intent(in) :: iunit
      
      write(iunit,'(2(1x,I9))')
     &  this%realiz,
     &  this%sub_realiz
      
      return
      end subroutine sifmt_realiz_write
      
      
      
      
c     sifmt_realiz_read - Read from a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
      
      subroutine sifmt_realiz_read (this, iunit)
      implicit none
      
      type(sifmt_realiz), pointer :: this
      integer, intent(in) :: iunit
      
      read(iunit,*)
     &  this%realiz,
     &  this%sub_realiz
      
      return
      end subroutine sifmt_realiz_read
      


      
c     sifmt_faultsize_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
      
      subroutine sifmt_faultsize_new (this)
      implicit none
      
      type(sifmt_faultsize), pointer :: this
      
      ! Allocate the object
      
      allocate(this)
      
      return
      end subroutine sifmt_faultsize_new
      
      
      
      
c     sifmt_faultsize_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine sifmt_faultsize_delete (this)
      implicit none
      
      type(sifmt_faultsize), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine sifmt_faultsize_delete
      
      
      
      
c     sifmt_faultsize_write - Write to a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
      
      subroutine sifmt_faultsize_write (this, iunit)
      implicit none
      
      type(sifmt_faultsize), pointer :: this
      integer, intent(in) :: iunit
      
      write(iunit,'(2(1x,I9),2(1x,1pe15.6))')
     &  this%cell_count_s,
     &  this%cell_count_d,
     &  this%fault_size_s,
     &  this%fault_size_d
      
      return
      end subroutine sifmt_faultsize_write
      
      
      
      
c     sifmt_faultsize_read - Read from a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
      
      subroutine sifmt_faultsize_read (this, iunit)
      implicit none
      
      type(sifmt_faultsize), pointer :: this
      integer, intent(in) :: iunit
      
      read(iunit,*)
     &  this%cell_count_s,
     &  this%cell_count_d,
     &  this%fault_size_s,
     &  this%fault_size_d
      
      return
      end subroutine sifmt_faultsize_read
      


      
c     sifmt_hypo_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
      
      subroutine sifmt_hypo_new (this)
      implicit none
      
      type(sifmt_hypo), pointer :: this
      
      ! Allocate the object
      
      allocate(this)
      
      return
      end subroutine sifmt_hypo_new
      
      
      
      
c     sifmt_hypo_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine sifmt_hypo_delete (this)
      implicit none
      
      type(sifmt_hypo), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine sifmt_hypo_delete
      
      
      
      
c     sifmt_hypo_write - Write to a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
      
      subroutine sifmt_hypo_write (this, iunit)
      implicit none
      
      type(sifmt_hypo), pointer :: this
      integer, intent(in) :: iunit
      
      write(iunit,'(2(1x,I9),4(1x,1pe15.6))')
     &  this%hypo_s,
     &  this%hypo_d,
     &  this%coord_hypo_s,
     &  this%coord_hypo_d,
     &  this%hypo_radius_s,
     &  this%hypo_radius_d
      
      return
      end subroutine sifmt_hypo_write
      
      
      
      
c     sifmt_hypo_read - Read from a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
      
      subroutine sifmt_hypo_read (this, iunit)
      implicit none
      
      type(sifmt_hypo), pointer :: this
      integer, intent(in) :: iunit
      
      read(iunit,*)
     &  this%hypo_s,
     &  this%hypo_d,
     &  this%coord_hypo_s,
     &  this%coord_hypo_d,
     &  this%hypo_radius_s,
     &  this%hypo_radius_d
      
      return
      end subroutine sifmt_hypo_read
      


      
c     sifmt_friction_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
      
      subroutine sifmt_friction_new (this)
      implicit none
      
      type(sifmt_friction), pointer :: this
      
      ! Allocate the object
      
      allocate(this)
      
      return
      end subroutine sifmt_friction_new
      
      
      
      
c     sifmt_friction_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine sifmt_friction_delete (this)
      implicit none
      
      type(sifmt_friction), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine sifmt_friction_delete
      
      
      
      
c     sifmt_friction_write - Write to a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
      
      subroutine sifmt_friction_write (this, iunit)
      implicit none
      
      type(sifmt_friction), pointer :: this
      integer, intent(in) :: iunit
      
      write(iunit,'(2(1x,I9),12(1x,1pe15.6))')
     &  this%node_s, this%node_d,
     &  this%coord_node_s, this%coord_node_d,
     &  this%normal_stress,
     &  this%shear_stress_s, this%shear_stress_d,
     &  this%stress_ratio_s, this%stress_ratio_d,
     &  this%sta_coeff, this%dyn_coeff,
     &  this%sw_crit,
     &  this%cohesion,
     &  this%rupture_time
      
      return
      end subroutine sifmt_friction_write
      
      
      
      
c     sifmt_friction_read - Read from a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
      
      subroutine sifmt_friction_read (this, iunit)
      implicit none
      
      type(sifmt_friction), pointer :: this
      integer, intent(in) :: iunit
      
      read(iunit,*)
     &  this%node_s, this%node_d,
     &  this%coord_node_s, this%coord_node_d,
     &  this%normal_stress,
     &  this%shear_stress_s, this%shear_stress_d,
     &  this%stress_ratio_s, this%stress_ratio_d,
     &  this%sta_coeff, this%dyn_coeff,
     &  this%sw_crit,
     &  this%cohesion,
     &  this%rupture_time
      
      return
      end subroutine sifmt_friction_read
      
      
      
      
      end module sifmt_data








c----------------------------------------------------------------------
c
c     sinf - Stress input file builder.




c     sinf_data - Data module.

      module sinf_data
      
      use value2d_include
      use fzsc_data
      implicit none
      
      
      !===== Pre-scaling of stress
      
      
      ! Structure to hold weights for scaling stress
      
      type :: sinf_weights
        
        ! Constant term
        
        real(8) :: constant
        
        ! Weights for average statistics: entire, main, stopping
      
        type(fzsc_info), pointer :: avg_entire_fzsc
        type(fzsc_info), pointer :: avg_main_fzsc
        type(fzsc_info), pointer :: avg_stop_fzsc
        
        ! Weights for individual statistics: entire, main, stopping
      
        type(fzsc_info), pointer :: indv_entire_fzsc
        type(fzsc_info), pointer :: indv_main_fzsc
        type(fzsc_info), pointer :: indv_stop_fzsc
      
      end type sinf_weights
      
      
      ! Target value of scale
      
      real(8), save :: sinf_target_scale
      
      ! Target value of level
      
      real(8), save :: sinf_target_level
      
      ! Weights for stress scale
      
      type(sinf_weights), pointer, save :: sinf_weight_scale
      
      ! Weights for stress level
      
      type(sinf_weights), pointer, save :: sinf_weight_level
      
      ! Minimum pre-scaled stress on the fault
      
      type(value2d), pointer, save :: sinf_presc_min
      
      ! Maximum pre-scaled stress on the fault
      
      type(value2d), pointer, save :: sinf_presc_max
      
      
      !===== Hypocenter and nucleation
      
      
      ! Nucleation method
      
      integer, save :: sinf_nuke_method
      
      ! Time-drop method
      
      integer, parameter :: sinf_nuke_method_time_drop = 1
      
      ! Overstress method
      
      integer, parameter :: sinf_nuke_method_overstress = 2
      
      ! No nucleation in output file
      
      integer, parameter :: sinf_nuke_method_none = 3
      
      ! Nucleation overstress parameter
      
      real(8), save :: sinf_nuke_mu
      
      ! Velocity of forced rupture, meter/second, along strike, dip
      
      real(8), save :: sinf_nuke_vel_s
      real(8), save :: sinf_nuke_vel_d
      
      ! Formula for hypocenter radius
      
      integer, save :: sinf_hrad_form
      
      ! Radius equals nucleation factor F
      ! R = F
      
      integer, parameter :: sinf_hrad_form_fixed = 1
      
      ! Radius given by the Day formula muliplied by F
      ! R = F * (7*pi/24) * lame_mu * (sta_coeff - dyn_coeff) * sw_crit
      !      / ( (avg_presc - dyn_coeff)^2 * normal_stress )
      ! where avg_presc is the average prescaled stress
      
      integer, parameter :: sinf_hrad_form_day = 2
      
      ! Nucleation factor, for computing radius
      
      real(8), save :: sinf_nuke_factor
      
      ! Nucleation zone aspect ratio, hypo_radius_s/hypo_radius_d
      
      real(8), save :: sinf_nuke_aspect
      
      ! Shear modulus, used in computation of hypocenter radius
      
      real(8), save :: sinf_nuke_lame_mu
      
      ! Radius at which to begin taper
      ! 0.0 = hypocenter, 1.0 = outer edge of nuke zone.
      
      real(8), save :: sinf_nuke_taper_radius
      
      ! Radius at which to begin, end taper of critical slip distance
      ! 0.0 = hypocenter, 1.0 = outer edge of nuke zone.
      ! Values larger than 1.0 are allowed and extend taper beyond the
      !  nuke zone.  Use -1.0 if no taper is desired.
      
      real(8), save :: sinf_dc_taper_radius_lo
      real(8), save :: sinf_dc_taper_radius_hi
      
      ! Level for taper of critical slip distance
      ! The level is a multipler applied to sinf_sw_crit.
      ! Use 1.0 if no taper is desired. 
      
      real(8), save :: sinf_dc_taper_level
      
      ! Hypocenter selection mask radius along strike, dip
      ! In units of cells, 0 selectes a single point
      
      integer, save :: sinf_hypo_mrad_s
      integer, save :: sinf_hypo_mrad_d
      
      ! Hypocenter selection mask standard deviation along strike, dip
      ! In units of cells
      
      real(8), save :: sinf_hypo_msdev_s
      real(8), save :: sinf_hypo_msdev_d
      
      ! Hypoocenter selection mask
      ! sinf_hypo_mask(-sinf_hypo_mrad_s:sinf_hypo_mrad_s,
      !                -sinf_hypo_mrad_d:sinf_hypo_mrad_d)
      
      real(8), allocatable, save :: sinf_hypo_mask(:,:)
      
      ! Limits of hypocenter location along strike, dip
      
      integer, save :: sinf_hypo_s_lo
      integer, save :: sinf_hypo_s_hi
      
      integer, save :: sinf_hypo_d_lo
      integer, save :: sinf_hypo_d_hi
      
      ! Limits of nucleation zone along strike, dip
      ! Note:  The hypocenter selection mask, when positioned at any
      !  point within the hypocenter limits, must lie within the
      !  nucleation zone limits.
      
      integer, save :: sinf_nuke_s_lo
      integer, save :: sinf_nuke_s_hi
      
      integer, save :: sinf_nuke_d_lo
      integer, save :: sinf_nuke_d_hi
      
      ! Limits for hypocenter stress averaging along strike, dip
      
      integer, save :: sinf_stravg_s_lo
      integer, save :: sinf_stravg_s_hi
      
      integer, save :: sinf_stravg_d_lo
      integer, save :: sinf_stravg_d_hi
      
      
      !===== Friction parameters
      
      
      ! Stress multipliers along strike, dip
      ! This allows selection of right/left lateral, down/up normal
      
      real(8), save :: sinf_smult_s
      real(8), save :: sinf_smult_d
      
      ! Value of normal stress on the fault
      
      type(value2d), pointer, save :: sinf_normal_stress
      
      ! Value of static coefficient of friction on the fault
      
      type(value2d), pointer, save :: sinf_sta_coeff
      
      ! Value of dynamic coefficient of friction on the fault
      
      type(value2d), pointer, save :: sinf_dyn_coeff
      
      ! Value of critical slip distance on the fault
      
      type(value2d), pointer, save :: sinf_sw_crit
      
      ! Value of cohesion on the fault
      
      type(value2d), pointer, save :: sinf_cohesion
      
      ! Value of depth conditioning on the fault
      
      type(value2d), pointer, save :: sinf_depth_cond
      
      ! Value of min strength excess coefficient on the fault
      ! The static coefficient of friction is increased if necessary so
      !  it is at least as large as the pre-scaled shear stress plus
      !  min_excess.
      
      type(value2d), pointer, save :: sinf_min_excess


      contains
      


      
c     sinf_weights_new - Constructor.
c     Parameters:
c      this = Variable to receive pointer.
      
      subroutine sinf_weights_new (this)
      implicit none
      
      type(sinf_weights), pointer :: this
      
      ! Allocate the object
      
      allocate(this)
      this%avg_entire_fzsc => null()
      this%avg_main_fzsc => null()
      this%avg_stop_fzsc => null()
      this%indv_entire_fzsc => null()
      this%indv_main_fzsc => null()
      this%indv_stop_fzsc => null()
      
      call fzsc_info_new (this%avg_entire_fzsc)
      call fzsc_info_new (this%avg_main_fzsc)
      call fzsc_info_new (this%avg_stop_fzsc)
      call fzsc_info_new (this%indv_entire_fzsc)
      call fzsc_info_new (this%indv_main_fzsc)
      call fzsc_info_new (this%indv_stop_fzsc)
      
      return
      end subroutine sinf_weights_new
      
      
      
      
c     sinf_weights_delete - Destructor.
c     Parameters:
c      this = Pointer.
      
      subroutine sinf_weights_delete (this)
      implicit none
      
      type(sinf_weights), pointer :: this
      
      ! If non-null ...
      
      if (associated(this)) then
      
        ! Deallocate our held objects
        
        if (associated(this%indv_stop_fzsc)) then
          call fzsc_info_delete (this%indv_stop_fzsc)
        endif
        
        if (associated(this%indv_main_fzsc)) then
          call fzsc_info_delete (this%indv_main_fzsc)
        endif
        
        if (associated(this%indv_entire_fzsc)) then
          call fzsc_info_delete (this%indv_entire_fzsc)
        endif
        
        if (associated(this%avg_stop_fzsc)) then
          call fzsc_info_delete (this%avg_stop_fzsc)
        endif
        
        if (associated(this%avg_main_fzsc)) then
          call fzsc_info_delete (this%avg_main_fzsc)
        endif
        
        if (associated(this%avg_entire_fzsc)) then
          call fzsc_info_delete (this%avg_entire_fzsc)
        endif
        
        ! Deallocate our object
        
        deallocate(this)
        this => null()
      
      endif
      
      return
      end subroutine sinf_weights_delete
      
      
      
      
c     sinf_weights_read - Read from a file.
c     Parameters:
c      this = Pointer.
c      iunit = Fortran unit number.
c
c     File format:
c      constant
c      avg_entire_fzsc  [see fzsc_info_read]
c      avg_main_fzsc  [see fzsc_info_read]
c      avg_stop_fzsc  [see fzsc_info_read]
c      indv_entire_fzsc  [see fzsc_info_read]
c      indv_main_fzsc  [see fzsc_info_read]
c      indv_stop_fzsc  [see fzsc_info_read]
      
      subroutine sinf_weights_read (this, iunit)
      implicit none
      
      type(sinf_weights), pointer :: this
      integer, intent(in) :: iunit
      
      ! Read constant term
      
      read(iunit,*)
     &  this%constant
      
      ! Read weights for statistics
      
      call fzsc_info_read (this%avg_entire_fzsc, iunit)
      call fzsc_info_read (this%avg_main_fzsc, iunit)
      call fzsc_info_read (this%avg_stop_fzsc, iunit)
      
      call fzsc_info_read (this%indv_entire_fzsc, iunit)
      call fzsc_info_read (this%indv_main_fzsc, iunit)
      call fzsc_info_read (this%indv_stop_fzsc, iunit)
      
      return
      end subroutine sinf_weights_read
      
      
      
      
c     sinf_weights_apply - Apply weights to a stress distribution.
c     Parameters:
c      this = Pointer.
c      rawsd = Pointer to raw stress distribution.
c      vdot = Variable to receive dot product.
      
      subroutine sinf_weights_apply (this, rawsd, vdot)
      
      use sdlist_data
      implicit none
      
      type(sinf_weights), pointer :: this
      type(rawsdist_holder), pointer :: rawsd
      real(8), intent(out) :: vdot
      
      ! Compute dot product
      
      vdot = this%constant
     &       + fzsc_info_dot (this%avg_entire_fzsc, sdlist_entire_fzsc)
     &       + fzsc_info_dot (this%avg_main_fzsc, sdlist_main_fzsc)
     &       + fzsc_info_dot (this%avg_stop_fzsc, sdlist_stop_fzsc)
     &       + fzsc_info_dot (this%indv_entire_fzsc, rawsd%entire_fzsc)
     &       + fzsc_info_dot (this%indv_main_fzsc, rawsd%main_fzsc)
     &       + fzsc_info_dot (this%indv_stop_fzsc, rawsd%stop_fzsc)
      
      return
      end subroutine sinf_weights_apply
      
      
      
      
c     sinf_make_hypo_mask - Make the hypocenter selection mask.
      
      subroutine sinf_make_hypo_mask ()
      implicit none
      
      ! Index into mask
      
      integer :: mask_s
      integer :: mask_d
      
      ! Radius normalized by standard deviation along strike, dip
      
      real(8) :: rs
      real(8) :: rd
      
      ! Normalization factor
      
      real(8) :: norm
      
      ! Allocate the mask
      
      allocate (sinf_hypo_mask(-sinf_hypo_mrad_s:sinf_hypo_mrad_s,
     &                         -sinf_hypo_mrad_d:sinf_hypo_mrad_d))
      
      ! Construct mask as a gaussian
      
      norm = 0.0d0
      
      do mask_d = -sinf_hypo_mrad_d, sinf_hypo_mrad_d
        do mask_s = -sinf_hypo_mrad_s, sinf_hypo_mrad_s
          rs = dble(mask_s)/sinf_hypo_msdev_s
          rd = dble(mask_d)/sinf_hypo_msdev_d
          sinf_hypo_mask(mask_s, mask_d) = exp(-0.5d0*(rs*rs + rd*rd))
          norm = norm + sinf_hypo_mask(mask_s, mask_d)
        enddo
      enddo
      
      do mask_d = -sinf_hypo_mrad_d, sinf_hypo_mrad_d
        do mask_s = -sinf_hypo_mrad_s, sinf_hypo_mrad_s
          sinf_hypo_mask(mask_s, mask_d)
     &      = sinf_hypo_mask(mask_s, mask_d) / norm
        enddo
      enddo
      
      return
      end subroutine sinf_make_hypo_mask
      
      
      
      
c     sinf_find_hypo - Find hypocenter for a stress distribution.
c     Parameters:
c      rawsd = Pointer to raw stress distribution.
c      f_ok = Receives true if success, false if cannot find hypocenter.
c     This function fills in the hypocenter fields of rawsd.
c     Note: Stress must be pre-scaled before calling this function.
      
      subroutine sinf_find_hypo (rawsd, f_ok)
      
      use rawsdist_data
      implicit none
      
      type(rawsdist_holder), pointer :: rawsd
      logical, intent(out) :: f_ok
      
      ! Index into mask
      
      integer :: mask_s
      integer :: mask_d
      
      ! Index into fault
      
      integer :: node_s
      integer :: node_d
      
      ! Coordinates in meters
      
      real(8) :: xs
      real(8) :: xd
      
      ! Accumulators
      
      real(8) :: acc
      real(8) :: acc_row
      
      integer :: acc_cnt
      
      ! Average prescaled stress
      
      real(8) :: avg_presc
      
      ! Flag is true if we found a possible hypocenter
      
      logical :: f_found
      
      ! Current and maximum filtered stress
      
      real(8) :: cur_filt_stress
      real(8) :: max_filt_stress
      
      ! Current hypocenter radius
      
      real(8) :: hypo_radius_s
      real(8) :: hypo_radius_d
      
      real(8) :: hypo_radius_avg
      
      ! Nucleation zone lower and upper limits, in meters
      
      real(8) :: x_nuke_s_lo
      real(8) :: x_nuke_s_hi
      
      real(8) :: x_nuke_d_lo
      real(8) :: x_nuke_d_hi
      
      ! Friction parameters
      
      real(8) :: normal_stress
      real(8) :: sta_coeff
      real(8) :: dyn_coeff
      real(8) :: sw_crit
      real(8) :: cohesion
      
      ! If no nucleation, set fields for no hypocenter
      
      if (sinf_nuke_method == sinf_nuke_method_none) then
      
        ! No hypocenter location
      
        rawsd%hypo_s = -1
        rawsd%hypo_d = -1
        
        ! Use half-cell minimum radius, so division by radius is OK
      
        rawsd%hypo_radius_s = rawsd%cell_size_s * 0.5d0
        rawsd%hypo_radius_d = rawsd%cell_size_d * 0.5d0
      
        ! Success
      
        f_ok = .true.
      
        return
      
      endif
      
      ! Compute nucleation zone limits, we allow a half-cell extra
      
      x_nuke_s_lo = rawsd%cell_size_s * (dble(sinf_nuke_s_lo) - 0.5d0)
      x_nuke_s_hi = rawsd%cell_size_s * (dble(sinf_nuke_s_hi) + 0.5d0)
      
      x_nuke_d_lo = rawsd%cell_size_d * (dble(sinf_nuke_d_lo) - 0.5d0)
      x_nuke_d_hi = rawsd%cell_size_d * (dble(sinf_nuke_d_hi) + 0.5d0)
      
      ! Compute the average prescaled stress
      
      acc = 0.0d0
      acc_cnt = 0
      
      do node_d = sinf_stravg_d_lo, sinf_stravg_d_hi
      
        acc_row = 0.0d0
        
        do node_s = sinf_stravg_s_lo, sinf_stravg_s_hi
        
          acc_row = acc_row + rawsd%sd(node_s, node_d)
          acc_cnt = acc_cnt + 1
          
        enddo
        
        acc = acc + acc_row
        
      enddo
      
      if (acc_cnt == 0) then
        f_ok = .false.
        return
      endif
      
      avg_presc = acc / dble(acc_cnt)
      
      ! Search for hypocenter
      
      f_found = .false.
      
      max_filt_stress = 0.0d0
      
      do node_d = sinf_hypo_d_lo, sinf_hypo_d_hi
        do node_s = sinf_hypo_s_lo, sinf_hypo_s_hi
          
          ! If selection mask does not fit, skip it
          
          if (     node_d - sinf_hypo_mrad_d < sinf_nuke_d_lo
     &        .or. node_d + sinf_hypo_mrad_d > sinf_nuke_d_hi
     &        .or. node_s - sinf_hypo_mrad_s < sinf_nuke_s_lo
     &        .or. node_s + sinf_hypo_mrad_s > sinf_nuke_s_hi) then
            cycle
          endif
          
          ! Apply the selection mask at this point
          
          cur_filt_stress = 0.0d0
          
          do mask_d = -sinf_hypo_mrad_d, sinf_hypo_mrad_d
            do mask_s = -sinf_hypo_mrad_s, sinf_hypo_mrad_s
            
              cur_filt_stress = cur_filt_stress
     &            + ( sinf_hypo_mask(mask_s, mask_d)
     &                * rawsd%sd(node_s + mask_s, node_d + mask_d) )
              
            enddo
          enddo
          
          ! If not a new maximum, skip it
          
          if (f_found) then
            if (cur_filt_stress <= max_filt_stress) then
              cycle
            endif
          endif
          
          ! Get friction parameters
          
          xs = rawsd%cell_size_s * dble(node_s)
          xd = rawsd%cell_size_d * dble(node_d)
          
          call value2d_eval (sinf_normal_stress, normal_stress, xd, xs)
          call value2d_eval (sinf_sta_coeff, sta_coeff, xd, xs)
          call value2d_eval (sinf_dyn_coeff, dyn_coeff, xd, xs)
          call value2d_eval (sinf_sw_crit, sw_crit, xd, xs)
          call value2d_eval (sinf_cohesion, cohesion, xd, xs)
          
          ! Compute radius using selected formula
          
          select case (sinf_hrad_form)
          
          ! Fixed radius
          
          case (sinf_hrad_form_fixed)
          
            hypo_radius_avg = sinf_nuke_factor
          
          ! Day formula
          
          case (sinf_hrad_form_day)
            
            if (avg_presc <= dyn_coeff) then
              cycle
            endif
          
            hypo_radius_avg
     &        = sinf_nuke_factor * 0.9162978572970230d0
     &        * sinf_nuke_lame_mu * (sta_coeff - dyn_coeff) * sw_crit
     &        / ( (avg_presc - dyn_coeff)*(avg_presc - dyn_coeff)
     &            * normal_stress )
          
          end select
          
          ! Apply aspect ratio, and enforce half-cell minimum radius
          
          hypo_radius_s = hypo_radius_avg * sqrt(sinf_nuke_aspect)
          hypo_radius_d = hypo_radius_avg / sqrt(sinf_nuke_aspect)
          
          hypo_radius_s = max(hypo_radius_s, rawsd%cell_size_s * 0.5d0)
          hypo_radius_d = max(hypo_radius_d, rawsd%cell_size_d * 0.5d0)
          
          ! If not within nucleation limits, skip it
          
          if (     xd - hypo_radius_d < x_nuke_d_lo
     &        .or. xd + hypo_radius_d > x_nuke_d_hi
     &        .or. xs - hypo_radius_s < x_nuke_s_lo
     &        .or. xs + hypo_radius_s > x_nuke_s_hi) then
            cycle
          endif
          
          ! We found a possible hypocenter location
          
          f_found = .true.
          
          max_filt_stress = cur_filt_stress
          
          rawsd%hypo_s = node_s
          rawsd%hypo_d = node_d
          
          rawsd%hypo_radius_s = hypo_radius_s
          rawsd%hypo_radius_d = hypo_radius_d
        
        enddo
      enddo
      
      ! Error if we found no hypocenter location
      
      if (.not. f_found) then
        f_ok = .false.
        return
      endif
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine sinf_find_hypo
      
      
      
      
c     sinf_prescale_stress - Pre-scale a stress distribution.
c     Parameters:
c      rawsd = Pointer to raw stress distribution.
c      f_ok = Receives true if success, false if cannot pre-scale.
c      cnt_clip_min = Receives number of points clipped at minimum.
c      cnt_clip_max = Receives number of points clipped at maximum.
c     This function adjusts the values of the stress distribution.
      
      subroutine sinf_prescale_stress (
     &  rawsd, f_ok, cnt_clip_min, cnt_clip_max)
      
      use rawsdist_data
      implicit none
      
      type(rawsdist_holder), pointer :: rawsd
      logical, intent(out) :: f_ok
      integer, intent(out) :: cnt_clip_min
      integer, intent(out) :: cnt_clip_max
      
      ! Index into fault
      
      integer :: node_s
      integer :: node_d
      
      ! Coordinates in meters
      
      real(8) :: xs
      real(8) :: xd
      
      ! Scale and level of the raw distribution
      
      real(8) :: raw_scale
      real(8) :: raw_level
      
      ! Scale and offset to be applied for prescaling
      
      real(8) :: presc_scale
      real(8) :: presc_offset
      
      ! Minimum and maximum prescaled values
      
      real(8) :: presc_min
      real(8) :: presc_max
      
      ! Stress value
      
      real(8) :: v
      
      ! Initialize counts of clipped points
      
      cnt_clip_min = 0
      cnt_clip_max = 0
      
      ! Apply weights to get raw scale and level
      
      call sinf_weights_apply (sinf_weight_scale, rawsd, raw_scale)
      call sinf_weights_apply (sinf_weight_level, rawsd, raw_level)
      
      ! Scale and offset to transform to target values
      
      presc_scale = sinf_target_scale / raw_scale
      presc_offset = sinf_target_level - (raw_level * presc_scale)
      
      ! Loop over fault surface
      
      do node_d = 0, rawsd%cell_count_d
        do node_s = 0, rawsd%cell_count_s
          
          ! Get range limit
          
          xs = rawsd%cell_size_s * dble(node_s)
          xd = rawsd%cell_size_d * dble(node_d)
          
          call value2d_eval (sinf_presc_min, presc_min, xd, xs)
          call value2d_eval (sinf_presc_max, presc_max, xd, xs)
          
          ! Apply scaling and range limit
          
          v = (rawsd%sd(node_s, node_d) * presc_scale) + presc_offset
          
          if (v < presc_min) then
            v = presc_min
            cnt_clip_min = cnt_clip_min + 1
          endif
          
          if (v > presc_max) then
            v = presc_max
            cnt_clip_max = cnt_clip_max + 1
          endif
          
          rawsd%sd(node_s, node_d) = v
        
        enddo
      enddo
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine sinf_prescale_stress
      
      
      
      
c     sinf_fill_realiz - Fill a realization record.
c     Parameters:
c      rec_realiz = Pointer to record to fill.
c      rawsd = Pointer to raw stress distribution.
c      f_ok = Receives true if success, false if error.
      
      subroutine sinf_fill_realiz (rec_realiz, rawsd, f_ok)
      
      use rawsdist_data
      use sifmt_data
      implicit none
      
      type(sifmt_realiz), pointer :: rec_realiz
      type(rawsdist_holder), pointer :: rawsd
      logical, intent(out) :: f_ok
      
      ! Fill record
      
      rec_realiz%realiz = rawsd%realiz
      rec_realiz%sub_realiz = rawsd%sub_realiz
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine sinf_fill_realiz
      
      
      
      
c     sinf_fill_faultsize - Fill a fault size record.
c     Parameters:
c      rec_faultsize = Pointer to record to fill.
c      rawsd = Pointer to raw stress distribution.
c      f_ok = Receives true if success, false if error.
      
      subroutine sinf_fill_faultsize (rec_faultsize, rawsd, f_ok)
      
      use rawsdist_data
      use sifmt_data
      implicit none
      
      type(sifmt_faultsize), pointer :: rec_faultsize
      type(rawsdist_holder), pointer :: rawsd
      logical, intent(out) :: f_ok
      
      ! Fill record
      
      rec_faultsize%cell_count_s = rawsd%cell_count_s
      rec_faultsize%cell_count_d = rawsd%cell_count_d
      
      rec_faultsize%fault_size_s
     &  = dble(rawsd%cell_count_s) * rawsd%cell_size_s
      rec_faultsize%fault_size_d
     &  = dble(rawsd%cell_count_d) * rawsd%cell_size_d
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine sinf_fill_faultsize
      
      
      
      
c     sinf_fill_hypo - Fill a hypocenter record.
c     Parameters:
c      rec_hypo = Pointer to record to fill.
c      rawsd = Pointer to raw stress distribution.
c      f_ok = Receives true if success, false if error.
      
      subroutine sinf_fill_hypo (rec_hypo, rawsd, f_ok)
      
      use rawsdist_data
      use sifmt_data
      implicit none
      
      type(sifmt_hypo), pointer :: rec_hypo
      type(rawsdist_holder), pointer :: rawsd
      logical, intent(out) :: f_ok
      
      ! Fill record
      
      rec_hypo%hypo_s = rawsd%hypo_s
      rec_hypo%hypo_d = rawsd%hypo_d
      
      rec_hypo%coord_hypo_s
     &  = dble(rawsd%hypo_s) * rawsd%cell_size_s
      rec_hypo%coord_hypo_d
     &  = dble(rawsd%hypo_d) * rawsd%cell_size_d
      
      rec_hypo%hypo_radius_s = rawsd%hypo_radius_s
      rec_hypo%hypo_radius_d = rawsd%hypo_radius_d
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine sinf_fill_hypo
      
      
      
      
c     sinf_fill_friction - Fill a friction record.
c     Parameters:
c      rec_friction = Pointer to record to fill.
c      rawsd = Pointer to raw stress distribution.
c      node_s = Node position along-strike.
c      node_d = Node position along-dip.
c      f_ok = Receives true if success, false if error.
c      cnt_nuke_node = Variable incremented for a nucleation zone node.
c      cnt_sta_adjust = Variable incremented for static coeff adjustment
c                       (except for nucleation overstress).
      
      subroutine sinf_fill_friction (
     &  rec_friction, rawsd, node_s, node_d, f_ok,
     &  cnt_nuke_node, cnt_sta_adjust)
      
      use math_constants
      use rawsdist_data
      use sifmt_data
      implicit none
      
      type(sifmt_friction), pointer :: rec_friction
      type(rawsdist_holder), pointer :: rawsd
      integer, intent(in) :: node_s
      integer, intent(in) :: node_d
      logical, intent(out) :: f_ok
      integer, intent(inout) :: cnt_nuke_node
      integer, intent(inout) :: cnt_sta_adjust
      
      ! Coordinates in meters
      
      real(8) :: xs
      real(8) :: xd
      
      ! Friction parameters
      
      real(8) :: normal_stress
      real(8) :: sta_coeff
      real(8) :: dyn_coeff
      real(8) :: sw_crit
      real(8) :: cohesion
      real(8) :: depth_cond
      real(8) :: min_excess
      
      ! Prescaled shear stress
      
      real(8) :: presc_shear
      
      ! Rupture time 
      
      real(8) :: rupt_s
      real(8) :: rupt_d
      
      real(8) :: rupture_time
      
      ! Distances to hypocenter, and normalized distances
      
      real(8) :: r_hypo_s
      real(8) :: r_hypo_d
      
      real(8) :: nr_hypo_s
      real(8) :: nr_hypo_d
      real(8) :: nr_hypo
      
      ! Friction limit
      
      real(8) :: fric_limit
          
      ! Get friction parameters
      
      xs = rawsd%cell_size_s * dble(node_s)
      xd = rawsd%cell_size_d * dble(node_d)
      
      call value2d_eval (sinf_normal_stress, normal_stress, xd, xs)
      call value2d_eval (sinf_sta_coeff, sta_coeff, xd, xs)
      call value2d_eval (sinf_dyn_coeff, dyn_coeff, xd, xs)
      call value2d_eval (sinf_sw_crit, sw_crit, xd, xs)
      call value2d_eval (sinf_cohesion, cohesion, xd, xs)
      call value2d_eval (sinf_depth_cond, depth_cond, xd, xs)
      call value2d_eval (sinf_min_excess, min_excess, xd, xs)
        
      ! Pre-scaled shear stress
      
      presc_shear = rawsd%sd(node_s, node_d)
        
      ! Assume no rupture time
      
      rupture_time = sifmt_no_rupture_time
      
      ! Calculate distances to hypocenter
      
      r_hypo_s = (rawsd%cell_size_s * dble(rawsd%hypo_s)) - xs
      r_hypo_d = (rawsd%cell_size_d * dble(rawsd%hypo_d)) - xd
      
      nr_hypo_s = r_hypo_s / rawsd%hypo_radius_s
      nr_hypo_d = r_hypo_d / rawsd%hypo_radius_d
      
      nr_hypo = sqrt(nr_hypo_s*nr_hypo_s + nr_hypo_d*nr_hypo_d)
      
      ! If we're in the nucleation zone ...
      
      if (nr_hypo <= 1.0d0
     &    .and. sinf_nuke_method /= sinf_nuke_method_none) then
        
        ! Count a nucleation node
        
        cnt_nuke_node = cnt_nuke_node + 1
        
        ! Switch based on nucleation method ...
        
        select case (sinf_nuke_method)
        
        ! Time-drop method ...
        
        case (sinf_nuke_method_time_drop)
        
          ! If in taper zone ...
          ! (Note +0.0001 ensures sinf_nuke_taper_radius <= 0.9999
          !  and nr_hypo >= 0.0001)
          
          if (nr_hypo > sinf_nuke_taper_radius + 0.0001d0) then
        
            ! Rupture time based on distance to hypocenter,
            ! with half velocity within the taper zone
          
            rupt_s = r_hypo_s / sinf_nuke_vel_s
            rupt_d = r_hypo_d / sinf_nuke_vel_d
        
            rupture_time = sqrt(rupt_s*rupt_s + rupt_d*rupt_d)
     &        * (nr_hypo + nr_hypo - sinf_nuke_taper_radius) / nr_hypo
          
            ! Make sure the shear stress exceeds the dynamic friction
            ! (including cohesion) by at least the overstress parameter,
            ! but taper this effect
            
            fric_limit
     &        = dyn_coeff + sinf_nuke_mu + (cohesion/normal_stress)
          
            if (presc_shear < fric_limit) then
     
              presc_shear = lin_interp (
     &          nr_hypo, sinf_nuke_taper_radius, 1.0d0,
     &          fric_limit, presc_shear)
     
              rawsd%sd(node_s, node_d) = presc_shear
            
            endif
          
          ! Otherwise, in inner zone ...
          
          else
        
            ! Rupture time based on distance to hypocenter
          
            rupt_s = r_hypo_s / sinf_nuke_vel_s
            rupt_d = r_hypo_d / sinf_nuke_vel_d
        
            rupture_time = sqrt(rupt_s*rupt_s + rupt_d*rupt_d)
          
            ! Make sure the shear stress exceeds the dynamic friction
            ! (including cohesion) by at least the overstress parameter
            
            fric_limit
     &        = dyn_coeff + sinf_nuke_mu + (cohesion/normal_stress)
          
            if (presc_shear < fric_limit) then
     
              presc_shear = fric_limit
     
              rawsd%sd(node_s, node_d) = presc_shear
            
            endif
          
          endif
        
          ! If shear stress plus min_excess exceeds yield strength
          ! (neglecting cohesion), increase the static coefficient
        
          if (sta_coeff < presc_shear + min_excess) then
        
            sta_coeff = presc_shear + min_excess
          
            cnt_sta_adjust = cnt_sta_adjust + 1
          
          endif
          
          ! Note no depth conditioning in nuke zone
        
        ! Overstress method ...
        
        case (sinf_nuke_method_overstress)
        
          ! If in taper zone ...
          ! (Note +0.0001 ensures sinf_nuke_taper_radius <= 0.9999
          !  and nr_hypo >= 0.0001)
          
          if (nr_hypo > sinf_nuke_taper_radius + 0.0001d0) then
        
            ! Shear stress is at least static friction (including
            ! cohesion) plus overstress parameter, but taper this effect
            
            fric_limit
     &        = sta_coeff + sinf_nuke_mu + (cohesion/normal_stress)
            
            if (presc_shear < fric_limit) then
     
              presc_shear = lin_interp (
     &          nr_hypo, sinf_nuke_taper_radius, 1.0d0,
     &          fric_limit, presc_shear)
       
              rawsd%sd(node_s, node_d) = presc_shear
              
            endif
          
          ! Otherwise, in inner zone ...
          
          else
        
            ! Shear stress is at least static friction (including
            ! cohesion) plus overstress parameter
            
            fric_limit
     &        = sta_coeff + sinf_nuke_mu + (cohesion/normal_stress)
            
            if (presc_shear < fric_limit) then
       
              presc_shear = fric_limit
       
              rawsd%sd(node_s, node_d) = presc_shear
              
            endif
          
          endif
          
          ! Note no depth conditioning in nuke zone
        
        end select
      
      ! Otherwise, we're not in the nucleation zone ...
      
      else
        
        ! If shear stress plus min_excess exceeds yield strength
        ! (neglecting cohesion), increase the static coefficient
        
        if (sta_coeff < presc_shear + min_excess) then
        
          sta_coeff = presc_shear + min_excess
          
          cnt_sta_adjust = cnt_sta_adjust + 1
          
        endif
        
        ! Apply depth conditioning
        
        presc_shear = presc_shear * depth_cond
      
      endif
      
      ! If we're in the critical slip distance taper zone ...
      
      if (nr_hypo < sinf_dc_taper_radius_hi
     &    .and. sinf_nuke_method /= sinf_nuke_method_none) then
        
        ! If in outer zone ...
        ! (Note +0.0001 ensures sinf_dc_taper_radius_lo <= 
        !  sinf_dc_taper_radius_hi - 0.0001 and nr_hypo >= 0.0001)
        
        if (nr_hypo > sinf_dc_taper_radius_lo + 0.0001d0) then
          
          ! Apply to critical slip distance, but taper this effect
          
          sw_crit = sw_crit * lin_interp (
     &      nr_hypo, sinf_dc_taper_radius_lo, sinf_dc_taper_radius_hi,
     &      sinf_dc_taper_level, 1.0d0)
        
        ! Otherwise, in inner zone ...
        
        else
          
          ! Apply to critical slip distance
          
          sw_crit = sw_crit * sinf_dc_taper_level
        
        endif
      
      endif
      
      ! Fill record
      
      rec_friction%node_s = node_s
      rec_friction%node_d = node_d
      
      rec_friction%coord_node_s = xs
      rec_friction%coord_node_d = xd
      
      rec_friction%normal_stress = normal_stress
      
      rec_friction%shear_stress_s
     &  = presc_shear * sinf_smult_s * normal_stress
      rec_friction%shear_stress_d
     &  = presc_shear * sinf_smult_d * normal_stress
      
      rec_friction%stress_ratio_s = presc_shear * sinf_smult_s
      rec_friction%stress_ratio_d = presc_shear * sinf_smult_d
      
      rec_friction%sta_coeff = sta_coeff
      rec_friction%dyn_coeff = dyn_coeff
      
      rec_friction%sw_crit = sw_crit
      
      rec_friction%cohesion = cohesion
      
      rec_friction%rupture_time = rupture_time
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine sinf_fill_friction
      
      

      
      end module sinf_data




c     sinf_read_param - Read parameters for stress input file builder.
c     Parameters:
c      iunit = Fortran unit number.
c      f_ok = Variable receives true if success, false if error.
c
c     File format:
c       Pre-scaled stress target scale
c       Pre-scaled stress target level
c       Weights for stress scale [see sinf_weights_read]
c       Weights for stress level [see sinf_weights_read]
c       Minimum pre-scaled stress [see value2d_new]
c       Maximum pre-scaled stress [see value2d_new]
c       Nucleation method
c       Nucleation overstress parameter
c       Velocity of forced rupture
c       Formula for hypocenter radius
c       Nucleation factor
c       Nucleation zone aspect ratio strike/dip
c       Shear modulus for hypocenter radius calculation
c       Hypocenter selection mask radius along dip,strike
c       Hypocenter selection mask std dev along dip,strike
c       Limits for hypocenter location along dip lo,hi
c       Limits for hypocenter location along strike lo,hi
c       Limits for nucleation zone along dip lo,hi
c       Limits for nucleation zone along strike lo,hi
c       Limits for hypocenter stress averaging along dip lo,hi
c       Limits for hypocenter stress averaging along strike lo,hi
c       Stress multipliers along dip,strike
c       Normal stress [see value2d_new]
c       Static coefficient of friction [see value2d_new]
c       Dynamic coefficient of friction [see value2d_new]
c       Critical slip distance [see value2d_new]
c       Cohesion [see value2d_new]
c       Depth conditioning [see value2d_new]
c       Minimum strength excess coefficient [see value2d_new]

      subroutine sinf_read_param (iunit, f_ok)
      
      use io_unit_data
      use sinf_data
      implicit none
      
      integer, intent(in) :: iunit
      logical, intent(out) :: f_ok
      
      ! Target scale and level
      
      read(iunit,*) sinf_target_scale
      read(iunit,*) sinf_target_level
      
      ! Weights for stress scale
      
      call sinf_weights_new (sinf_weight_scale)
      call sinf_weights_read (sinf_weight_scale, iunit)
      
      ! Weights for stress level
      
      call sinf_weights_new (sinf_weight_level)
      call sinf_weights_read (sinf_weight_level, iunit)
      
      ! Minimum and maximum pre-scaled stress
      
      call value2d_new (sinf_presc_min, iunit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      call value2d_new (sinf_presc_max, iunit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Nucleation method
      
      read(iunit,*) sinf_nuke_method
      
      select case (sinf_nuke_method)
        case (sinf_nuke_method_time_drop)
        case (sinf_nuke_method_overstress)
        case (sinf_nuke_method_none)
        case default
          f_ok = .false.
          return
      end select
      
      ! Nucleation overstress parameter
      
      read(iunit,*) sinf_nuke_mu
      
      ! Velocity of forced rupture along dip, strike
      
      read(iunit,*) sinf_nuke_vel_d, sinf_nuke_vel_s
      
      ! Formula for hypocenter radius
      
      read(iunit,*) sinf_hrad_form
      
      select case (sinf_hrad_form)
        case (sinf_hrad_form_fixed)
        case (sinf_hrad_form_day)
        case default
          f_ok = .false.
          return
      end select
      
      ! Nucleation factor
      
      read(iunit,*) sinf_nuke_factor
      
      ! Nucleation zone aspect ratio
      
      read(iunit,*) sinf_nuke_aspect
      
      ! Shear modulus for hypocenter radius calculation
      
      read(iunit,*) sinf_nuke_lame_mu
      
      ! Nucleation zone taper radius
      
      read(iunit,*) sinf_nuke_taper_radius
      
      ! Critical slip distance taper radius lo,hi
      
      read(iunit,*) sinf_dc_taper_radius_lo, sinf_dc_taper_radius_hi
      
      ! Critical slip distance taper level
      
      read(iunit,*) sinf_dc_taper_level
      
      ! Hypocenter selection mask radius along dip,strike
      
      read(iunit,*) sinf_hypo_mrad_d, sinf_hypo_mrad_s
       
      ! Hypocenter selection mask standard deviation along dip,strike
      
      read(iunit,*) sinf_hypo_msdev_d, sinf_hypo_msdev_s
      
      ! Make hypocenter selection mask
      
      call sinf_make_hypo_mask ()
      
      ! Limits of hypocenter location
      
      read(iunit,*) sinf_hypo_d_lo, sinf_hypo_d_hi
      read(iunit,*) sinf_hypo_s_lo, sinf_hypo_s_hi
      
      ! Limits of nucleation zone
      
      read(iunit,*) sinf_nuke_d_lo, sinf_nuke_d_hi
      read(iunit,*) sinf_nuke_s_lo, sinf_nuke_s_hi
      
      ! Limits of hypocenter stress averaging
      
      read(iunit,*) sinf_stravg_d_lo, sinf_stravg_d_hi
      read(iunit,*) sinf_stravg_s_lo, sinf_stravg_s_hi
      
      ! Stress multipliers
      
      read(iunit,*) sinf_smult_d, sinf_smult_s
      
      ! Normal stress
      
      call value2d_new (sinf_normal_stress, iunit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Static coefficient of friction
      
      call value2d_new (sinf_sta_coeff, iunit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Dynamic coefficient of friction
      
      call value2d_new (sinf_dyn_coeff, iunit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Slip-weakening critical distance
      
      call value2d_new (sinf_sw_crit, iunit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Cohesion
      
      call value2d_new (sinf_cohesion, iunit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Depth conditioning
      
      call value2d_new (sinf_depth_cond, iunit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Minimum strength excess coefficient
      
      call value2d_new (sinf_min_excess, iunit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Success
      
      f_ok = .true.
      
      return
      end subroutine sinf_read_param








c----------------------------------------------------------------------
c
c     test_gen01 - Generator test #1.
c     Parameters:
c      in_unit = Unit from which to read parameters.
c      f_ok = Variable receives true if success, false if error.
c     The test generates the stress perturbation, selects the fault
c      surface and global phase using the selection template, and writes
c      the along-strike and along-dip perturbations into table files.
c
c     Parameter file format:
c       Random number generator parameters
c       Output filename for strike component
c       Output filename for dip component
c       Fault and Fourier parameters
c       Selection template parameters
c       Rolloff formula parameters

      subroutine test_gen01 (in_unit, f_ok)
      
      use io_unit_data
      use math_constants
      use ft_data
      use fault_data
      implicit none
      
      integer, intent(in) :: in_unit
      logical, intent(out) :: f_ok
      
      ! Fourier domain for stresses
      
      real(8), allocatable :: z(:,:,:)
      
      ! Fourier domain for template
      
      real(8), allocatable :: v(:,:,:)
      
      ! Fault plane
      
      real(8), allocatable :: w(:,:,:)
      
      ! Output filenames

      character(max_path) :: out_filename_strike
      character(max_path) :: out_filename_dip
      
      ! Output unit
      
      integer :: out_unit
      
      ! Mode generation outputs
      
      real(8) :: tau_ms_cont
      
      real(8) :: tau_ms_mean
      real(8) :: tau_ms_actual
      integer :: org0
      integer :: org1
      real(8) :: sel_conv
      
      real(8) :: stress_scale
      
      ! Read the random number generator parameters
      
      call rgen_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the output filenames

      read(in_unit,*) out_filename_strike
      read(in_unit,*) out_filename_dip
      
      ! Read the fault and Fourier parameters
      
      call fault_ft_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the selection template parameters
      
      call stpl_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the rolloff formula parameters
      
      call rolloff_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Allocate Fourier domain for stresses
      
      allocate(z(0:1, 0:ft_N1-1, 0:ft_N0-1))
      
      ! Allocate Fourier domain for template
      
      allocate(v(0:1, 0:ft_N1-1, 0:ft_N0-1))
      
      ! Allocate fault plane
      
      allocate(w(0:1, 0:cell_count1, 0:cell_count0))
      
      ! Calculate selection template
      
      call stpl_calc_v (v)
      
      ! Get continuum mean-square shear stress
      
      call rolloff_get_tau_ms_cont (tau_ms_cont)
      
      ! Generate modes and select optimum location and phase
      
      call ft_gen_sel_modes (z, tau_ms_mean, tau_ms_actual,
     &                       v, org0, org1, sel_conv)
      
      ! Transform stresses to spatial domain
      
      call ft_reverse_fft (z)
      
      ! Extract into fault surface
      
      stress_scale = sqrt(tau_ms_cont / tau_ms_mean)
      
      call fault_extract_from_ft (w, z, org0, org1, stress_scale)
      
      ! Write into tables
      
      call open_unit_w (out_unit, out_filename_strike)
      
      call fault_write_table_re (out_unit, w)
      
      call close_unit (out_unit)
      
      call open_unit_w (out_unit, out_filename_dip)
      
      call fault_write_table_im (out_unit, w)
      
      call close_unit (out_unit)
      
      ! Deallocate arrays
      
      deallocate(w)
      deallocate(v)
      deallocate(z)
      
      ! Success
        
      f_ok = .true.
      
      return
      end subroutine test_gen01







c----------------------------------------------------------------------
c
c     test_gen02 - Generator test #2.
c     Parameters:
c      in_unit = Unit from which to read parameters.
c      f_ok = Variable receives true if success, false if error.
c     The test generates the stress perturbation, selects the fault
c      surface and global phase using the selection template, and writes
c      the along-strike and along-dip perturbations into table files.
c      Then it calculates and writes statistics.
c
c     Parameter file format:
c       Random number generator parameters
c       Output filename for statistics
c       Output filename for strike component
c       Output filename for dip component
c       Fault and Fourier parameters
c       Selection template parameters
c       Rolloff formula parameters
c       Statistics parameters

      subroutine test_gen02 (in_unit, f_ok)
      
      use io_unit_data
      use math_constants
      use ft_data
      use fault_data
      use sstat_include
      use fzsc_data
      implicit none
      
      integer, intent(in) :: in_unit
      logical, intent(out) :: f_ok
      
      ! Fourier domain for stresses
      
      real(8), allocatable :: z(:,:,:)
      
      ! Fourier domain for template
      
      real(8), allocatable :: v(:,:,:)
      
      ! Fault plane
      
      real(8), allocatable :: w(:,:,:)
      
      ! Arbitrary fault plane
      
      real(8), allocatable :: w_arb(:,:,:)
      
      ! Output filenames

      character(max_path) :: out_filename_statistics
      character(max_path) :: out_filename_strike
      character(max_path) :: out_filename_dip
      
      ! Output unit
      
      integer :: out_unit
      
      ! Mode generation outputs
      
      real(8) :: tau_ms_cont
      
      real(8) :: tau_ms_mean
      real(8) :: tau_ms_actual
      integer :: org0
      integer :: org1
      real(8) :: sel_conv
      
      real(8) :: stress_scale
      
      ! Statistics outputs
      
      type(fzsc_info), pointer :: entire_fzsc
      type(fzsc_info), pointer :: main_fzsc
      type(fzsc_info), pointer :: stop_fzsc
      
      ! Allocate fixed-size objects
      
      call fzsc_info_new (entire_fzsc)
      call fzsc_info_new (main_fzsc)
      call fzsc_info_new (stop_fzsc)
      
      ! Read the random number generator parameters
      
      call rgen_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the output filenames

      read(in_unit,*) out_filename_statistics
      read(in_unit,*) out_filename_strike
      read(in_unit,*) out_filename_dip
      
      ! Read the fault and Fourier parameters
      
      call fault_ft_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the selection template parameters
      
      call stpl_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the rolloff formula parameters
      
      call rolloff_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the statistics parameters
      
      call sstat_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Allocate Fourier domain for stresses
      
      allocate(z(0:1, 0:ft_N1-1, 0:ft_N0-1))
      
      ! Allocate Fourier domain for template
      
      allocate(v(0:1, 0:ft_N1-1, 0:ft_N0-1))
      
      ! Allocate fault plane
      
      allocate(w(0:1, 0:cell_count1, 0:cell_count0))
      
      ! Allocate arbitrary fault plane
      
      allocate(w_arb(0:1, 0:cell_count1, 0:cell_count0))
      
      ! Calculate selection template
      
      call stpl_calc_v (v)
      
      ! Get continuum mean-square shear stress
      
      call rolloff_get_tau_ms_cont (tau_ms_cont)
      
      ! Generate modes and select optimum location and phase
      
      call ft_gen_sel_modes (z, tau_ms_mean, tau_ms_actual,
     &                       v, org0, org1, sel_conv)
      
      ! Transform stresses to spatial domain
      
      call ft_reverse_fft (z)
      
      ! Extract into fault surface
      
      stress_scale = sqrt(tau_ms_cont / tau_ms_mean)
      
      call fault_extract_from_ft (w, z, org0, org1, stress_scale)
      
      ! Extract arbitrary fault surface
      
      call fault_extract_from_ft (w_arb, z, 0, 0, stress_scale)
      
      ! Write statistics
      
      call open_unit_w (out_unit, out_filename_statistics)
      
      call rolloff_write_param (out_unit)
      
      write(out_unit,'(A)') '***** Stress generation results'
      write(out_unit,*)
      write(out_unit,'(A,1pe15.6)') 'tau_ms_cont = ',tau_ms_cont
      write(out_unit,'(A,1pe15.6)') 'tau_ms_mean = ',tau_ms_mean
      write(out_unit,'(A,1pe15.6)') 'tau_ms_actual = ',tau_ms_actual
      write(out_unit,'(A,1pe15.6)') 'stress_scale = ',stress_scale
      write(out_unit,'(A,I9)') 'org0 = ',org0
      write(out_unit,'(A,I9)') 'org1 = ',org1
      write(out_unit,'(A,1pe15.6)') 'sel_conv = ',sel_conv
      write(out_unit,*)
      
      write(out_unit,*)
      write(out_unit,*)
      write(out_unit,'(A)')
     &  '===== Statistics for selected fault surface ====='
      write(out_unit,*)
      write(out_unit,*)
      
!      call sstat_calc_and_write (out_unit, w)

      call sstat_calc_and_write_2 (
     &  out_unit, w, entire_fzsc, main_fzsc, stop_fzsc)
      
      write(out_unit,*)
      write(out_unit,*)
      write(out_unit,'(A)')
     &  '===== Statistics for Fourier domain ====='
      write(out_unit,*)
      write(out_unit,*)
      
      call sstat_calc_and_write_ft (out_unit, z, stress_scale)
      
!      write(out_unit,*)
!      write(out_unit,*)
!      write(out_unit,'(A)')
!     &  '===== Statistics for arbitrary fault surface ====='
!      write(out_unit,*)
!      write(out_unit,*)
!      
!      call sstat_calc_and_write (out_unit, w_arb)
      
      call close_unit (out_unit)
      
      ! Write into tables
      
      call open_unit_w (out_unit, out_filename_strike)
      
      call fault_write_table_re (out_unit, w)
      
      call close_unit (out_unit)
      
      call open_unit_w (out_unit, out_filename_dip)
      
      call fault_write_table_im (out_unit, w)
      
      call close_unit (out_unit)
      
      ! Deallocate objects
      
      deallocate(w_arb)
      deallocate(w)
      deallocate(v)
      deallocate(z)
      
      call fzsc_info_delete (stop_fzsc)
      call fzsc_info_delete (main_fzsc)
      call fzsc_info_delete (entire_fzsc)
      
      ! Success
        
      f_ok = .true.
      
      return
      end subroutine test_gen02







c----------------------------------------------------------------------
c
c     run_gen01 - Run generator method #1.
c     (Note, this is derived from test_gen02.)
c     Parameters:
c      in_unit = Unit from which to read parameters.
c      f_ok = Variable receives true if success, false if error.
c
c     Parameter file format:
c       Random number generator parameters:
c           Realization number
c           Seed filename
c       Maximum sub-realization number
c       Raw stress filename base, prefix
c       Statistics output filename base, prefix
c       Summary filename base, prefix
c       Fault and Fourier parameters
c       Selection template parameters
c       Rolloff formula parameters
c       Statistics parameters

      subroutine run_gen01 (in_unit, f_ok)
      
      use io_unit_data
      use math_constants
      use ft_data
      use fault_data
      use sstat_include
      use fzsc_data
      use vsel_data
      use rawsdist_data
      implicit none
      
      integer, intent(in) :: in_unit
      logical, intent(out) :: f_ok
      
      ! Fourier domain for stresses
      
      real(8), allocatable :: z(:,:,:)
      
      ! Fourier domain for template
      
      real(8), allocatable :: v(:,:,:)
      
      ! Fourier domain to hold template across sub-realizations
      
      real(8), allocatable :: v_hold(:,:,:)
      
      ! Fault plane
      
      real(8), allocatable :: w(:,:,:)
      
      ! Rotated fault plane
      
      real(8), allocatable :: w_rot(:,:,:)
      
      ! Raw stress distribution holder
      
      type(rawsdist_holder), pointer :: rawsd
      
      ! Parameters
      
      integer :: max_sub_realiz  ! Maximum sub-realization
      character(max_path) :: out_basename  ! Output file basename
      
      ! Raw stress filename
      
      character(max_path) :: raw_basename
      character(max_path) :: raw_postname
      
      ! Statistics output filename
      
      character(max_path) :: statout_basename
      character(max_path) :: statout_postname
      
      ! Summary filename
      
      character(max_path) :: summary_basename
      character(max_path) :: summary_postname
      
      ! Output filename

      character(max_path) :: out_filename
      
      ! Output unit
      
      integer :: out_unit
      integer :: sum_unit
      
      ! Realization
      
      integer :: realiz  ! Current realization
      integer :: sub_realiz  ! Current sub-realization
      
      ! Mode generation outputs
      
      real(8) :: tau_ms_cont
      
      real(8) :: tau_ms_mean
      real(8) :: tau_ms_actual
      integer :: org0
      integer :: org1
      real(8) :: sel_conv
      
      real(8) :: stress_scale
      
      integer :: rot_result

      ! Mode generation temporaries

      integer :: rot_org0
      integer :: rot_org1
      
      ! Statistics outputs
      
      type(fzsc_info), pointer :: entire_fzsc
      type(fzsc_info), pointer :: main_fzsc
      type(fzsc_info), pointer :: stop_fzsc
      
      ! Allocate fixed-size objects
      
      call fzsc_info_new (entire_fzsc)
      call fzsc_info_new (main_fzsc)
      call fzsc_info_new (stop_fzsc)
      
      ! Read the random number generator parameters
      
      call rgen_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      call rgen_get_realiz (realiz)
      
      ! Read the maximum sub-realization number

      read(in_unit,*) max_sub_realiz
      
      ! Read the raw stress distribution filename

      read(in_unit,*) raw_basename, raw_postname
      
      ! Read the statistics output filename

      read(in_unit,*) statout_basename, statout_postname
      
      ! Read the summary filename

      read(in_unit,*) summary_basename, summary_postname
      
      ! Read the fault and Fourier parameters
      
      call fault_ft_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the selection template parameters
      
      call stpl_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the rotation template parameters
      
      call rotpl_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the rolloff formula parameters
      
      call rolloff_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the statistics parameters
      
      call sstat_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Allocate Fourier domain for stresses
      
      allocate(z(0:1, 0:ft_N1-1, 0:ft_N0-1))
      
      ! Allocate Fourier domain for template
      
      allocate(v(0:1, 0:ft_N1-1, 0:ft_N0-1))
      
      ! Allocate half-size Fourier domain for holding template
      
      allocate(v_hold(0:1, 0:ft_N1-1, 0:ft_N0/2))
      
      ! Allocate fault plane
      
      allocate(w(0:1, 0:cell_count1, 0:cell_count0))
      
      ! Allocate rotated fault plane
      
      allocate(w_rot(0:1, 0:cell_count1, 0:cell_count0))
      
      ! Allocate raw stress distribution holder
      
      call rawsdist_holder_new (rawsd)
      
      ! Calculate selection template
      
      call stpl_calc_v (v)
      
      call ft_save_xreal (v, v_hold)
      
      ! Open the summary file
        
      call make_realiz_filename (out_filename, summary_basename,
     &    realiz, -1, summary_postname)
      
      call open_unit_w (sum_unit, out_filename)
      
      ! Loop over sub-realizations
      
      do sub_realiz = 0, max_sub_realiz
      
        ! Get continuum mean-square shear stress
      
        call rolloff_get_tau_ms_cont (tau_ms_cont)
      
        ! Generate modes and select optimum location and phase
      
        call ft_restore_xreal (v, v_hold)
      
        call ft_gen_sel_modes (z, tau_ms_mean, tau_ms_actual,
     &                         v, org0, org1, sel_conv)
      
        ! Transform stresses to spatial domain
      
        call ft_reverse_fft (z)
      
        ! Extract into fault surface
      
        stress_scale = sqrt(tau_ms_cont / tau_ms_mean)
      
        call fault_extract_from_ft (w, z, org0, org1, stress_scale)
        
        ! Do rotation if desired
        
        !!!call rotpl_do_rotation (w, vsel_real, rot_result)
        
        call rotpl_calc_rot_org (org0, org1, rot_org0, rot_org1)

        call fault_extract_rot_from_ft (
     &          w_rot, z, rot_org0, rot_org1, stress_scale)

        call rotpl_do_rotation_2 (w, w_rot, vsel_real, rot_result)
      
        ! Write statistics
        
        call make_realiz_filename (out_filename, statout_basename,
     &      realiz, sub_realiz, statout_postname)
      
        call open_unit_w (out_unit, out_filename)
      
        write(out_unit,'(A)') '***** Realization'
        write(out_unit,*)
        write(out_unit,'(A,I9)') 'realiz = ',realiz
        write(out_unit,'(A,I9)') 'sub_realiz = ',sub_realiz
        write(out_unit,*)
      
        call rolloff_write_param (out_unit)
      
        write(out_unit,'(A)') '***** Stress generation results'
        write(out_unit,*)
        write(out_unit,'(A,1pe15.6)') 'tau_ms_cont = ',tau_ms_cont
        write(out_unit,'(A,1pe15.6)') 'tau_ms_mean = ',tau_ms_mean
        write(out_unit,'(A,1pe15.6)') 'tau_ms_actual = ',tau_ms_actual
        write(out_unit,'(A,1pe15.6)') 'stress_scale = ',stress_scale
        write(out_unit,'(A,I9)') 'org0 = ',org0
        write(out_unit,'(A,I9)') 'org1 = ',org1
        write(out_unit,'(A,1pe15.6)') 'sel_conv = ',sel_conv
        write(out_unit,'(A,I9)') 'rot_result = ',rot_result
        write(out_unit,*)
      
        write(out_unit,*)
        write(out_unit,*)
        write(out_unit,'(A)')
     &    '===== Statistics for selected fault surface ====='
        write(out_unit,*)
        write(out_unit,*)

        call sstat_calc_and_write_2 (
     &    out_unit, w, entire_fzsc, main_fzsc, stop_fzsc)
      
        write(out_unit,*)
        write(out_unit,*)
        write(out_unit,'(A)')
     &    '===== Statistics for Fourier domain ====='
        write(out_unit,*)
        write(out_unit,*)
      
        call sstat_calc_and_write_ft (out_unit, z, stress_scale)
      
        call close_unit (out_unit)
        
        ! Fill in the raw statistics holder
        
        call rawsdist_holder_fill (
     &    rawsd, sub_realiz, w, vsel_real, sel_conv,
     &    entire_fzsc, main_fzsc, stop_fzsc)
      
        ! Write the raw statistics file
        
        call make_realiz_filename (out_filename, raw_basename,
     &      realiz, sub_realiz, raw_postname)
      
        call open_unit_w (out_unit, out_filename)
        
        call rawsdist_holder_write (rawsd, out_unit)
      
        call close_unit (out_unit)
        
        ! Write a record in the summary file
        
        write(sum_unit,'(I7,1x,I5,2(1x,1pe15.6))')
     &    realiz,
     &    sub_realiz,
     &    sel_conv,
     &    (main_fzsc%str_mean - stop_fzsc%str_mean)/main_fzsc%str_stddev
      
      enddo  ! end loop over sub-realizations
      
      ! Close the summary file
      
      call close_unit (sum_unit)
      
      ! Deallocate objects
      
      call rawsdist_holder_delete (rawsd)
      
      deallocate(w)
      deallocate(v_hold)
      deallocate(v)
      deallocate(z)
      
      call fzsc_info_delete (stop_fzsc)
      call fzsc_info_delete (main_fzsc)
      call fzsc_info_delete (entire_fzsc)
      
      ! Success
        
      f_ok = .true.
      
      return
      end subroutine run_gen01







c----------------------------------------------------------------------
c
c     run_fric01 - Run friction method #1.
c     Parameters:
c      in_unit = Unit from which to read parameters.
c      f_ok = Variable receives true if success, false if error.
c
c     Parameter file format:
c       Stress distribution list parameters:
c           Realization number lo, hi
c           Raw stress filename base, prefix
c       Pre-scaled stress filename base, prefix
c       Stress input filename base, prefix
c       Statistics output filename base, prefix
c       Summary filename base, prefix
c       Stress input file builder parameters
c       Statistics parameters (used for recalculation)

      subroutine run_fric01 (in_unit, f_ok)
      
      use io_unit_data
      use math_constants
      use fzsc_data
      use sstat_include
      use rawsdist_data
      use sdlist_data
      use sifmt_data
      use sinf_data
      implicit none
      
      integer, intent(in) :: in_unit
      logical, intent(out) :: f_ok
      
      ! Records for stress input file
      
      type(sifmt_realiz), pointer :: rec_realiz
      type(sifmt_faultsize), pointer :: rec_faultsize
      type(sifmt_hypo), pointer :: rec_hypo
      type(sifmt_friction), pointer :: rec_friction
      
      ! Pre-scaled stress filename
      
      character(max_path) :: presc_basename
      character(max_path) :: presc_postname
      
      ! Stress input filename
      
      character(max_path) :: sinfile_basename
      character(max_path) :: sinfile_postname
      
      ! Statistics output filename
      
      character(max_path) :: statout_basename
      character(max_path) :: statout_postname
      
      ! Summary filename
      
      character(max_path) :: summary_basename
      character(max_path) :: summary_postname
      
      ! Output filename
      
      character(max_path) :: out_filename
      
      ! Output units
      
      integer :: out_unit
      integer :: sum_unit
      
      ! Current raw stress distribution
      
      type(rawsdist_holder), pointer :: rawsd
      
      ! Stress input file builder outputs
      
      integer :: cnt_clip_min
      integer :: cnt_clip_max
      
      integer :: cnt_nuke_node
      integer :: cnt_sta_adjust
      
      ! Index into fault
      
      integer :: node_s
      integer :: node_d
      
      ! Allocate fixed-size objects
      
      call sifmt_realiz_new (rec_realiz)
      call sifmt_faultsize_new (rec_faultsize)
      call sifmt_hypo_new (rec_hypo)
      call sifmt_friction_new (rec_friction)
      
      ! Read the stress distribution list parameters
      
      call sdlist_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the pre-scaled stress filename

      read(in_unit,*) presc_basename, presc_postname
      
      ! Read the stress input filename

      read(in_unit,*) sinfile_basename, sinfile_postname
      
      ! Read the statistics output filename

      read(in_unit,*) statout_basename, statout_postname
      
      ! Read the summary filename

      read(in_unit,*) summary_basename, summary_postname
      
      ! Read the stress input file builder parameters
      
      call sinf_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Read the statistics parameters
      
      call sstat_read_param (in_unit, f_ok)
      
      if (.not. f_ok) then
        return
      endif
      
      ! Open the summary file
        
      call make_realiz_filename (out_filename, summary_basename,
     &    sdlist_head%realiz, -1, summary_postname)
      
      call open_unit_w (sum_unit, out_filename)
      
      ! Loop over raw stress distributions
      
      rawsd => sdlist_head
      
      do while (associated(rawsd))
        
        ! Pre-scale the distribution
        
        call sinf_prescale_stress (
     &    rawsd, f_ok, cnt_clip_min, cnt_clip_max)
        
        if (.not. f_ok) then
          return
        endif
        
        ! Find hypocenter location
        
        call sinf_find_hypo (rawsd, f_ok)
        
        if (.not. f_ok) then
          return
        endif
        
        ! Open stress input file
        
        call make_realiz_filename (out_filename, sinfile_basename,
     &      rawsd%realiz, rawsd%sub_realiz, sinfile_postname)
      
        call open_unit_w (out_unit, out_filename)
        
        ! Write the realization record
        
        call sinf_fill_realiz (rec_realiz, rawsd, f_ok)
        
        if (.not. f_ok) then
          return
        endif
        
        call sifmt_realiz_write (rec_realiz, out_unit)
        
        ! Write the fault size record
        
        call sinf_fill_faultsize (rec_faultsize, rawsd, f_ok)
        
        if (.not. f_ok) then
          return
        endif
        
        call sifmt_faultsize_write (rec_faultsize, out_unit)
        
        ! Write the hypocenter record
        
        call sinf_fill_hypo (rec_hypo, rawsd, f_ok)
        
        if (.not. f_ok) then
          return
        endif
        
        call sifmt_hypo_write (rec_hypo, out_unit)
        
        ! Write the friction records, one per node
        
        cnt_nuke_node = 0
        cnt_sta_adjust = 0
      
        do node_d = 0, rawsd%cell_count_d
          do node_s = 0, rawsd%cell_count_s
            
            call sinf_fill_friction (
     &          rec_friction, rawsd, node_s, node_d, f_ok,
     &          cnt_nuke_node, cnt_sta_adjust)
        
            if (.not. f_ok) then
              return
            endif
            
            call sifmt_friction_write (rec_friction, out_unit)
            
          enddo
        enddo
        
        ! Close the stress input file
      
        call close_unit (out_unit)
      
        ! Write statistics
        
        call make_realiz_filename (out_filename, statout_basename,
     &      rawsd%realiz, rawsd%sub_realiz, statout_postname)
      
        call open_unit_w (out_unit, out_filename)
      
        write(out_unit,'(A)') '***** Realization'
        write(out_unit,*)
        write(out_unit,'(A,I9)') 'realiz = ',rawsd%realiz
        write(out_unit,'(A,I9)') 'sub_realiz = ',rawsd%sub_realiz
        write(out_unit,*)
      
        write(out_unit,'(A)') '***** Friction builder results'
        write(out_unit,*)
        write(out_unit,'(A,I9)') 'cnt_clip_min = ',cnt_clip_min
        write(out_unit,'(A,I9)') 'cnt_clip_max = ',cnt_clip_max
        write(out_unit,'(A,I9)') 'cnt_nuke_node = ',cnt_nuke_node
        write(out_unit,'(A,I9)') 'cnt_sta_adjust = ',cnt_sta_adjust
        write(out_unit,*)
      
        write(out_unit,*)
        write(out_unit,*)
        write(out_unit,'(A)')
     &    '===== Statistics for pre-scaled stress ====='
        write(out_unit,*)
        write(out_unit,*)

        call rsdh_recalc_stats_and_write (rawsd, out_unit)
      
        write(out_unit,*)
        write(out_unit,*)
      
        call close_unit (out_unit)
      
        ! Write the pre-scaled statistics file
        
        call make_realiz_filename (out_filename, presc_basename,
     &      rawsd%realiz, rawsd%sub_realiz, presc_postname)
      
        call open_unit_w (out_unit, out_filename)
        
        call rawsdist_holder_write (rawsd, out_unit)
      
        call close_unit (out_unit)
        
        ! Write a record in the summary file
        
        write(sum_unit,'(I7,1x,I5,8(1x,1pe15.6))')
     &    rawsd%realiz,
     &    rawsd%sub_realiz,
     &    rawsd%main_fzsc%str_mean,
     &    rawsd%main_fzsc%str_stddev,
     &    rawsd%stop_fzsc%str_mean,
     &    rawsd%stop_fzsc%str_stddev,
     &    rawsd%hypo_d * rawsd%cell_size_d,
     &    rawsd%hypo_s * rawsd%cell_size_s,
     &    rawsd%hypo_radius_d,
     &    rawsd%hypo_radius_s
        
        ! Next stress distribution
        
        rawsd => rawsd%next
      
      enddo  ! end loop over raw stress distributions
      
      ! Close the summary file
      
      call close_unit (sum_unit)
      
      ! Deallocate objects
      
      call sifmt_friction_delete (rec_friction)
      call sifmt_hypo_delete (rec_hypo)
      call sifmt_faultsize_delete (rec_faultsize)
      call sifmt_realiz_delete (rec_realiz)
      
      ! Success
        
      f_ok = .true.
      
      return
      end subroutine run_fric01







c----------------------------------------------------------------------
c
c     ths_gen_main - Main program to generate heterogeneous stress.
c
c     The program reads parameters from standard input, one per line.
c     The first parameter is a command that specifies the operation to
c     perform.  Subsequent parameters vary depending on the command.
c     Commands are case-sensitive (darn Fortran!).
c
c     Hello command:
c       Command = "hello"
c
c     Test 2D FFT command:
c       Command = "test_fft_2d"
c       Extent in dimension #1, a power of 2
c       Extent in dimension #2, a power of 2
c       Initial index in dimension #1 (0-based)
c       Initial index in dimension #2 (0-based)
c       Index step in dimension #1
c       Index step in dimension #2
c       Number of tests
c
c     Test exponential H calculation:
c       Command = "test_texpd_H"
c       Number of intervals (at least 1)
c
c     Test log-square rolloff formula:
c       Command = "test_roll_lsq"
c       Correlation wavelength, in meters
c       Seismogenic wavelength constant, in meters
c       Seismogenic wavelength slope
c
c     Test rolloff parameters:
c       Command = "test_roll_param"
c       Rolloff parameters [see rolloff_read_param]
c
c     Test selection template and plot:
c       Command = "test_stpl_plot"
c       Output filename (for table file)
c       Fault and Fourier parameters
c       Selection template parameters
c
c     Generator test #1:
c       Command = "test_gen01"
c       Random number generator parameters
c       Output filename for strike component
c       Output filename for dip component
c       Fault and Fourier parameters
c       Selection template parameters
c       Rolloff formula parameters
c
c     Generator test #2:
c       Command = "test_gen02"
c       Random number generator parameters
c       Output filename for strike component
c       Output filename for dip component
c       Fault and Fourier parameters
c       Selection template parameters
c       Rolloff formula parameters
c       Statistics parameters
c
c     Run generator method #1
c       Command = "run_gen01"
c       Random number generator parameters:
c           Realization number
c           Seed filename
c       Maximum sub-realization number
c       Raw stress filename base, prefix
c       Statistics output filename base, prefix
c       Summary filename base, prefix
c       Fault and Fourier parameters
c       Selection template parameters
c       Rolloff formula parameters
c       Statistics parameters
c
c     Run friction method #1
c       Command = "run_fric01"
c       Stress distribution list parameters:
c           Realization number lo, hi
c           Raw stress filename base, prefix
c       Pre-scaled stress filename base, prefix
c       Stress input filename base, prefix
c       Statistics output filename base, prefix
c       Summary filename base, prefix
c       Stress input file builder parameters
c       Statistics parameters (used for recalculation)
      
      subroutine ths_gen_main ()
      
      use io_unit_data
      implicit none
      
      ! Command

      character(300) :: prog_command
      
      ! Flag for function returns
      
      logical :: f_ok
      
      ! Parameters for test 2D FFT command
      
      integer :: tfft2d_ext_1
      integer :: tfft2d_ext_2
      integer :: tfft2d_init_1
      integer :: tfft2d_init_2
      integer :: tfft2d_step_1
      integer :: tfft2d_step_2
      integer :: tfft2d_n_test
      
      ! Workspace for test 2D FFT command
      
      integer :: tfft2d_iret  ! return code
      
      ! Parameters for test exponential H calculation
      
      integer :: texpdH_n
      
      ! Workspace for test exponential H calculation
      
      real(8) :: texpdH_H
      real(8) :: texpdH_g
      real(8) :: texpdH_g2
      real(8) :: texpdH_err
      integer :: texpdH_i  ! index for test number
      integer :: texpdH_m  ! index for power series evaluation
      real(8) :: texpdH_x  ! power for power series evaluation
      
      ! Parameters for test log-square rolloff formula
      
      real(8) :: tlsqr_corr_lambda
      real(8) :: tlsqr_seis_b
      real(8) :: tlsqr_seis_m
      integer :: tlsqr_n_test
      
      ! Workspace for test log-square rolloff formula
      
      real(8) :: tlsqr_roll_lambda  ! Rolloff wavelength in meters
      real(8) :: tlsqr_seis_lambda  ! Seismogenic wavelength in meters
      real(8) :: tlsqr_integ_ksq  ! Integral of (k^2)*g(k)
      integer :: tlsqr_i  ! index for test number
      real(8) :: tlsqr_corr_lambda_i  ! wavelength for current test
      
      ! Workspace for test rolloff parameters
      
      real(8) :: trollp_corr_lambda  ! Correlation wavelength in meters
      real(8) :: trollp_roll_lambda  ! Rolloff wavelength in meters
      real(8) :: trollp_seis_lambda  ! Seismogenic wavelength in meters
      real(8) :: trollp_integ_ksq  ! Integral of (k^2)*g(k)
      
      
      ! Read program command from standard input      

      read(stdin,*) prog_command
      
      
      !===== Hello command =====
      
      if (prog_command == "hello") then
      
      write(stdout,'(A)')
     &  'Hello from thermodynamic stress generation!'
      
      
      !===== Test 2D FFT command =====
      
      else if (prog_command == "test_fft_2d") then
      
      write(stdout,'(A)')
     &  'Testing 2D FFT functions ...'
      
      read(stdin,*) tfft2d_ext_1
      read(stdin,*) tfft2d_ext_2
      read(stdin,*) tfft2d_init_1
      read(stdin,*) tfft2d_init_2
      read(stdin,*) tfft2d_step_1
      read(stdin,*) tfft2d_step_2
      read(stdin,*) tfft2d_n_test
      
      ! Run the test
      
      tfft2d_iret = 0  ! assume failure
      
      call test_fft_2dim (
     &  tfft2d_iret, tfft2d_ext_1, tfft2d_ext_2,
     &  tfft2d_init_1, tfft2d_init_2, tfft2d_step_1, tfft2d_step_2,
     &  tfft2d_n_test)
      
      ! Display the result
      
      if (tfft2d_iret /= 0) then
        write(stdout,'(A)')
     &    'Test of 2D FFT functions PASSED.'
      else
        write(stdout,'(A)')
     &    'Test of 2D FFT functions FAILED.'
      endif
      
      
      !===== Test exponential H calculation =====
      
      else if (prog_command == "test_texpd_H") then
      
      read(stdin,*) texpdH_n
      
      write(stdout,'(A)')
     &  'Testing trexpd H calculation, output = G, H, G2, ERR ...'
      
      ! Loop over tests
      
      do texpdH_i = 0, texpdH_n
        
        ! Invoke the function
        
        texpdH_g = dble(texpdH_i)/dble(texpdH_n)
        call trexpd_calc_H (texpdH_H, texpdH_g)
        
        ! Calculate the g from the returned H
        
        if (texpdH_i == texpdH_n) then
        
          ! Value should be 1 for last test
          
          texpdH_g2 = 1.0d0
        
        else if (texpdH_H > 0.1d0) then
        
          ! Calculate from formula
          
          texpdH_g2=1.0d0+(1.0d0-texpdH_H)*log(1.0d0-texpdH_H)/texpdH_H
          
        else
        
          ! Calculate from power series
          
          texpdH_g2 = 0.0d0
          texpdH_x = 1.0d0
          
          do texpdH_m = 1, 16
            texpdH_x = texpdH_x * texpdH_H
            texpdH_g2 = texpdH_g2 + texpdH_x/dble(texpdH_m*(texpdH_m+1))
          enddo
        
        endif
        
        ! Error
        
        texpdH_err = texpdH_g - texpdH_g2
        
        ! Display the result
      
        write(stdout,'(4(1x,1pe24.15))')
     &    texpdH_g, texpdH_H, texpdH_g2, texpdH_err
      
      enddo
      
      
      !===== Test log-square rolloff formula =====
      
      else if (prog_command == "test_roll_lsq") then
      
      read(stdin,*) tlsqr_corr_lambda
      read(stdin,*) tlsqr_seis_b
      read(stdin,*) tlsqr_seis_m
      read(stdin,*) tlsqr_n_test
      
      write(stdout,'(A)')
     &  'Testing log-square rolloff formula, output = '
     &  // 'corr_lambda, roll_lambda, seis_lambda, integ_ksq'
      
      ! Loop over tests
      
      do tlsqr_i = 1, tlsqr_n_test
      
        ! Correlation wavelength for this test
        
        tlsqr_corr_lambda_i
     &    = tlsqr_corr_lambda*dble(tlsqr_i)/dble(tlsqr_n_test)
        
        ! Set up the formula
        
        call rolloff_set_logsq (
     &    tlsqr_corr_lambda_i, tlsqr_seis_b, tlsqr_seis_m,
     &    tlsqr_roll_lambda, tlsqr_seis_lambda)
        
        ! Get the integral
        
        call rolloff_get_integ_ksq (tlsqr_integ_ksq)
        
        ! Display the result
      
        write(stdout,'(4(1x,1pe24.15))')
     &    tlsqr_corr_lambda_i, tlsqr_roll_lambda,
     &    tlsqr_seis_lambda, tlsqr_integ_ksq
      
      enddo
      
      
      !===== Test rolloff parameters =====
      
      else if (prog_command == "test_roll_param") then
      
      call rolloff_read_param (stdin, f_ok)
      
      if (.not. f_ok) then
        write(stdout,'(A)')
     &    'Call to rolloff_read_param failed.'
        return
      endif
      
      write(stdout,'(A)')
     &  'Testing rolloff parameters, output = '
     &  // 'corr_lambda, roll_lambda, seis_lambda, integ_ksq'
      
      ! Get the parameters
      
      call rolloff_get_corr_lambda (trollp_corr_lambda)
      call rolloff_get_roll_lambda (trollp_roll_lambda)
      call rolloff_get_seis_lambda (trollp_seis_lambda)
      
      ! Get the integral
        
      call rolloff_get_integ_ksq (trollp_integ_ksq)
        
      ! Display the result
      
      write(stdout,'(4(1x,1pe24.15))')
     &  trollp_corr_lambda, trollp_roll_lambda,
     &  trollp_seis_lambda, trollp_integ_ksq
      
      
      !===== Test selection template and plot =====
      
      else if (prog_command == "test_stpl_plot") then
      
      call stpl_test_plot (stdin, f_ok)
      
      if (.not. f_ok) then
        write(stdout,'(A)')
     &    'Call to stpl_test_plot failed.'
        return
      endif
      
      
      !===== Generator test #1 =====
      
      else if (prog_command == "test_gen01") then
      
      call test_gen01 (stdin, f_ok)
      
      if (.not. f_ok) then
        write(stdout,'(A)')
     &    'Call to test_gen01 failed.'
        return
      endif
      
      
      !===== Generator test #2 =====
      
      else if (prog_command == "test_gen02") then
      
      call test_gen02 (stdin, f_ok)
      
      if (.not. f_ok) then
        write(stdout,'(A)')
     &    'Call to test_gen02 failed.'
        return
      endif
      
      
      !===== Run generator method #1 =====
      
      else if (prog_command == "run_gen01") then
      
      call run_gen01 (stdin, f_ok)
      
      if (.not. f_ok) then
        write(stdout,'(A)')
     &    'Call to run_gen01 failed.'
        return
      endif
      
      
      !===== Run friction method #1 =====
      
      else if (prog_command == "run_fric01") then
      
      call run_fric01 (stdin, f_ok)
      
      if (.not. f_ok) then
        write(stdout,'(A)')
     &    'Call to run_fric01 failed.'
        return
      endif
     
      
      !===== Unrecognized command =====
      
      else
      
      write(stdout,'(A,A,A)')
     &  'Unrecognized command: "',
     &  prog_command(1:len_trim(prog_command)),
     &  '"'
      
      endif
      
    
      ! Done
      
      return
      end subroutine ths_gen_main











